
 /*
 File: NIF.C

 An example usage of the "info" structure pointer passed to the
 "main" function. Information is acquired on each node and sent
 to the root for printing. Probes are sent and recieved on the
 links to find the adjacent node numbers and link connect. The
 output is in the Logical Systems NIF file format. If redirected
 to a file, it can be used as a NIF file template for the
 existing network.

 Written by Scott Lloyd of Alta Technology Corp.

 4-30-93
 */

#define ANSI
#include <stdio.h>
#include <conc.h>
#include <flood.h>

#define HOST  0
#define NONE -1
#define WS   (10*sizeof(int))  /* work space for probe processes */

enum {FALSE,TRUE};

typedef struct {
    int node;
    int link;
} PROBE;

typedef struct {    /* structure passed to root for printing */
    int id;
    PROBE probe[4];
    int proc;
    int memsize;
    int last;
} NODE;

void printnode(NODE *node)  /* print NODE structure */
{
    int i,tmp;
    static char *proctype[] = {
        "T2X2","T414","T800","T425","T805","T801","T???","T225","T400"
    /*   -30    -20    -10      0     10     20     30     40     50   */
    };

    printf("%6d,        ?.tld,     R0",node->id);
    for (i = 0; i < 4; i++) {
        tmp = node->probe[i].node;
        switch (tmp) {
        case NONE: printf(",       "); break;
        case HOST: printf(",      0"); break;
        default:   printf(",%4d[%d]",tmp,node->probe[i].link);
        }
    }
    printf("; %s",proctype[node->proc/10+3]); /* strip mask rev, make index */
    printf("%7dK\n",node->memsize);
}

void outprobe(Process *p, Channel *c, PROBE probe)
{
    ChanOut(c,&probe,(int)sizeof(PROBE));
}

void inprobe(Process *p, Channel *c, PROBE *probe)
{
    ChanIn(c,probe,(int)sizeof(PROBE));
}

void findlinks(NODE *node, NDST *info)
{
    int i,j,k;
    Channel *in,*out;
    PROBE prb;
    Process *pout,*pin;
    Process *plist[9];

    prb.node = info->nodeid;
    j = 0;
    for (i = 0; i < 4; i++) {            /* send probes between adjacent */
        out = (Channel *)info->minint+i; /* nodes to identify link connect. */
        in = (Channel *)info->minint+4+i;
        if (info->nodeid == 1 && (Channel *)info->bootin == in) {
            node->probe[i].node = HOST;
        } else {
            node->probe[i].node = NONE;  /* init probe structures. */
            prb.link = i;
            pout = ProcAlloc(outprobe,(int)WS,3,out,prb);
            pin = ProcAlloc(inprobe,(int)WS,2,in,&node->probe[i]);
            for (k = 0; info->bootout[k] != 0 && info->bootout[k] != out; k++);
            if (info->bootin == in || info->bootout[k] == out) {
                plist[j++] = pout;       /* run later and resync */
                plist[j++] = pin;        /* if link part of boot tree. */
            } else {
                ProcRun(pout);           /* run now if link not part */
                ProcRun(pin);            /* of boot tree. */
            }
        }
    }
    plist[j] = (Process *)0;             /* terminate list. */
    if (j > 0) ProcParList(plist);       /* exchange probes in boot tree. */
}

int getdevid(void)
{
#pragma asm
        .val    TEMPA,0
        .val    TEMPB,1
        ajw     -2
        ldc     -10             ; pre-assign id for T800
        ldc     -20             ; pre-assign id for T212, T222, T414
        ldc     0
        lddevid                 ; load device id
        stl     TEMPA           ; TEMPA = possible device id
        stl     TEMPB           ; TEMPB = possible pre-assigned device id
        ldl     TEMPB
        cj      @L1             ; if lddevid not valid
        ldl     TEMPB           ; TEMPA = one of pre-assigned ids
        stl     TEMPA
        ldl     TEMPA           ; if id equ -20
        eqc     -20
        cj      @L1
        ldc     1               ; if word size is 2 bytes
        bcnt
        eqc     2
        cj      @L1
        ldc     -30             ; TEMPA = id for T2X2
        stl     TEMPA
L1:     ldl     TEMPA           ; return device id
        ajw     2
#pragma endasm
}


void main(int argc, char *argv[], char *envp[], NDST *info)
{
    int i;
    NODE node;

    findlinks(&node,info);
    node.id = info->nodeid;
    node.memsize = (info->memory - info->minint) / 1024;
    node.proc = getdevid();
    node.last = FALSE;
    if (info->nodeid == 1) {
        printf("; Node     Load-file   Reset   Link0"
            "   Link1   Link2   Link3  Proc  Memory\n");
        printnode(&node);                          /* print this node */
    } else {
        ChanOut((Channel *)info->bootin-4,         /* send this node up */
            &node,(int)sizeof(node));
    }
    for (i = 0; info->bootout[i] != 0; i++) {      /* for each branch */
        for (;;) {
            ChanIn((Channel *)info->bootout[i]+4,  /* recv other node */
                &node,(int)sizeof(node));
            if (node.last == TRUE) break;
            if (info->nodeid == 1) {
                printnode(&node);                  /* print other node */
            } else {
                ChanOut((Channel *)info->bootin-4, /* send other node up */
                    &node,(int)sizeof(node));
            }
        }
    }
    if (info->nodeid != 1) {
        node.last = TRUE;                /* signal no more nodes on branch */
        ChanOut((Channel *)info->bootin-4,
            &node,(int)sizeof(node));
    }
}
