
 /*
 File: MAND.C

 Mandelbrot compute program. This program takes an area to be
 rendered and sends out job packets of pixel blocks covering the
 area to all available nodes in the network. The pixel blocks are
 sent back to the host. The job packets are load balanced across
 all the nodes. Job packets from the feed process are routed to
 only one job process. Data packets are routed to all the job
 processes.

 Written by Scott Lloyd of Alta Technology Corp.

 4-30-92
 */

#define ANSI
#include <string.h>
#include <flood.h>
#include <ldbal.h>
#include "mdefs.h"

typedef struct {
    int x,y,w,h;
} BLOCK, *BLOCKPTR;

typedef struct {
    real Cx,Cy,Sx,Sy;
} AREA, *AREAPTR;

typedef struct {
    int mod,off;
} PMOD, *PMODPTR;

typedef struct {
    real r,i;
} COMPLEX, *COMPLEXPTR;

#define BPW ((int)sizeof(int)) /* bytes per word */
#define FWS (25*BPW)           /* feed workspace */
#define JWS (80*BPW)           /* job workspace */

#define FBS ((int)sizeof(AREA))        /* feed buffer size, largest message */
#define JBS ((int)sizeof(BLOCK)+MAXPIX)/* job buffer size, largest message */

#define loop        for (;;)
#define LinkNo(c)   ((c) >> 2 & 3)  /* Link No. from (Channel*) */

enum {FALSE,TRUE};


void main(NDST *info)
{
    void feed(),job();
    int *ptr;
    int uplk;   /* up link number to root */
    int dnlk;   /* down links to leaves (bit set) */
    LBPRC fd;   /* parameters for creating feed process */
    LBPRC jb;   /* parameters for creating job process */

    fd.proc = feed; fd.wkspace = FWS; fd.bufsize = FBS;
    jb.proc = job;  jb.wkspace = JWS; jb.bufsize = JBS;
    uplk = LinkNo(info->bootin);
    for (ptr = info->bootout, dnlk = 0; *ptr; ptr++) {
        dnlk |= 1 << LinkNo(*ptr);
    }
    switch (info->nodeid) { /* init load-balance, process structure */
    case 1:
        InitBal(&fd,&jb,NOLBPRC,FBS,JBS,uplk,dnlk);     /* root; feed and job*/
    default:
        InitBal(NOLBPRC,&jb,NOLBPRC,FBS,JBS,uplk,dnlk); /* others; job only */
    }
}                                               /* InitBal never returns */

void feed(Process *p, char *buf)
{
    int xend;   /* right edge of viewport */
    int yend;   /* bottom edge of viewport */
    int pbw;    /* pixel block width */
    int pbh;    /* pixel block height */
    BLOCK vp;   /* viewport */
    BLOCK pb;   /* pixel block */
    LBHDR hdr;  /* header for load-balance messages */

    loop {
        FeedIn(&hdr,buf);   /* recieve message from host */
        switch (hdr.type) {
        case RENDER_P:  /* send out pixel blocks covering viewport area */
            pbw = ((int *)buf)[0];
            pbh = ((int *)buf)[1];
            if (pbw * pbh > MAXPIX) continue;   /* ignore if block to large */
            hdr.len = sizeof(pb);
            for (pb.y = vp.y, yend = vp.y+vp.h; pb.y < yend; pb.y += pbh) {
                pb.h = (pb.y+pbh > yend) ? yend-pb.y : pbh;
                for (pb.x = vp.x, xend = vp.x+vp.w; pb.x < xend; pb.x += pbw) {
                    pb.w = (pb.x+pbw > xend) ? xend-pb.x : pbw;
                    FeedOut(&hdr,&pb);
                }
            }
            hdr.len = 0;            /* flush all job packets through farm */
            hdr.type = FLUSH_P;
            break;
        case VIEWPORT_P:
            vp = *(BLOCKPTR)buf;    /* save viewport */
            break;
        }
        FeedOut(&hdr,buf);  /* send to job processes in farm */
    }
}

void job(Process *p, char *buf)
{
    int imax;   /* maximum iterations */
    PMOD pm;    /* pixel modifier */
    BLOCK vp;   /* viewport */
    BLOCK pb;   /* pixel block */
    AREA ar;    /* area */
    LBHDR hdr;  /* header for load-balance message */

    loop {
        JobIn(&hdr,buf);
        switch (hdr.type) {
        case RENDER_P:                  /* render packet to work on */
            {
                extern int iterate(COMPLEX c, int max);
                BLOCKPTR pb;
                char *pix;
                int x,y;
                real xinc,yinc;
                real xbeg,ybeg;
                COMPLEX c;

                pb = (BLOCKPTR)buf;         /* save pointer to pixel block */
                pix = buf+sizeof(BLOCK);    /* compute value of each pixel */
                xinc = ar.Sx*(real)2.0/(vp.w-1);
                yinc = ar.Sy*(real)2.0/(vp.h-1);
                xbeg = (pb->x - vp.x)*xinc - ar.Sx + ar.Cx;
                ybeg = (vp.y + vp.h - pb->y - pb->h)*yinc - ar.Sy + ar.Cy;
                for (y = pb->h-1; y >= 0; y--) {
                    c.i = y*yinc + ybeg;
                    for (x = 0; x < pb->w; x++) {
                        c.r = x*xinc + xbeg;
                        *pix++ = iterate(c,imax) % pm.mod + pm.off;
                    }
                }
                hdr.len += pb->w * pb->h;
            }
            break;                      /* send results back to host */
        case VIEWPORT_P:
            vp = *(BLOCKPTR)buf;        /* save viewport */
            continue;
        case AREA_P:
            ar = *(AREAPTR)buf;         /* save area */
            continue;
        case PMOD_P:
            pm = *(PMODPTR)buf;         /* save pixel modifier */
            continue;
        case MAXITER_P:
            imax = *(int *)buf;         /* save max iterations */
            continue;
        }
        JobOut(&hdr,buf);               /* send message to root or host */
    }
}

int iterate(COMPLEX c, int max)         /* iterate on chaos */
{
    int i;
    COMPLEX z;
    real zrsq,zisq;

    i = 1;
    z.r = c.r;
    z.i = c.i;
    zrsq = z.r*z.r; zisq = z.i*z.i;
    while ((i < max) && ((zrsq+zisq) < (real)4.0)) {
        z.i = z.r*z.i*(real)2.0 + c.i;
        z.r = zrsq - zisq + c.r;
        zrsq = z.r*z.r; zisq = z.i*z.i;
        i++;
    }
    if (i == max) i = 0;
    return(i);
}
