
 /*
 File: MSRV.C

 Mandelbrot server for Mandelbrot compute program. This server
 sends the area to be rendered to the compute nodes which return
 pixel blocks to be displayed on the screen. A mouse is used to
 point to the area to be magnified.

 Written by Scott Lloyd of Alta Technology Corp.

 4-30-93
 */

#include <graph.h>  /* _setcolor, _setpixel, _setvideomode, _getvideomode, */
                    /* _remapallpalette */
#include <signal.h> /* SIGTERM, SIGINT, SIGABRT, signal */
#include <stdlib.h> /* exit, getenv, malloc */
#include <stdio.h>  /* NULL, cputs, sprintf */
#include "mdefs.h"

#define LINK_NAME "LINKNAME"

#define aspect 0.75     /* aspect ratio (y/x) of screen */

#define loop for (;;)

int id = -1;            /* id for link */
int crs = 0;            /* use cross cursor */
char *lname;            /* link name */
char tmsg[80] = {0};    /* terminate message */
short trymode[] = {0x00,0x12,0x10,0x06,0x04};   /* video modes to try */

int atox(char *str);
void terminate(int sig);
void fexit(char *msg1, char *msg2);
void mandel();


void main(int argc, char *argv[])
{
    int i,aok = 1;
    char *s;

    for (i = 0; aok && i < argc; i++) {
        s = argv[i];
        if (*s == '-') {
            switch (*++s) {
            case 'm':
                aok = (trymode[0] = atox(s+1)) != -1;
                break;
            case 'c':
                crs = 1;
                break;
            default:
                aok = 0;
            }
        } else if (*s == '?') aok = 0;
    }
    if (!aok) {
        printf("Usage: msrv -c -m<XX>\n");
        printf("  -m  graphics mode, <XX> in hex (e.g. -m10, -m13)\n");
        printf("  -c  use cross cursor\n");
        printf("  \"LINKNAME\" in environment specifies link\n");
        printf("  buttons: left = zoom, right = outer frame, both = exit\n");
        exit(1);
    }
    lname = getenv(LINK_NAME);
    if ((id = OpenLink(lname)) < 0) fexit("can't open link: ",lname);
    signal(SIGINT,terminate);
    mandel();
    terminate(SIGTERM);
}

int atox(char *str)                 /* convert hex string to number */
{
    int c,n;

    if (str[0] == '\0') return(-1);
    n = 0;
    while ((c = *str++) != '\0') {
        if (c >= '0' && c <= '9') c -= '0';
        else {
            c &= 0xDF;
            if (c >= 'A' && c <= 'F') c -= 'A' - 10;
            else return(-1);
        }
        n = (n << 4) + c;
    }
    return(n);
}

void terminate(int sig)                 /* clean up before exit */
{
    _setvideomode(_DEFAULTMODE);
    if (id >= 0) CloseLink(id);
    if (tmsg[0]) cputs(tmsg);
    while (kbhit()) getch();
    if (sig == SIGTERM) exit(0); else exit(1);
}

void fexit(char *msg1, char *msg2)      /* fatal exit with message */
{
    sprintf(tmsg," -- %s%s\r\n",msg1,msg2);
    terminate(SIGABRT);
}

typedef struct {
    long len,type;
} LBHDR;

typedef struct {
    long x,y,w,h;
} BLOCK, *BLOCKPTR;

typedef struct {
    real Cx,Cy,Sx,Sy;
} AREA, *AREAPTR;

typedef struct {
    long mod,off;
} PMOD, *PMODPTR;

void HostOut(LBHDR *hdr, void *buf)     /* send message to nodes */
{
    WriteLink(id,(char *)hdr,sizeof(LBHDR),0);
    if ((int)hdr->len > 0) WriteLink(id,(char *)buf,(int)hdr->len,0);
}

void HostIn(LBHDR *hdr, void *buf)      /* recieve message from nodes */
{
    ReadLink(id,(char *)hdr,sizeof(LBHDR),0);
    if ((int)hdr->len > 0) ReadLink(id,(char *)buf,(int)hdr->len,0);
}

void makepalette(                       /* set up a rainbow of colors */
    long *colors, int idx, int len,
    int r, int g, int b,
    int dr, int dg, int db,
    int lo, int hi)
{
    register char *ptr = (char *)&colors[idx];
    register char *end = (char *)&colors[idx+len];

    while (ptr < end) {
        *ptr++ = r; *ptr++ = g; *ptr++ = b; *ptr++ = 0;
        r += dr; g += dg; b += db;
        if (r > hi) {r = hi - (r-hi); dr = -dr;}
        else if (r < lo) {r = lo + (lo-r); dr = -dr;}
        if (g > hi) {g = hi - (g-hi); dg = -dg;}
        else if (g < lo) {g = lo + (lo-g); dg = -dg;}
        if (b > hi) {b = hi - (b-hi); db = -db;}
        else if (b < lo) {b = lo + (lo-b); db = -db;}
    }
}

void getpixels()                    /* get pixel blocks and display */
{
    register int x,y;
    int xe,ye;
    unsigned char *ptr;
    LBHDR hdr;
    struct {
        BLOCK pb;
        char pix[MAXPIX];
    } m;

    loop {
        ptr = m.pix;
        HostIn(&hdr,&m);
        if ((int)hdr.type != RENDER_P) break;
        xe = m.pb.x + m.pb.w;
        ye = m.pb.y + m.pb.h;
        for (y = m.pb.y; y < ye; y++) {
            for (x = m.pb.x; x < xe; x++) {
                _setcolor((short)*ptr++);
                _setpixel((short)x,(short)y);
            }
        }
    }
}

int ResetM(int *nbut);
void ShowM();
void HideM();
void GetPosM(int *x, int *y, int *b);
void SetPointerM(int hotx, int hoty, short far *image);

long *colors;           /* pointer to allocated color palette */
struct videoconfig vc;  /* video config parameters */

short cross[] = {                       /* mouse pointer image */
    0xFEFF, 0xFEFF, 0xFEFF, 0xFEFF,
    0xFEFF, 0xFEFF, 0xFEFF, 0x0001,
    0xFEFF, 0xFEFF, 0xFEFF, 0xFEFF,
    0xFEFF, 0xFEFF, 0xFEFF, 0xFFFF,

    0x0100, 0x0100, 0x0100, 0x0100,
    0x0100, 0x0100, 0x0100, 0xFFFE,
    0x0100, 0x0100, 0x0100, 0x0100,
    0x0100, 0x0100, 0x0100, 0x0000
};

void mandel()
{
    int x,y,b;  /* mouse x, y and buttons */
    int i,t,s;  /* temperaries */
    long imax;  /* maximum iterations */
    long pb[2]; /* pixel block width and height */
    LBHDR hdr;  /* header for load-balance messages */
    BLOCK vp;   /* viewport */
    PMOD pm;    /* pixel modifier */
    AREA ar;    /* area */

    for (i = 0; ; i++) {                            /* set video mode */
        if (i == sizeof(trymode)/sizeof(short))
            fexit("can't set a graphics mode","");
        if (trymode[i] && _setvideomode(trymode[i])) break;
    }
    _getvideoconfig((struct videoconfig far *)&vc);

    if (!ResetM(&b))                                /* init mouse */
        fexit("no mouse installed","");
    if (b < 2)
        fexit("need 2 or more mouse buttons","");
    if (crs) {                                      /* set cursor */
        t = (vc.numxpixels == 640) ? 7 : 14;
        if (vc.mode < 0x06) t = 6;
        SetPointerM(t,7,(short far *)cross);
    }
    if (vc.numcolors > 16) {
        pm.mod = vc.numcolors-16; pm.off = 16;
    } else {
        pm.mod = vc.numcolors- 1; pm.off = 0;
    }
    if (vc.numcolors > 4) {                         /* set up palette */
        if ((colors = (long *)malloc(sizeof(long)*vc.numcolors)) == NULL)
            fexit("can't allocate palette memory","");
        for (i = 0; i <= (int)pm.off; i++) colors[i] = 0x000000L;   /* black */
        colors[vc.numcolors > 16 ? 15 : vc.numcolors-1] = 0x3F3F3FL;/* white */
        makepalette(
            colors,(int)pm.off+1,(int)pm.mod-1,
            4,2,57,
            +5,+9,-6,
            1,63);
        _remapallpalette((long far *)colors);
    } else pm.mod++;
    hdr.type = PMOD_P; hdr.len = sizeof(pm);        /* send pixel modifier */
    HostOut(&hdr,&pm);
    hdr.type = VIEWPORT_P; hdr.len = sizeof(vp);    /* send viewport */
    vp.x = 0; vp.y = 0; vp.w = vc.numxpixels; vp.h = vc.numypixels;
    HostOut(&hdr,&vp);
    hdr.type = MAXITER_P; hdr.len = sizeof(imax);   /* send max iterations */
    imax = 256;
    HostOut(&hdr,&imax);
    ar.Cx = -0.75; ar.Cy = 0.00; ar.Sx = 1.25/aspect; ar.Sy = 1.25;
    loop {
        hdr.type = AREA_P; hdr.len = sizeof(ar);    /* send area to compute */
        HostOut(&hdr,&ar);
        hdr.type = RENDER_P; hdr.len = sizeof(pb);  /* send render command */
        pb[0] = 16; pb[1] = 16;
        HostOut(&hdr,pb);
        getpixels();                                /* get pixels back */
        ShowM();
        do GetPosM(&x,&y,&b); while (b == 0);       /* get button press */
        do {GetPosM(&t,&t,&s); b |= s;} while (s != 0);
        HideM();
        if (b == 1) {                   /* zoom 10X on selected point */
            if ((int)vp.w == 320) x /= 2;
            ar.Cx += (x - (int)vp.x)
                * (ar.Sx*2.0/((int)vp.w-1)) - ar.Sx;
            ar.Cy += (((int)vp.h-1) - (y - (int)vp.y))
                * (ar.Sy*2.0/((int)vp.h-1)) - ar.Sy;
            ar.Sx *= 0.10;
            ar.Sy = ar.Sx*aspect;
        } else if (b == 2) {            /* go back to start frame */
            ar.Cx = -0.75; ar.Cy = 0.00; ar.Sx = 1.25/aspect; ar.Sy = 1.25;
        } else break;                   /* exit */
    }
}

#include <dos.h>

int ResetM(int *nbut)               /* reset mouse and get status */
{
    union REGS regs;

    regs.x.ax = 0x0000;
    int86(0x33,&regs,&regs);
    *nbut = regs.x.bx;
    return(regs.x.ax);
}

void ShowM()                        /* show mouse pointer */
{
    union REGS regs;

    regs.x.ax = 0x0001;
    int86(0x33,&regs,&regs);
}

void HideM()                        /* hide mouse pointer */
{
    union REGS regs;

    regs.x.ax = 0x0002;
    int86(0x33,&regs,&regs);
}

void GetPosM(int *x, int *y, int *b)    /* get mouse position and buttons */
{
    union REGS regs;

    regs.x.ax = 0x0003;
    int86(0x33,&regs,&regs);
    *b = regs.x.bx;
    *x = regs.x.cx;
    *y = regs.x.dx;
}

void SetPointerM(int hotx, int hoty, short far *image)  /* set pointer image */
{
    union REGS regs;
    struct SREGS sregs;

    regs.x.ax = 0x0009;
    regs.x.bx = hotx;
    regs.x.cx = hoty;
    regs.x.dx = FP_OFF(image);
    sregs.es = FP_SEG(image);
    int86x(0x33,&regs,&regs,&sregs);
}
