/************************************************************************/
/*									*/
/*	File:	sorts.c							*/
/*									*/
/*	Another fine member of the LOGICAL SYSTEMS "C" library series.	*/
/*									*/
/*	Contains the following sorting routines and associated support	*/
/*	code:								*/
/*									*/
/*		"isort"	-	Linear Insertion sort.			*/
/*		"qsort"	-	Bounded stack Quick sort.		*/
/*		"ssort"	-	Binary Shell sort.			*/
/*									*/
/*	Copyright (C) 1986-1992 by Kirk Bailey of Logical Systems.	*/
/*									*/
/*	Written:	7/9/86						*/
/*	Last update:	2/12/92						*/
/*									*/
/************************************************************************/

#include	<stdlib.h>

/************************************************************************/
/*									*/
/*	Routine:	_swap						*/
/*									*/
/*	Exchange two blocks of memory.					*/
/*									*/
/*	Parameters:	"a"		Points to first block.		*/
/*			"b"		Points to second block.		*/
/*			"width"		Width of an element in bytes.	*/
/*									*/
/*	Returns:	None.						*/
/*  									*/
/************************************************************************/

static	void
_swap(register char *a,register char *b,size_t width)
	{
	register	int	temp;

	while(width--)
		{
		temp = *a;
		*a++ = *b;
		*b++ = temp;
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	isort						*/
/*									*/
/*	Linear Insertion SORT.						*/
/*									*/
/*	Parameters:	"base"		Points to start of structure.	*/
/*			"num"		Number of elements.		*/
/*			"width"		Width of an element in bytes.	*/
/*			"compare"	Points to the compare routine.	*/
/*									*/
/*	Returns:	None.						*/
/*  									*/
/************************************************************************/

void
isort(void *_base,size_t num,size_t width,int (*compare)(void *,void *))
	{
	char			*base;
	char			*i;
	register	char	*j;
	register	char	*k;
	register	int	temp;
	char			*top;
	size_t			_width;

	base = _base;
	top = base + ((num - 1) * width);
	for(i = top - width; i >= base; i -= width)
		{
		j = (k = i) + width;
		while((j <= top) && ((*compare)(k,j) > 0))
			{
			_width = width;
			while(_width--)
				{
				temp = *j;
				*j++ = *k;
				*k++ = temp;
				}
			}
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	qsort						*/
/*									*/
/*	Quick SORT routine.						*/
/*									*/
/*	Adapted from G.H. Gonnet's HANDBOOK OF ALGORITHMS AND DATA	*/
/*	STRUCTURES pp. 262-263 and from Horowitz and Sahni's		*/
/*	FUNDAMENTALS OF COMPUTER ALGORITHMS, pp. 121-127.  Some subtle	*/
/*	details on partitioning technique were derived from Sedgewick's	*/
/*	ALGORITHMS pp. 103-113.  This version utilizes the bounded	*/
/*	stack growth variation to prevent stack "accidents" with	*/
/*	pre-sorted files.  A final pass of Insertion sort is used to	*/
/*	allow the Quicksort to terminate when the partitions become	*/
/*	small.  Partitioning is done using the median-of-three		*/
/*	technique to improve performance on pre-sorted files.		*/
/*									*/
/*	Parameters:	"base"		Points to start of structure.	*/
/*			"num"		Number of elements.		*/
/*			"width"		Width of an element in bytes.	*/
/*			"compare"	Points to the compare routine.	*/
/*									*/
/*	Returns:	None.						*/
/*  									*/
/************************************************************************/

static	void
_qsort(char *base,size_t num,register size_t width,int (*compare)(void *,
	void *))
	{
	register	char	*i;
	register	char	*j;
	char			*top;
	size_t			u;

	top = base + ((num - 1) * width);

	while(top > base)
		{
/*
 *	If there are too few data items give up and let the final insertion
 *	sort pass fix things up.
 */
		if((top - base) <= (width << 3))
			return;
/*
 *	Otherwise, partition into two subfiles using the median-of-three
 *	technique for choosing the partition element.
 */
		i = (((((top - base) / width) >> 1) + 1) * width) + base;
		if((*compare)(base,i) > 0)
			_swap(base,i,width);
		if((*compare)(base,top) > 0)
			_swap(base,top,width);
		if((*compare)(i,top) > 0)
			_swap(i,top,width);
		base += width;
		_swap(base,i,width);
		i = base;
		j = top - width;
		while(i < j)
			{
			while((*compare)(j,base) > 0)
				j -= width;
			while((i < j) && (*compare)(i,base) <= 0)
				i += width;
			_swap(i,j,width);
			}
		_swap(base,i,width);
		base -= width;
/*
 *	Sort smallest subfile first.  Sort larger partition by looping back up
 *	up and executing the above procedure again (love that tail recursion).
 */
		if(i - base < top - i)
			{
			if((u = ((i - base) / width) + 1) > 9)
				_qsort(base,u,width,compare);
			base = i + width;
			}
		else
			{
			if((u = ((top - i) / width)) > 9)
				_qsort(i + width,u,width,compare);
			top = i - width;
			}
		}
	}

void
qsort(void *base,size_t num,size_t width,int (*compare)())
	{
	if(num > 9)
		_qsort(base,num,width,compare);

	isort(base,num,width,compare);
	}

/************************************************************************/
/*									*/
/*	Routine:	ssort						*/
/*									*/
/*	Shell SORT routine.						*/
/*									*/
/*	Parameters:	"base"		Points to start of structure.	*/
/*			"num"		Number of elements.		*/
/*			"width"		Width of an element in bytes.	*/
/*			"compare"	Points to the compare routine.	*/
/*									*/
/*	Returns:	None.						*/
/*  									*/
/************************************************************************/

void
ssort(void *_base,size_t num,size_t width,int (*compare)(void *,void *))
	{
	char			*base;
	char			*end;
	register	size_t	gap;
	char			*i;
	register	char	*j;
	register	char	*jg;

	base = _base;
	end = base + (num * width);
	while(gap = (num >>= 1) * width)
		{
		for(i = base + gap; i < end; i += width)
			{
			for(j = i - gap; j >= base; j -= gap)
				{
				jg = j + gap;
				if((*compare)(j,jg) <= 0)
					break;
				_swap(j,jg,width);
				}
			}
		}
	}
