/*
 *	tcio.c
 *
 *	Client (Transputer) side of "C" remote file I/O server.  Stream I/O
 *	routines are located in "tstream.c" which uses the basic server code in
 *	this file also.
 *
 *	Copyright 1987-1992 by Logical Systems.
 */

#include	<time.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<conc.h>
#include	<cio.h>
#include	<cioext.h>
#include	<inline.h>

#define	UCHAR	unsigned char

/*
 *	Include file consistency check for seek origin values.
 */

#if	(SEEK_CUR!=_SEEK_CUR) || (SEEK_END!=_SEEK_END) || (SEEK_SET!=_SEEK_SET)
#error	"FATAL: Seek origin inconsistency between 'cio.h' and 'stdio.h'"
#endif	/* Seek origin check */

/*
 *	Client I/O channel protection semaphore.
 */
Semaphore	_cio_prot =	SEMAPHOREINIT;

/*
 *	In order to make FILE * pointers actually be a pointer, rather than
 *	carry around an integer in a pointer, which must get converted (with
 *	random results), we store the integer from the server in an array, and
 *	make the (FILE *) pointers point to the array item.
 *
 *	Note that nothing is really stored in this array.  We are simply using
 *	the address of the array items.  As C allows you to reference past the
 *	end of the array, there is no reason to make the array overly large.
 */
FILE		_fio[1];

/*
 *	Static array for "gmtime"/"localtime" use.
 */
static	struct	tm	ttemp;

/*
 *	Function definitions.
 */

void	_brkpt(int,int,int);	/* Report that we hit a breakpoint	*/
int	_dec_2b(UCHAR *);	/* Convert 2 little endian bytes to int	*/
long	_dec_4b(UCHAR *);	/* Convert 4 little endian bytes to long*/
void	_enc_strm(UCHAR *,FILE *); /* Convert strm index into 2 bytes	*/
void	_enc_2b(UCHAR *,int);	/* Convert int to little endian 2 bytes	*/
void	_enc_4b(UCHAR *,long);	/* Convert long to little endian 4 bytes*/
void	_get_str(UCHAR *,int);	/* Get specified # of bytes from host	*/
FILE	*_get_strm(void);	/* Get stream pointer from host		*/
void	_get_time(void);	/* Get/cnvt broken down time from host	*/
int	_get_zeof(int);		/* Get one byte worth of 0/EOF from host*/
int	_get_1b(void);		/* Get one byte from host		*/
int	_get_2b(void);		/* Get 2 bytes from host		*/
int	_get_2beof(void);	/* Get 2 bytes/EOF from host		*/
long	_get_4b(void);		/* Get 4 bytes from host		*/
void	_put_str(UCHAR *,int);	/* Send "raw" string to host		*/
void	_put_1b(int);		/* Send "raw" byte to host		*/
void	_request_errno(void);	/* Get host "errno" value		*/
void	_send_cmd(UCHAR *,int,int); /* Send command to host		*/
int	_unlock_int(int);	/* Unlock client/server channel	(int)	*/
char	*_unlock_str(char *);	/* Unlock client/server channel	(string)*/
FILE	*_unlock_strm(FILE *);	/* Unlock client/server channel	(stream)*/
void	_unlock_void(void);	/* Unlock client/server channel	(void)	*/

/*
 *	Get "argv" table and argument lists and place at specified memory
 *	address.  See the "start.c" function and the "cio.c" I/O server
 *	source for more information.  Returns the value of "argc".
 */
int
_arg_get(UCHAR *addr)
	{
	UCHAR			buf[4];
	int			length;
	int			result;

	_enc_4b(buf,(long) addr);
	_send_cmd(buf,FN_ARG_GET,4);
	_get_str(buf,4);		/* Get return value and data size */
	result = _dec_2b(buf);		/* "argc" return value */
	length = _dec_2b(&buf[2]);	/* Total argument data length */
	_get_str((UCHAR *) addr,length);	/* Get argument data */
	return (_unlock_int(result));
	}

/*
 *	Ask the "cio.c" I/O server for the total storage that the "argv"
 *	pointer table and associated values will take (in bytes, rounded to the
 *	nearest word).  See the "start.c" function and the "cio.c" server
 *	source for more information.  Also, tell the host the breakpoint
 *	function address, and how many bytes are in each word on this node.
 */
int
_arg_size(void)
	{
	UCHAR			buf[6];

	_enc_4b(buf,(long) _brkpt);	/* Tell host breakpoint address */
	_enc_2b(&buf[4],(int) sizeof(char **));	/* Tell host bytes/word */
	_send_cmd(buf,FN_ARG_SIZE,6);
	return (_unlock_int(_get_2b()));
	}

/*
 *	Tell CIO that we have encountered a breakpoint and provide the
 *	corresponding status information, then die.
 */
void
_brkpt(int a,int b,int c)
	{
#ifndef	ISTC
	#error	"Breakpoint code is EXTREMELY processor specific"
#else	/* ISTC */
	UCHAR			buf[_DEBUG_REGS * 4];

	_enc_4b(&buf[_DEBUG_WS * 4],((long) (&c + 1)) |
		((long) ProcGetPriority()));		/* WS */
	_enc_4b(&buf[_DEBUG_PC * 4],(long) *(&a - 1));	/* PC */
	_enc_4b(&buf[_DEBUG_A * 4],(long) a);		/* Areg */
	_enc_4b(&buf[_DEBUG_B * 4],(long) b);		/* Breg */
	_enc_4b(&buf[_DEBUG_C * 4],(long) c);		/* Creg */
/*
 *	Preserve the floating point status (if appropriate).
 */
#ifdef	T800				/* T800 CPU:  There be Magik here! */
	#pragma	asm
	_asm_eval1(&buf[_DEBUG_FA_LOW * 4]);
	FPSTTEST			/* Store FAreg & test bits */
	_asm_eval1(&buf[_DEBUG_FB_LOW * 4]);
	FPSTTEST			/* Store FBreg & test bits */
	_asm_eval1(&buf[_DEBUG_FC_LOW * 4]);
	FPSTTEST			/* Store FCreg & test bits */
	#pragma	endasm
#else	/* !T800 */
#ifndef	T212				/* Must be T414 CPU */
	#pragma	asm
	.EXT	_FPstate
	.LDC	_FPstate+0		/* Address of simulated FAreg */
	_asm_eval1(&buf[_DEBUG_FA_LOW * 4]);
	.LDC	8
	MOVE
	.LDC	_FPstate+8		/* Address of simulated FBreg */
	_asm_eval1(&buf[_DEBUG_FB_LOW * 4]);
	.LDC	8
	MOVE
	.LDC	_FPstate+16		/* Address of simulated FCreg */
	_asm_eval1(&buf[_DEBUG_FC_LOW * 4]);
	.LDC	12			/* Also copy the round mode bits */
	MOVE
	#pragma	endasm
#endif	/* !T212 */
#endif	/* T800 */
/*
 *	Send command to host.
 */
	_send_cmd(buf,FN_BRKPT,_DEBUG_REGS * 4);	/* Send brkpt msg */
	_unlock_void();			/* Unlock communication channel */
	ProcStop();			/* Kill this process */
#endif	/* !ISTC */
	}

int
_cioext(int code,void *sbuf,int sbuflen,void *rbuf,int rbuflen)
	{
	UCHAR			buf[8];
	int			rval;
	int			temp;

	if((sbuflen > MAX_MSG_DATA) || (rbuflen > MAX_MSG_DATA))
		return (-1);

	_enc_2b(buf,code);
	_enc_2b(&buf[2],sbuflen);
	_enc_2b(&buf[4],rbuflen);
	_send_cmd(buf,FN_CIOEXT,6);
	_put_str((UCHAR *) sbuf,sbuflen);

	_get_str(buf,8);
	rval = _dec_4b(buf);		/* Get desired return value */
	if((temp = _dec_4b(&buf[4])) != 0)
		errno = temp;		/* Host "errno" value */
	_get_str((UCHAR *) rbuf,rbuflen);	/* Read reply str */
	return (_unlock_int(rval));
	}

void
clearerr(FILE *stream)
	{
	UCHAR			buf[2];

	_enc_strm(buf,stream);
	_send_cmd(buf,FN_CLEARERR,2);
	_request_errno();
	_unlock_void();
	}

int
close(int handle)
	{
	UCHAR			buf[2];

	_enc_2b(buf,handle);
	_send_cmd(buf,FN_CLOSE,2);
	return (_unlock_int(_get_zeof(1)));
	}

int
creat(const char *pathname,int pmode)
	{
	UCHAR			buf[4];
	int			length;

	if((pathname == NULL) ||
		((length = (strlen(pathname) + 1)) > MAX_MSG_DATA))
		{
		return (EOF);
		}

	_enc_2b(buf,pmode);
	_enc_2b(&buf[2],length);
	_send_cmd(buf,FN_CREAT,4);
	_put_str((UCHAR *) pathname,length);
	return (_unlock_int(_get_2beof()));
	}

int
dup(int handle)
	{
	UCHAR			buf[2];

	_enc_2b(buf,handle);
	_send_cmd(buf,FN_DUP,2);
	return (_unlock_int(_get_2beof()));
	}

int
dup2(int handle1,int handle2)
	{
	UCHAR			buf[4];

	_enc_2b(buf,handle1);
	_enc_2b(&buf[2],handle2);
	_send_cmd(buf,FN_DUP2,4);
	return (_unlock_int(_get_zeof(1)));
	}

void
exit(int status)
	{
	UCHAR			buf[4];

	_enc_4b(buf,(long) status);
	_send_cmd(buf,FN_EXIT,4);
	_unlock_void();			/* Unlock client/server channel */
	ProcStop();			/* Kill Process off */
	}

int
fclose(FILE *stream)
	{
	UCHAR			buf[2];

	_enc_strm(buf,stream);
	_send_cmd(buf,FN_FCLOSE,2);
	return (_unlock_int(_get_zeof(1)));
	}

int
fcloseall(void)
	{
	_send_cmd(NULL,FN_FCLOSEALL,0);
	return (_unlock_int(_get_2beof()));
	}

FILE	*
fdopen(int handle,const char *mode)
	{
	UCHAR			buf[4];
	int			length;

	if((mode == NULL) || ((length = (strlen(mode) + 1)) > MAX_MSG_DATA))
		return ((FILE *) NULL);

	_enc_2b(buf,handle);
	_enc_2b(&buf[2],length);
	_send_cmd(buf,FN_FDOPEN,4);
	_put_str((UCHAR *) mode,length);
	return (_unlock_strm(_get_strm()));
	}

int
feof(FILE *stream)
	{
	UCHAR			buf[2];

	_enc_strm(buf,stream);
	_send_cmd(buf,FN_FEOF,2);
	return (_unlock_int(_get_zeof(0)));
	}

int
ferror(FILE *stream)
	{
	UCHAR			buf[2];

	_enc_strm(buf,stream);
	_send_cmd(buf,FN_FERROR,2);
	return (_unlock_int(_get_zeof(0)));
	}

int
fflush(FILE *stream)
	{
	UCHAR			buf[2];

	_enc_strm(buf,stream);
	_send_cmd(buf,FN_FFLUSH,2);
	return (_unlock_int(_get_zeof(1)));
	}

int
fileno(FILE *stream)
	{
	UCHAR			buf[2];

	_enc_strm(buf,stream);
	_send_cmd(buf,FN_FILENO,2);
	return (_unlock_int(_get_2beof()));
	}

FILE	*
fopen(const char *pathname,const char *mode)
	{
	UCHAR			buf[4];
	int			length1;
	int			length2;

	if((pathname == NULL) || (mode == NULL) ||
		((length1 = (strlen(pathname) + 1)) > MAX_MSG_DATA) ||
		((length2 = (strlen(mode) + 1)) > MAX_MSG_DATA))
		{
		return ((FILE *) NULL);
		}

	_enc_2b(buf,length1);
	_enc_2b(&buf[2],length2);
	_send_cmd(buf,FN_FOPEN,4);
	_put_str((UCHAR *) pathname,length1);
	_put_str((UCHAR *) mode,length2);
	return (_unlock_strm(_get_strm()));
	}

size_t
fread(void *ptr,size_t size,size_t nmemb,FILE *stream)
	{
	int			actual;
	UCHAR			buf[6];
	int			desired;
	int			per_read;
	size_t			total_membs;

	if((size == 0) || (size > MAX_MSG_DATA))
		return ((size_t) 0);

	total_membs = 0;
	per_read = MAX_MSG_DATA / size;

	_enc_strm(buf,stream);
	_enc_2b(&buf[2],(int) size);
/*
 *	Use multiple short "fread" commands to fake one long one.
 */
	while(nmemb > 0)
		{
		if(nmemb > per_read)
			desired = per_read;
		else
			desired = nmemb;
		nmemb -= desired;
		_enc_2b(&buf[4],desired);
		_send_cmd(buf,FN_FREAD,6);
		if((actual = _get_2b()) == 0)
			{
			if(total_membs == 0)
				_request_errno();
			_unlock_void();	/* Unlock client/server channel */
			break;
			}
		else
			{
			_get_str((UCHAR *) ptr,actual);
			_unlock_void();	/* Unlock client/server channel */
			ptr = ((UCHAR *) ptr) + actual;
			total_membs += (actual /= size);
			if(actual < desired)
				break;
			}
		}
	return (total_membs);
	}

FILE	*
freopen(const char *pathname,const char *mode,FILE *stream)
	{
	UCHAR			buf[6];
	int			length1;
	int			length2;

	if((pathname == NULL) || (mode == NULL) ||
		((length1 = (strlen(pathname) + 1)) > MAX_MSG_DATA) ||
		((length2 = (strlen(mode) + 1)) > MAX_MSG_DATA))
		{
		return ((FILE *) NULL);
		}

	_enc_strm(buf,stream);
	_enc_2b(&buf[2],length1);
	_enc_2b(&buf[4],length2);
	_send_cmd(buf,FN_FREOPEN,6);
	_put_str((UCHAR *) pathname,length1);
	_put_str((UCHAR *) mode,length2);
	return (_unlock_strm(_get_strm()));
	}

int
fseek(FILE *stream,long offset,int origin)
	{
	UCHAR			buf[7];

	_enc_strm(buf,stream);
	_enc_4b(&buf[2],offset);
	buf[6] = origin;
	_send_cmd(buf,FN_FSEEK,7);
	return (_unlock_int(_get_zeof(1)));
	}

long
ftell(FILE *stream)
	{
	UCHAR			buf[2];
	long			result;

	_enc_strm(buf,stream);
	_send_cmd(buf,FN_FTELL,2);
	result = _get_4b();
	if(result == RPC_EOF)
		{
		result = EOF;
		_request_errno();
		}
	_unlock_void();			/* Unlock client/server channel */
	return (result);
	}

size_t
fwrite(const void *ptr,size_t size,size_t nmemb,FILE *stream)
	{
	int			actual;
	UCHAR			buf[6];
	int			desired;
	int			per_length;
	int			per_write;
	size_t			total_membs;

	if((size == 0) || (size > MAX_MSG_DATA))
		return ((size_t) 0);

	total_membs = 0;
	per_write = MAX_MSG_DATA / size;

	_enc_strm(buf,stream);
	_enc_2b(&buf[2],(int) size);
/*
 *	Use multiple short "fwrite" commands to fake one long one.
 */
	while(nmemb > 0)
		{
		if(nmemb > per_write)
			desired = per_write;
		else
			desired = nmemb;
		nmemb -= desired;
		_enc_2b(&buf[4],desired);
		_send_cmd(buf,FN_FWRITE,6);
		_put_str((UCHAR *) ptr,(per_length = desired * size));
		if(((actual = _get_2b()) == 0) && (total_membs == 0))
			_request_errno();
		_unlock_void();	/* Unlock client/server channel */
		total_membs += actual;	/* # of items, not length */
		if(actual < desired)
			break;
		ptr = ((UCHAR *) ptr) + per_length;
		}
	return (total_membs);
	}

char	*
getenv(const char *name)
	{
	UCHAR			buf[2];
	int			length;
	char			*sptr;

	if((name == NULL) || ((length = (strlen(name) + 1)) > MAX_MSG_DATA))
		return ((char *) NULL);

	_enc_2b(buf,length);
	_send_cmd(buf,FN_GETENV,2);
	_put_str((UCHAR *) name,length);
	if((length = _get_2b()) == 0)	/* Figure out if string/NULL */
		return (_unlock_str((char *) NULL));
	else				/* Real environment string, read */
		{
		if((sptr = (char *) malloc((size_t) length)) == NULL)
			{
			_put_1b(RPC_NULL);	/* Signal failure */
			return (_unlock_str((char *) NULL));
			}
		else
			{
			_put_1b(RPC_NOTNULL);	/* Signal success */
			_get_str((UCHAR *) sptr,length);
			return (_unlock_str(sptr));
			}
		}
	}

#ifndef	T212
struct	tm	*
gmtime(const time_t *timer)
	{
	UCHAR			buf[4];

	_enc_4b(buf,(long) *timer);
	_send_cmd(buf,FN_TIMEGMT,4);
	if(_get_1b() != RPC_NULL)	/* Figure out if time/NULL */
		{
		_get_time();
		_unlock_void();		/* Unlock client/server channel */
		return (&ttemp);
		}
	else
		{
		_unlock_void();		/* Unlock client/server channel */
		return ((struct tm *) NULL);
		}
	}
#endif	/* !T212 */

#ifndef	T212
struct	tm	*
localtime(const time_t *timer)
	{
	UCHAR			buf[4];

	_enc_4b(buf,(long) *timer);
	_send_cmd(buf,FN_TIMELOC,4);
	_get_time();
	_unlock_void();			/* Unlock client/server channel */
	return (&ttemp);
	}
#endif	/* !T212 */

long
lseek(int handle,long offset,int origin)
	{
	UCHAR			buf[7];
	long			result;

	_enc_2b(buf,handle);
	_enc_4b(&buf[2],offset);
	buf[6] = origin;
	_send_cmd(buf,FN_LSEEK,7);
	result = _get_4b();
	if(result == RPC_EOF)
		{
		result = EOF;
		_request_errno();
		}
	_unlock_void();			/* Unlock client/server channel */
	return (result);
	}

int
open(const char *pathname,int oflag,int pmode)
	{
	UCHAR			buf[6];
	int			length;

	if((pathname == NULL) ||
		((length = (strlen(pathname) + 1)) > MAX_MSG_DATA))
		{
		return (EOF);
		}

	_enc_2b(buf,oflag);
	_enc_2b(&buf[2],pmode);
	_enc_2b(&buf[4],length);
	_send_cmd(buf,FN_OPEN,6);
	_put_str((UCHAR *) pathname,length);
	return (_unlock_int(_get_2beof()));
	}

int
read(int handle,char *s,size_t count)
	{
	int			actual;
	UCHAR			buf[4];
	int			desired;
	size_t			total_length;

	total_length = 0;

	_enc_2b(buf,handle);
/*
 *	Use multiple short "read" commands to fake one long one.
 */
	while(count > 0)
		{
		if(count > MAX_MSG_DATA)
			desired = MAX_MSG_DATA;
		else
			desired = count;

		_enc_2b(&buf[2],desired);
		_send_cmd(buf,FN_READ,4);

		if((actual = _get_2beof()) == EOF)
			return (_unlock_int(EOF));
		if(actual > 0)
			{
			_get_str((UCHAR *) s,actual);	/* Read data */
			_unlock_void();	/* Unlock client/server chan */
			s += actual;
			total_length += actual;
			count -= actual;
			}
		else
			{
			_unlock_void();	/* Unlock client/server chan */
			break;
			}
		}
	return ((int) total_length);
	}

int
remove(const char *pathname)
	{
	UCHAR			buf[2];
	int			length;

	if((pathname == NULL) ||
		((length = (strlen(pathname) + 1)) > MAX_MSG_DATA))
		{
		return (EOF);
		}

	_enc_2b(buf,length);
	_send_cmd(buf,FN_REMOVE,2);
	_put_str((UCHAR *) pathname,length);
	return (_unlock_int(_get_zeof(1)));
	}

int
rename(const char *old,const char *new)
	{
	UCHAR			buf[4];
	int			length1;
	int			length2;

	if((old == NULL) || (new == NULL) ||
		((length1 = (strlen(old) + 1)) > MAX_MSG_DATA) ||
		((length2 = (strlen(new) + 1)) > MAX_MSG_DATA))
		{
		return (EOF);
		}

	_enc_2b(buf,length1);
	_enc_2b(&buf[2],length2);
	_send_cmd(buf,FN_RENAME,4);
	_put_str((UCHAR *) old,length1);
	_put_str((UCHAR *) new,length2);
	return (_unlock_int(_get_zeof(1)));
	}

void
rewind(FILE *stream)
	{
	(void) fseek(stream,0L,SEEK_SET);	/* Seek to start */
	clearerr(stream);		/* Clear possible error flag */
	}

int
system(const char *string)
	{
	UCHAR			buf[2];
	int			length;
	char			*sptr;

	if(string == NULL)
		length = 0;
	else
		{
		if((length = (strlen(string) + 1)) > MAX_MSG_DATA)
			return (EOF);
		}

	_enc_2b(buf,length);
	_send_cmd(buf,FN_SYSTEM,2);
	_put_str((UCHAR *) string,length);
	return (_unlock_int((int) _get_4b()));
	}

#ifndef	T212
time_t
time(time_t *timer)
	{
	time_t			rval;

	_send_cmd(NULL,FN_TIME,0);
	rval = (time_t) _get_4b();
	_unlock_void();			/* Unlock client/server channel */
	if(timer != NULL)
		*timer = rval;
	return (rval);
	}
#endif	/* !T212 */

FILE	*
tmpfile(void)
	{
	_send_cmd(NULL,FN_TMPFILE,0);
	return (_unlock_strm(_get_strm()));
	}

char	*
tmpnam(char *string)
	{
	int			length;
	static char		*name_ptr = NULL;

	_send_cmd(NULL,FN_TMPNAM,0);
	if((length = _get_2b()) == 0)	/* Figure out if string/NULL */
		return (_unlock_str(NULL));
	else				/* A name has been generated, get it */
		{
		if(string == NULL)	/* Store it internally? */
			{
			if(name_ptr != NULL)
				free((void *) name_ptr);
			if((name_ptr = (char *) malloc((size_t) length)) ==
				NULL)
				{
				_put_1b(RPC_NULL);
				return (_unlock_str((char *) NULL));
				}
			else
				{
				_put_1b(RPC_NOTNULL);
				_get_str((UCHAR *) name_ptr,length);
				return (_unlock_str(name_ptr));
				}
			}
		else
			{
			_put_1b(RPC_NOTNULL);
			_get_str((UCHAR *) string,length);
			return (_unlock_str(string));
			}
		}
	}

int
ungetc(int c,FILE *stream)
	{
	UCHAR			buf[3];

	if(c == EOF)
		return (EOF);

	_enc_strm(buf,stream);
	buf[2] = c;
	_send_cmd(buf,FN_UNGETC,3);
	return (_unlock_int(_get_2beof()));
	}

int
unlink(const char *pathname)
	{
	return (remove(pathname));
	}

int
write(int handle,char *s,size_t count)
	{
	int			actual;
	UCHAR			buf[4];
	int			desired;
	size_t			total_length;

	total_length = 0;

	_enc_2b(buf,handle);
/*
 *	Use multiple short "write" commands to fake one long one.
 */
	while(count > 0)
		{
		if(count > MAX_MSG_DATA)
			desired = MAX_MSG_DATA;
		else
			desired = count;
		count -= desired;
		_enc_2b(&buf[2],desired);
		_send_cmd(buf,FN_WRITE,4);
		_put_str((UCHAR *) s,desired);
		if((actual = _get_2beof()) == EOF)
			return (_unlock_int(EOF));
		else
			{
			_unlock_void();	/* Unlock client/server channel */
			total_length += actual;
			if(actual < desired)
				break;
			s += desired;
			}
		}
	return ((int) total_length);
	}

/************************************************************************/
/*									*/
/*			COMMON UTILITY FUNCTIONS			*/
/*									*/
/************************************************************************/

/************************************************************************/
/*									*/
/*	Routine:	_dec_2b/_dec_4b					*/
/*									*/
/*	Translate a two/four byte little endian value into a int/long.	*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to unsigned char (the input buffer).	*/
/*									*/
/*	Returns:	The converted value.				*/
/*									*/
/************************************************************************/

int
_dec_2b(UCHAR *sptr)
	{
	int			rval;

	rval = *sptr++;
	return (rval | (*sptr << 8));
	}

long
_dec_4b(UCHAR *sptr)
	{
	long			rval;

	rval = *sptr++;
	rval |= (*sptr++ << 8);
#ifndef	T212
	rval |= (*sptr++ << 16);
	rval |= (*sptr << 24);
#endif	/* !T212 */
	return (rval);
	}

/************************************************************************/
/*									*/
/*	Routine:	_enc_strm					*/
/*									*/
/*	Translate a stream into a two byte table index and place into	*/
/*	the specified buffer.						*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to unsigned char (the buffer).	*/
/*		"strm"	The stream to convert.				*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

void
_enc_strm(UCHAR *sptr,FILE *strm)
	{
	_enc_2b(sptr,(int) (strm - _fio));
	}

/************************************************************************/
/*									*/
/*	Routine:	_enc_2b						*/
/*									*/
/*	Place a two byte signed value into the specified buffer.	*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to unsigned char (the buffer).	*/
/*		"value"	The value to place in the buffer.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

void
_enc_2b(UCHAR *sptr,int value)
	{
#ifndef	ISTC
	*sptr++ = value;
	*sptr = (value >>= 8);
#else	/* ISTC */
	#pragma	asm
	LDLP	[value]
	LB				;Low byte of "value"
	LDL	[sptr]
	SB				;Transfer byte over
	LDLP	[value]
	ADC	1
	LB				;High order byte of "value"
	LDL	[sptr]
	ADC	1
	SB				;Transfer byte over
	#pragma	endasm
#endif	/* !ISTC */
	}

/************************************************************************/
/*									*/
/*	Routine:	_enc_4b						*/
/*									*/
/*	Place a four byte signed value into the specified buffer.	*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to unsigned char (the buffer).	*/
/*		"value"	The value to place in the buffer.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

void
_enc_4b(UCHAR *sptr,long value)
	{
#ifndef	ISTC
	*sptr++ = value;
	*sptr++ = (value >>= 8);
	*sptr++ = (value >>= 8);
	*sptr = (value >> 8);
#else	/* ISTC */
	#pragma	asm
	LDLP	[value]
	LB				;Low byte of "value"
	LDL	[sptr]
	SB				;Transfer byte over
	LDLP	[value]
	ADC	1
	LB				;Next higher order byte of "value"
	LDL	[sptr]
	ADC	1
	SB				;Transfer byte over
#ifdef	T212
	LDL	[value]
	MINT
	AND				;Non zero if negative
	EQC	0			;0 if negative, 1 if positive
	ADC	-1			;FFFF if negative, 0000 if positive
	STL	[value]			;Save a copy
	LDL	[value]
	LDL	[sptr]
	ADC	2
	SB				;Sign extend into next high order byte
	LDL	[value]
	LDL	[sptr]
	ADC	3
	SB				;Sign extend into high order byte
#else	/* !T212 */
	LDLP	[value]
	ADC	2
	LB				;Next higher order byte of "value"
	LDL	[sptr]
	ADC	2
	SB				;Transfer byte over
	LDLP	[value]
	ADC	3
	LB				;High order byte of "value"
	LDL	[sptr]
	ADC	3
	SB				;Transfer byte over
#endif	/* T212 */
	#pragma	endasm
#endif	/* !ISTC */
	}

/************************************************************************/
/*									*/
/*	Routine:	_get_str					*/
/*									*/
/*	Read a string of specified length from the host system.		*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to unsigned char (the buffer to send).*/
/*		"len"	The buffer length to read.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

void
_get_str(UCHAR *sptr,int len)
	{
	if(len > 0)
		_HostIn(_host_chan_in,sptr,len);	/* Read data */
	}

/************************************************************************/
/*									*/
/*	Routine:	_get_strm					*/
/*									*/
/*	Read a two byte representation of a "stream" from the host	*/
/*	system and return it.  If the return value is "RPC_NULL", get	*/
/*	a current version of the host system "errno" and return NULL.	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The resulting stream pointer.			*/
/*									*/
/************************************************************************/

FILE	*
_get_strm(void)
	{
	UCHAR			buf[2];
	int			result;

	_HostIn(_host_chan_in,buf,2);	/* Read data */
	result = buf[0] | (buf[1] << 8);
	if(result == (RPC_NULL & 0xFFFF))
		{
		_request_errno();
		return ((FILE *) NULL);
		}
	else
		return (&_fio[(int) result]);
	}

/************************************************************************/
/*									*/
/*	Routine:	_get_time					*/
/*									*/
/*	Read a broken down time value from the host system.		*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

void
_get_time(void)
	{
	struct	rpc_time	t;

	_HostIn(_host_chan_in,&t,RPC_TIME_DATA);	/* Read data */

	ttemp.tm_sec = t.tm_sec;
	ttemp.tm_min = t.tm_min;
	ttemp.tm_hour = t.tm_hour;
	ttemp.tm_mday = t.tm_mday;
	ttemp.tm_mon = t.tm_mon;
	ttemp.tm_year = t.tm_year;
	ttemp.tm_wday = t.tm_wday;
	ttemp.tm_yday = (t.low_tm_yday + (t.high_tm_yday << 8));
	ttemp.tm_isdst = t.tm_isdst;
	}

/************************************************************************/
/*									*/
/*	Routine:	_get_zeof					*/
/*									*/
/*	Read a 1 byte value from the host.  If the value is non-zero,	*/
/*	return a value of EOF, otherwise just zero.  If the "error"	*/
/*	parameter is TRUE, request "errno" if non-zero has been read,	*/
/*	otherwise don't bother.						*/
/*									*/
/*	Parameters:							*/
/*		"error"	TRUE if "errno" should be requested on non-zero.*/
/*									*/
/*	Returns:	The byte value read/converted.			*/
/*									*/
/************************************************************************/

int
_get_zeof(int error)
	{
	signed	char		result;

	_HostIn(_host_chan_in,&result,1);
	if(result != 0)
		{
		if(error)
			_request_errno();
		return (EOF);
		}
	else
		return ((int) result);
	}

/************************************************************************/
/*									*/
/*	Routine:	_get_1b						*/
/*									*/
/*	Read a 1 byte value from the host.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The byte value read.				*/
/*									*/
/************************************************************************/

int
_get_1b(void)
	{
	signed	char		result;

	_HostIn(_host_chan_in,&result,1);
	return ((int) result);
	}

/************************************************************************/
/*									*/
/*	Routine:	_get_2b						*/
/*									*/
/*	Read a 2 byte value from the host.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The two byte value read (little endian).	*/
/*									*/
/************************************************************************/

int
_get_2b(void)
	{
	UCHAR			buf[2];

	_HostIn(_host_chan_in,buf,2);	/* Read data */
	return ((int) (buf[0] | (buf[1] << 8)));
	}

/************************************************************************/
/*									*/
/*	Routine:	_get_2beof					*/
/*									*/
/*	Read a 2 byte value from the host.  If the value is equal to	*/
/*	RPC_EOF, request "errno" and return EOF, otherwise return value.*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The two bytes read/converted (little endian).	*/
/*									*/
/************************************************************************/

int
_get_2beof(void)
	{
	UCHAR			buf[2];
	int			result;

	_HostIn(_host_chan_in,buf,2);	/* Read data */
	result = buf[0] | (buf[1] << 8);
	if(result == (RPC_EOF & 0xFFFF))
		{
		_request_errno();
		return (EOF);
		}
	else
		return (result);
	}

/************************************************************************/
/*									*/
/*	Routine:	_get_4b						*/
/*									*/
/*	Read a 4 byte signed value from the host.			*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The four byte value read (little endian).  If	*/
/*			the code is running on a T2 processor the top	*/
/*			two bytes are simply thrown away.		*/
/*									*/
/************************************************************************/

long
_get_4b(void)
	{
	long			temp[2];	/* temp[1] gets excess on T2 */

	_HostIn(_host_chan_in,temp,4);	/* Read four bytes */
	return (temp[0]);
	}

/************************************************************************/
/*									*/
/*	Routine:	_put_str					*/
/*									*/
/*	Send a string of specified length to the host system.		*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to unsigned char (the buffer to send).*/
/*		"len"	The buffer length to send.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

void
_put_str(UCHAR *sptr,int len)
	{
	if(len > 0)
		_HostOut(_host_chan_out,sptr,len);	/* Send data */
	}

/************************************************************************/
/*									*/
/*	Routine:	_put_1b						*/
/*									*/
/*	Send a single byte to the host system.				*/
/*									*/
/*	Parameters:							*/
/*		"c"	The character to send.				*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

void
_put_1b(int c)
	{
	char			ch;

	ch = c;
	_HostOut(_host_chan_out,&ch,1);	/* Send byte */
	}

/************************************************************************/
/*									*/
/*	Routine:	_request_errno					*/
/*									*/
/*	Request the host system value of "errno" to update the local	*/
/*	copy.								*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

void
_request_errno(void)
	{
	static	UCHAR		header[2] = {	FN_ERRNO,	0	};

	int	temp;

	_HostOut(_host_chan_out,header,2);	/* Command header */
	if((temp = _get_4b()) != 0)
		errno = temp;		/* Host "errno" value */
	}

/************************************************************************/
/*									*/
/*	Routine:	_send_cmd					*/
/*									*/
/*	Send a I/O command to the host system.  The command consists of	*/
/*	a two byte header which specifies the desired function to	*/
/*	perform and a one byte length for the remainder of the command.	*/
/*	If we are using virtual channels we use a four byte header to	*/
/*	also allow the transmission of "_node_number".			*/
/*									*/
/*	Parameters:							*/
/*		"buf"	A pointer to the buffer to send.  NULL if no	*/
/*			additional data beyond the header is sent.	*/
/*		"func"	The function code to send.			*/
/*		"len"	Length in bytes of the additional data to send.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

void
_send_cmd(UCHAR *buf,int func,int len)
	{
	UCHAR			header[4];
/*
 *	If we aren't the root node, or running with virtual channels, just
 *	quit here as we presumably can't communicate with the host server.
 */
	if((_node_number != 1) && (_host_chan_in == _boot_chan_in))
		ProcStop();
/*
 *	Send I/O request to host.
 */
	header[0] = func;
	header[1] = len;
	HSemP(_cio_prot);		/* Lock client/server channel */
	_HostOut(_host_chan_out,header,2);	/* Command header */
/*
 *	Send the remainder of the data in this command (if any).
 */
	if(len > 0)
		_HostOut(_host_chan_out,buf,len);
	}

/************************************************************************/
/*									*/
/*	Routine:	_unlock_{int|str|strm|void}			*/
/*									*/
/*	Unlock the client/server I/O channel semaphore which has been	*/
/*	previously set by a call to "_send_cmd".  Four variations of	*/
/*	the function are provided to allow a pass through argument.	*/
/*									*/
/*	Parameters:							*/
/*		"rval"	The value to return (if any).			*/
/*									*/
/*	Returns:	The parameter value (if any).			*/
/*									*/
/************************************************************************/

int
_unlock_int(int rval)
	{
	HSemV(_cio_prot);		/* Unlock client/server channel */
	return (rval);
	}

char	*
_unlock_str(char *rval)
	{
	HSemV(_cio_prot);		/* Unlock client/server channel */
	return (rval);
	}

FILE	*
_unlock_strm(FILE *rval)
	{
	HSemV(_cio_prot);		/* Unlock client/server channel */
	return (rval);
	}

void
_unlock_void(void)
	{
	HSemV(_cio_prot);		/* Unlock client/server channel */
	}
