/************************************************************************/
/*									*/
/*	File:	vchan.c							*/
/*									*/
/*	Virtual channel communication kernel routines.			*/
/*									*/
/*	The code in this file implements a software emulation of the	*/
/*	T9000 virtual channel communication facility suitable for use	*/
/*	with T2, T4 and T8 Transputers.					*/
/*									*/
/*	Copyright (c) 1991-1992 by Logical Systems.			*/
/*									*/
/*	Written by Kirk Bailey of Logical Systems.			*/
/*									*/
/*	Last update:	11/21/92					*/
/*									*/
/************************************************************************/

#include	<stdlib.h>
#include	<string.h>
#include	<inline.h>
#include	<conc.h>
#include	<cio.h>

#define	TRUE		1
#define	FALSE		0
#define	FOREVER		for(;;)

#define	BOOL		int
#define	UINT		unsigned int

/*
 *	Configuration parameters.
 */

#define	IDLE_CAL_TICKS	100		/* # 64 usec idle time cal ticks*/

#ifdef	T212
#define	DEBUG_STACK	(50+(_DEBUG_REGS*2))	/* _DebugKernel stk size*/
#else	/* !T212 */
#define	DEBUG_STACK	(50+_DEBUG_REGS)	/* _DebugKernel stk size*/
#endif	/* T212 */


#define	_VCC_ISTK	50		/* "_LinkInput" stk size (words)*/
#define	_VCC_OSTK	40		/* "_LinkOutput" stk size(words)*/

/*
 *	Packet header structure for virtual channel messages between adjacent
 *	processors.
 */
typedef	struct
	{
	UINT	len	:16;		/* Packet length		*/
	UINT	chan	:16;		/* Destination virtual channel	*/
	} _PHdr;

/*
 *	"Routing" header structure for "routed" virtual channel messages.
 */
typedef	struct
	{
	UINT	node	:16;		/* "final" dest. node addr	*/
	UINT	chan	:16;		/* "final" dest. virtual chan	*/
	UINT	tag	:16;		/* Buffer tag ID for relayer	*/
	} _RHdr;

/*
 *	Queued output message packet data structure.
 */
struct	_qohdr
	{
	void		*buf;		/* Message body			*/
/*
 *	The "in_link" field is set negative if this message packet originated
 *	on this node.  Otherwise, it holds the input link that the message
 *	was "routed" in on.
 */
	int		in_link;
	int		len;		/* Message length		*/
/*
 *	The "misc" field holds the blocked process descriptor, if the packet
 *	originated on this node, or the "routing" buffer index number (the
 *	"tag"), associated with this packet if the data came from another node.
 */
	int		misc;	
	struct	_qohdr	*next;		/* Point to next msg in output Q*/
	_PHdr		p_hdr;		/* packet header		*/
	_RHdr		r_hdr;		/* "routing" packet header	*/
	};

typedef	struct	_qohdr	_QOHdr;

/*
 *	The following 9 element structure is used to control a physical or
 *	logical link between "adjacent" processors.  One of these structures
 *	will exist for each Transputer which can be directly connected to this
 *	node (either via a T or H-series direct, or C004 connection, or via a
 *	C104 "routing" chip).  In the case of a H1/C104-based network this can
 *	be a large number of structures!  Of course, since this structure is
 *	only required to handle software virtual channel "routing", the larger
 *	H1/C104	cases will generally not occur since the network will contain
 *	a C104 to handle the message "routing".  The worst case could be
 *	provoked by having an extremely large C104 based network, one of whose
 *	nodes was additionally directly connected to another node.  Given this
 *	circumstance, the "middle" node would have one of these structures
 *	for each of the	other nodes connected to the C104 that are "routing"
 *	messages to the	non-C104 connected node by relaying through the
 *	"middle" node.
 */
typedef	struct
	{
	_QOHdr		*ba_head;	/* Buffer available ack Q head	*/
	char		*b_stat;	/* Adjacent node buffer status	*/
	Channel		*chan_in;	/* Input link address		*/
	Channel		*chan_out;	/* Output link address		*/
	PDes		link_out;	/* _LinkOutput process ID	*/
	int		node;		/* Neighboring node identifier	*/
	_QOHdr		*q_head;	/* Output packet queue head	*/
	_QOHdr		*q_tail;	/* Output packet queue tail	*/
	UINT		traffic;	/* Performance monitoring info	*/
#ifdef	T212
	UINT		traffic2;	/* High order 16 bits for T2	*/
#endif	/* T212 */
	} _LLCB;

/*
 *	Data Structures.
 *
 *	First the variables required by the performance monitoring stuff.
 */
static	UINT	_IdleCalib	= 0;	/* # _IdleTime runs in 10 msec	*/
static	UINT	_IdleCount	= 0;	/* # times CPU was found idle	*/
static	int	_IdleWS[4]	= {0,0,0,0};	/* _IdleTime WS		*/
static	BOOL	_PerfMon	= FALSE;/* TRUE/enable performance mon.	*/
static	int	_PerfTime;		/* Time of last data upload	*/

/*
 *	Now a pointer to the adjacency matrix routing	table downloaded by
 *	LD-NET.  The pointer is actually set by "_VChanInit".  The table
 *	data layout is very simple, 2 bits for each possible destination
 *	indicating which _LLCB entry should be used.  This scheme depends on
 *	four or fewer _LLCB's per node.  The next set of comments re H1/
 *	C104 support is germane here also.
 */
static	UINT	*_VC_AMRT	= NULL;

/*
 *	The following structure implements the _LLCB table needed for non-
 *	H1/C104 Transputers.  When the H1/C104 exists this table will have
 *	to also be downloaded by the network loader in the same fashion as
 *	the arrays pointed to by "_VC_AMRT" and "_VC_VLCB"...
 */
static	_LLCB	_VC_TempLLCB[MAX_LINKS] =
{
	{	NULL,NULL,LINK0IN,LINK0OUT,NOPROCESS,-1,NULL,NULL,0	},
	{	NULL,NULL,LINK1IN,LINK1OUT,NOPROCESS,-1,NULL,NULL,0	},
	{	NULL,NULL,LINK2IN,LINK2OUT,NOPROCESS,-1,NULL,NULL,0	},
	{	NULL,NULL,LINK3IN,LINK3OUT,NOPROCESS,-1,NULL,NULL,0	},
};
static	_LLCB	*_VC_LLCB	= _VC_TempLLCB;	/* Log Lnk Cntl Blk ptr	*/

/*
 *	Maximum packet size configuration info for connections which are
 *	either to adjacent or remote processors.
 */
static	int	_VC_PLen_A	= 0;	/* Adjacent destination		*/
static	int	_VC_PLen_R	= 0;	/* Remote destination		*/

/*
 *	"routing" buffer configuration stuff.
 */
static	_QOHdr	*_VC_RBufs	= NULL;	/* Buffer linked list pointer	*/
static	int	_VC_ROffset	= 0;	/* Path length offset		*/

/*
 *	Virtual Channel Control Block table stuff.
 */
static	_VLCB	*_VC_VLCB	= NULL;	/* The _VLCB base pointer	*/
static	UINT	_VC_NVLCB	= 0;	/* # of _VLCB's in table	*/

/*
 *	Function forward declarations.
 */
static	void	_DebugKernel(void);
static	int	_Dec4B(unsigned char *);
static	void	_GoBrkpt(int,int,int);
static	void	_IdleTime(void);
static	void	_IdleTimeStart(void);
static	void	_LinkInput(int);
static	void	_LinkOutput(int);
static	_QOHdr	*_LoadOutput(_QOHdr *,_VLCB *,void *,int,int);
static	int	_LowPriTime(void);
static	int	_RouteLink(int);
static	void	_QueueOutput(int,_QOHdr *);
static	void	_VChanIn(_VLCB *,void *,int);
int		__VChanIn(_VLCB *,void *,int);
void		*_VChanInit(unsigned char *,int,_VLCB *);
void		_VChanOut(_VLCB *,void *,int);
static	void	__VChanOut(_VLCB *,void *,int,int);

/*
 *	_VChanInit
 *
 *	Called from either "_vcmain" or "_ns_vcmain" to initialize and
 *	configure the Virtual Channel communication kernel.  See the LD-NET
 *	source code (function "vchan_download"), for a description of the
 *	memory layout assumed by this function.
 *
 *	Parameters:
 *		"ptr"	A place holder useful only for its address.
 *		"icio"	Non-zero if to initialize host server.
 *		"vptr"	A place holder useful only for its address.
 *
 *	Returns:
 *		The user specified workspace address, less two words.
 */
void	*
_VChanInit(unsigned char *ptr,int icio,_VLCB *vptr)
	{
	char			buf[4];
	int			i;
	int			j;
	void			*mptr;
/*
 *	Create pointer to the beginning of the downloaded Virtual Channel
 *	configuration information.
 */
	ptr = (unsigned char *) &ptr + (sizeof(int) * 3);
/*
 *	Set node identification.
 */
	_node_number = *((int *) ptr);
/*
 *	Set maximum packet size for adjacent destinations.
 */
	ptr += sizeof(int);
	_VC_PLen_A = *((int *) ptr);
/*
 *	Set the maximum packet size for remote ("routed"), destinations.
 */
	ptr += sizeof(int);
	_VC_PLen_R = *((int *) ptr);
/*
 *	Allocate the appropriate # of "routing" buffers for this node.
 */
	ptr += sizeof(int);
	i = *((int *) ptr);

	while(i--)
		{
		if((mptr = malloc(sizeof(_QOHdr) + ((size_t) _VC_PLen_R))) ==
			NULL)
			{
			PStop();
			}
/*
 *	Make the second part of each allocated structure be the data region and
 *	thread the structures into a linked list.
 */
		((_QOHdr *) mptr)->next = _VC_RBufs;
		_VC_RBufs = mptr;
		_VC_RBufs->buf = (((char *) _VC_RBufs) + sizeof(_QOHdr));
		}
/*
 *	Record the initial virtual channel path length offset.
 */
	ptr += sizeof(int);
	_VC_ROffset = *((int *) ptr);
/*
 *	Update the LLCB table with the number of the node connected to the
 *	corresponding link, and an array of integers (one for each available
 *	"routing" buffer on the adjacent node connected to each link).
 */
	for(i = 0; i < MAX_LINKS; i++)
		{
		ptr += sizeof(int);
		_VC_LLCB[i].node = *((int *) ptr);
		ptr += sizeof(int);
		if((j = *((int *) ptr)) != 0)
			{
			if((mptr = malloc((size_t) j)) == NULL)
				PStop();
			_VC_LLCB[i].b_stat = mptr;
/*
 *	Indicate all the buffers are initially empty.
 */
			while(j--)
				*((char *) mptr)++ = 0;
			}
		}
/*
 *	If we are the root node, and are instructed to, inform the host server
 *	(CIO), that it should use the virtual channel protocol in the future,
 *	and what the appropriate maximum packet sizes should be.
 */
	if(icio && (_node_number == 1))
		{
		buf[0] = FN_VCHAN_MODE;
		buf[1] = 4;
		ChanOut(_boot_chan_out,buf,2);
		buf[0] = _VC_PLen_A;
		buf[1] = (_VC_PLen_A >> 8);
		buf[2] = _VC_PLen_R;
		buf[3] = (_VC_PLen_R >> 8);
		ChanOut(_boot_chan_out,buf,4);
		}
/*
 *	Allocate and initialize Virtual Channel routing table.
 */
	ptr += sizeof(int);
	if((_VC_AMRT = malloc((size_t) (i = *((int *) ptr)))) == NULL)
		PStop();
	ptr += sizeof(int);
	(void) memcpy((void *) _VC_AMRT,(void *) ptr,(size_t) i);
/*
 *	Allocate and initialize Virtual Channel VLCB table for this node.
 */
	ptr += i;
	_VC_NVLCB = i = *((int *) ptr) / 4;	/* # of VLCB's needed */
	ptr += sizeof(int);
	if(i != 0)			/* Any VLCB's required? */
		{
		if((vptr = _VC_VLCB = malloc(sizeof(_VLCB) * i)) == NULL)
			PStop();
		while(i--)		/* While more VLCB entries */
			{
			vptr->chan = *ptr | (*(ptr + 1) << 8);
			ptr += 2;
			vptr->dest = *ptr | (*(ptr + 1) << 8);
			ptr += 2;
			vptr->status = 0;
			vptr->in_ws = vptr->out_ws = NOPROCESS;
			vptr++;
			}
		}
	else
		_VC_VLCB = NULL;	/* No VLCB's for this node */
/*
 *	Configure host server I/O channels (if needed).  If so, we also compute
 *	a calibration constant for the processor idle time performance
 *	monitoring stuff.  This adds 6.4 milliseconds to the initialization
 *	time for each node, but this shouldn't be too objectionable...
 */
	if(icio)
		{
		if((_host_chan_in = VChan(VC_HOSTIO)) == NULL)
			PStop();
		_HostIn = VChanIn;
		if((_host_chan_out = VChan(VC_HOSTIO)) == NULL)
			PStop();
		_HostOut = VChanOut;
/*
 *	Run the "_IdleTime" process for 6.4 milliseconds.  The resulting
 *	calibration constant records how many times the "_IdleTime" process can
 *	execute in 6.4 ms when no other processes are active.  This value
 *	will be a function of processor type and speed, program memory layout
 *	and speed, etc.  Note that the "_IdleTime" process is stopped by
 *	explicitly deleting it from the low priority scheduling queue!  This is
 *	safe to do since we know that it is the only other low priority process
 *	running at present!  Don't try this at home kids...
 */
		_IdleTimeStart();	/* Start "_IdleTime" process */
		ProcWait(IDLE_CAL_TICKS);	/* Wait ~6.4 milliseconds */
		SetLoPriQ((void *) NOPROCESS,(void *) NOPROCESS);
		_IdleCalib = _IdleCount;/* # times _IdleTime runs in 6.4 ms */
		_IdleCount = 0;		/* Reset for later use */
		}
/*
 *	Start appropriate link sentinel processes.
 */
	for(i = 0; i < MAX_LINKS; i++)
		{
		if(_VC_LLCB[i].node >= 0)	/* If required */
			{
			if((mptr = (unsigned char *) malloc(_VCC_ISTK *
				sizeof(int))) == NULL)
				{
				PStop();
				}
			PRun(PSetup(mptr,_LinkInput,
				(int) (_VCC_ISTK * sizeof(int)),1,i));
			}
		}
/*
 *	Configure host debug I/O channel (if needed).
 */
	if(icio)
		{
		if((mptr = malloc(DEBUG_STACK * sizeof(int))) == NULL)
			PStop();
		PRun(PSetup(mptr,_DebugKernel,(int)
			(DEBUG_STACK * sizeof(int)),0));
		}

	return ((void *) ptr);
	}

/*
 *	_LinkInput
 *
 *	This process serves as an input "sentinel" for each logical or
 *	physical link connected to this processor upon which software virtual
 *	channel communication will take place.
 *
 *	Returns:
 *		None.
 */
static	void
_LinkInput(int link)
	{
	_LLCB			*active_LLCB;
	_VLCB			*active_VLCB;
	Channel			*chan_in;
	int			length;
	_PHdr			packet_header;
	void			*vptr;

	active_LLCB = &_VC_LLCB[link];
	chan_in = active_LLCB->chan_in;
/*
 *	Allocate space for the output process associated with this link.
 *	If we can't allocate the space, don't spawn the process.
 */
	if((vptr = malloc(_VCC_OSTK * sizeof(int))) != NULL)
		{
		PRun(PSetup(vptr,_LinkOutput,(int) (_VCC_OSTK * sizeof(int)),
			(int) (sizeof(int) / _WORD_SIZE),link));
		}
/*
 *	Execute once for each message read from the input link.
 */
	FOREVER
		{
/*
 *	Read packet header.
 */
		ChanIn(chan_in,(void *) &packet_header,
			(int) (sizeof(packet_header)));
/*
 *	If the packet means that a remote "routing" buffer is again available,
 *	clear out the corresponding entry in the "routing buffer available"
 *	table.  If the associated "_LinkOutput" process is blocked, wake it up.
 */
		if(packet_header.len == (_VCH_CNTL | _VCH_RRDY))
			{
			active_LLCB->b_stat[packet_header.chan] = FALSE;
			if(active_LLCB->link_out != NOPROCESS)
				{
				PRun(active_LLCB->link_out);
				active_LLCB->link_out = NOPROCESS;
				}
			}
/*
 *	Otherwise, see if the packet belongs to a "routed" virtual channel.
 */
		else if(packet_header.chan & _VCH_CNTL)
			{
/*
 *	The packet will need to be "routed" to another output link.  Grab a
 *	"routing" buffer from the free list and copy the data we have already
 *	read into it.
 */
			vptr = _VC_RBufs;	/* Get first item on list */
			_VC_RBufs = _VC_RBufs->next;	/* Remove from list */
			((_QOHdr *) vptr)->p_hdr = packet_header;
/*
 *	Read second "routing" header.
 */
			ChanIn(chan_in,(void *) &((_QOHdr *) vptr)->r_hdr,6);
/*
 *	If the packet isn't a control message, and there is anything to read,
 *	read the data into the routing process input buffer.
 */
			if(((packet_header.len & _VCH_CNTL) == 0) &&
				((length = (packet_header.len & (~_VCH_EOMP)))
					!= 0))
				{
				ChanIn(chan_in,(void *) ((_QOHdr *) vptr)->buf,
					length);
				}
			else
				length = 0;

			((_QOHdr *) vptr)->in_link = link;	/* Inp link */
			((_QOHdr *) vptr)->len = length;	/* Data len */
/*
 *	Note that we need remember the "routing buffer table index" (the
 *	"tag"), which was passed by the relaying processor so that we can tell
 *	it which "routing" buffer is again available when this message has
 *	been sent along to the next node.
 */
			((_QOHdr *) vptr)->misc =
				((_QOHdr *) vptr)->r_hdr.tag;
/*
 *	Send the packet along to the final destination.
 */
			_QueueOutput(_RouteLink((int)
				((_QOHdr *) vptr)->r_hdr.node),
					(_QOHdr *) vptr);
			}
		else
			{
/*
 *	The packet is for a regular virtual channel on this node.
 */
			active_VLCB = &_VC_VLCB[packet_header.chan];
/*
 *	If the message is part of the protocol (a "control" message), perform
 *	the specified function.
 */
			if(packet_header.len & _VCH_CNTL)
				{
				switch ((int) (packet_header.len &
					((~_VCH_CNTL) & 0xFFFF)))
					{
					case	_VCH_IRDY:
						if(active_VLCB->out_ws ==
							NOPROCESS)
							{
							active_VLCB->status |=
								_VCS_IRDY;
							}
						else
							{
							PRun(active_VLCB->
								out_ws);
							active_VLCB->out_ws =
								NOPROCESS;
							}
						break;

					case	_VCH_ORDY:
						active_VLCB->status |=
							_VCS_ORDY;
/*
 *	If there is an ALT construct waiting on this input channel, write to
 *	the "in_ws" word which the ALT will be using as the "fake" input
 *	channel.  This will cause the microcode to trigger the ALT.  The
 *	various virtual channel ALT wrapper functions will restart this
 *	process if this actually happens (with a RUNP).
 */
						if(active_VLCB->status &
							_VCS_IALT)
							{
							ChanOut(&active_VLCB->
								in_ws,
								&packet_header,
								1);
							}
						break;
/*
 *	If we have a protocol error, kill the process.
 */
					default:
						PStop();
					}
				}
			else
				{
/*
 *	If the packet wasn't a control message read any message body it might
 *	have into the final destination.  We must also make sure the input
 *	buffer is large enough.  If not, "malloc" up some temporary storage to
 *	read the message into...
 */
				if(length = (packet_header.len & (~_VCH_EOMP)))
					{
					if((active_VLCB->in_len + length) >
						active_VLCB->in_max)
						{
/*
 *	Too long, get a VERY temporary buffer, then read and discard the data.
 */
						if((vptr = malloc((size_t)
							length)) == NULL)
							{
							PStop();
							}
						ChanIn(chan_in,vptr,length);
						free(vptr);
/* Signal too long */				active_VLCB->in_len =
							active_VLCB->in_max+1;
						}
					else
						{
						ChanIn(chan_in,(void *)
							active_VLCB->in_buf,
								length);
/*
 *	Advance the buffer pointer.  Set the input message length as
 *	appropriate.
 */
/* Bump buffer pointer */			active_VLCB->in_buf += length;
/* Bump count */				active_VLCB->in_len += length;
						}
					}
/*
 *	If the packet is the last in a given message, schedule the input
 *	process for execution.  If no input process is blocked we set the
 *	"in_ws" process descriptor to zero so that "_VChanIn" knows that
 *	it can skip blocking (ie. the message has already been received).
 */
				if(packet_header.len & _VCH_EOMP)
					{
					if(active_VLCB->in_ws == NOPROCESS)
						active_VLCB->in_ws = 0;
					else
						PRun(active_VLCB->in_ws);
					}
				}
			}
		}
	}

/*
 *	_LinkOutput
 *
 *	This process is awakened by the corresponding "_LinkInput" process to
 *	route and deliver message packets for a specified output link.
 *
 *	Returns:
 *		None.
 */
static	void
_LinkOutput(int link)
	{
	_LLCB			*active_LLCB;
	Channel			*chan_out;
	char			*cptr;
	_QOHdr			*packet;
	_PHdr			packet_header;
	_QOHdr			**qptr;
	int			temp;

	active_LLCB = &_VC_LLCB[link];
	chan_out = active_LLCB->chan_out;
/*
 *	Send packets out the link.
 */
	FOREVER
		{
/*
 *	First see if we have any "routing buffer again available" messages to
 *	send.  If so, send them in preference to data messages.
 */
		if(active_LLCB->ba_head != NULL)
			{
/*
 *	Make a local copy of the data to send, remove the packet from the
 *	output queue, and return it to the global free list.
 */
			packet = active_LLCB->ba_head;
			packet_header = packet->p_hdr;
			active_LLCB->ba_head = packet->next;
			packet->next = _VC_RBufs;
			_VC_RBufs = packet;
/*
 *	Send the "routing buffer again available" message.
 */
			ChanOut(chan_out,(void *) &packet_header,
				(int) (sizeof(packet_header)));
			continue;
			}
/*
 *	Well, we don't have any "routing buffer again available" messages.
 *	If we also don't have any data messages to send, it is time to block.
 */
		if(active_LLCB->q_head == NULL)
			{
			PHalt(&active_LLCB->link_out);
			continue;
			}
/*
 *	We have some data to send!
 *
 *	Go through the output packets in order and see if each may be sent.
 *	When one is identified that can be sent, do so.  To be sent, the packet
 *	either must not require a "routing" buffer, or a buffer of the
 *	appropriate "class" (in the "Structured Buffer Pool Deadlock Avoidance"
 *	sense), must be available on the adjacent node.  If no output packet is
 *	available that meets these requirements, block.
 */
		cptr = active_LLCB->b_stat;	/* "routing" buffer table */
		qptr = &active_LLCB->q_head;	/* Queue head ptr ptr */
		while(*qptr != NULL)	/* While something left in the queue */
			{
			packet = *qptr;
/*
 *	See if we need to use the "routed" or regular protocol.
 *
 *	An optimization:  We don't need to use up a "routing" buffer on the
 *	adjacent node if it will be the final destination (last hop), of a
 *	"routed" message.  Thus if the adjacent node is the "final" destination
 *	of a message, use the non-"routed" protocol.
 */
			if(packet->r_hdr.node == active_LLCB->node)
				{
				packet->p_hdr.chan = packet->r_hdr.chan;
/*
 *	Send message packet header.
 */
				ChanOut(chan_out,(void *) &packet->p_hdr,
					(int) (sizeof(packet->p_hdr)));
				break;
				}
			else
				{
/*
 *	The packet needs a "routing" buffer on the next node.  See if one is
 *	available.
 */
				temp = ((int) (packet->p_hdr.chan &
						((~_VCH_CNTL) & 0xFFFF)) + 1) -
					((int) (cptr - active_LLCB->b_stat));

				while(temp-- > 0)
					{
					if(! *cptr++)
/* Quit if we find an empty buffer */		break;
					}
				if(temp >= 0)
/*
 *	We have an available buffer!  Indicate that it is now in use and store
 *	the associated table index in the "routing" packet header.
 */
					{
/* Buffer is used */			*--cptr = TRUE;
/* Tag to return */			packet->r_hdr.tag = cptr -
						active_LLCB->b_stat;
/* Update hop count */			packet->p_hdr.chan++;
/*
 *	Send message packet header.
 */
					ChanOut(chan_out,
						(void *) &packet->p_hdr,
						(int) (sizeof(packet->p_hdr)));
/*
 *	Send "routing" packet header.
 */
					ChanOut(chan_out,
						(void *) &packet->r_hdr,6);
					break;
					}
				else
					{
/* 
 *	No "routing" buffer of the appropriate class was available for this
 *	packet, time to check the next packet.
 */
					qptr = &(packet->next);
					}
				}
			}
		if(*qptr == NULL)
			{
/*
 *	No packet is waiting that doesn't require a "routing" buffer, or if
 *	"routing" buffers are available, they are not small enough for the
 *	any of the current packets to use.  Time to sleep.
 */
			PHalt(&active_LLCB->link_out);
			continue;	/* Start the process over */
			}
/*
 *	Send actual message data (if any to send).
 */
		if(packet->len != 0)
			{
			ChanOut(chan_out,packet->buf,packet->len);
/*
 *	If performance monitoring is enabled, update the count of bytes sent in
 *	this monitoring period.  If the processor is a 16 bit type, perform
 *	a double precision accumulate operation.
 */
			if(_PerfMon)
				{
#ifdef	T212
				temp = (int) &active_LLCB->traffic;

				#pragma	asm
				_asm_eval1(packet->len);
				.LDC	0	/* Zero carry-in */
				REV
				LDL	[temp]	/* Address of "traffic" */
				LDNL	0	/* Get "traffic" LSW */
				LSUM		/* Add LSW's */
				LDL	[temp]	/* Address of "traffic" LSW */
				STNL	0	/* Store result LSW */
				LDL	[temp]	/* Address of "traffic" LSW */
				LDNL	1	/* Get "traffic" MSW */
				.LDC	0	/* Other high order MSW */
				LSUM		/* Don't want error set! */
				LDL	[temp]	/* Address of "traffic" LSW */
				STNL	1	/* Store "traffic" MSW */
				#pragma	endasm
#else	/* !T212 */
				active_LLCB->traffic += packet->len;
#endif	/* T212 */
				}
			}
/*
 *	Delete this packet from the output queue.
 */
		if((packet->next == NULL) && (packet != active_LLCB->q_head))
			{
/*
 *	It is necessary to update the queue tail pointer also since we are
 *	deleting the last item in a queue with more than one packet in it.
 *	Note that we aren't using the queue in a "pure" fashion in this
 *	routine...
 */
			active_LLCB->q_tail = (_QOHdr *) (((char *) qptr) -
				(((char *) &packet->next) -
					((char *) packet)));
			}
		*qptr = packet->next;	/* Update regular link pointer */
/*
 *	If the message has been "re-routed", we must inform the previous node
 *	that the routing buffer is again available.
 */
		if((temp = packet->in_link) >= 0)
			{
/*
 *	Prepare the "routing buffer again available" message using the
 *	previously recorded routing buffer table index (the "tag" passed by the
 *	forwarding node).
 */
			packet->p_hdr.len = (_VCH_CNTL | _VCH_RRDY);
			packet->p_hdr.chan = packet->misc;
/*
 *	Queue the message for output back to the relaying node.
 */
			packet->next = _VC_LLCB[temp].ba_head;
			_VC_LLCB[temp].ba_head = packet;
/*
 *	Wake up the corresponding "_LinkOutput" process if it is currently
 *	blocked.
 */
			if(_VC_LLCB[temp].link_out != NOPROCESS)
				{
				PRun(_VC_LLCB[temp].link_out);
				_VC_LLCB[temp].link_out = NOPROCESS;
				}
			}
/*
 *	We have finished with this packet.  If the packet originated on this
 *	node we need to wake up the associated "QueueOutput" output process.
 */
		if(temp < 0)
			PRun(packet->misc);
		}
	}

/*
 *	_LoadOutput
 *
 *	Initialize a "_QOHdr" structure for later transmission.  This function
 *	is only called for message packets which originate on THIS node!
 *
 *	Returns:
 *		A pointer to the "_QOHdr" structure which was initialized.
 */
static	_QOHdr	*
_LoadOutput(_QOHdr *packet,_VLCB *vchan,void *buf,int len,int status)
	{
	packet->buf = buf;		/* Packet data contents */
	packet->in_link = -1;		/* Packet originated on this node */
	packet->len = len;		/* Packet data length */
	packet->r_hdr.node = vchan->dest;	/* Destination node */
	packet->r_hdr.chan = vchan->chan;	/* Destination "vchan" */
	packet->p_hdr.len = status | len;	/* Header "status" field */
/*
 *	Note that in the case of a "routed" message the "p_hdr.chan" field
 *	holds the count of how many links the packet has already traversed.
 *	We are allowed to "lie" about this by biasing the value to a non-zero
 *	value to support the existance of "extra" buffers which have been
 *	allocated on a per-link basis.  The "_VCH_CNTL" is OR'ed in so that
 *	the packet defaults to being a "routed" packet.  Note that if the
 *	packet ends up NOT being "routed", then the "r_hdr" structure will
 *	be mostly ignored (except the "r_hdr.chan" field which will be used
 *	in place of "p_hdr.chan").
 */
	packet->p_hdr.chan = _VCH_CNTL + _VC_ROffset;

	return (packet);
	}

/*
 *	_RouteLink
 *
 *	Determine the correct output link to use to go towards the specified
 *	destination node.
 *
 *	Returns:
 *		The desired output link.
 */
static	int
_RouteLink(int dest)
	{
	int			link;
/*
 *	Use the "routing" adjacency matrix to determine which physical link
 *	to send the message on.
 */
	link = (sizeof(UINT) * 8) / MAX_LINKSLOG;

	return ((int) ((_VC_AMRT[dest / link] >>
		((dest % link) * MAX_LINKSLOG)) & ((1 << MAX_LINKSLOG) - 1)));
	}

/*
 *	_QueueOutput
 *
 *	Queue a packet for transmission on the specified output link.
 *
 *	Returns:	None.
 */
static	void
_QueueOutput(int link,_QOHdr *qohdr)
	{
	_LLCB			*active_LLCB;

	active_LLCB = &_VC_LLCB[link];

	qohdr->next = NULL;		/* Will be the last packet in the Q */
	if(active_LLCB->q_head == NULL)	/* Output queue is empty */
		active_LLCB->q_head = active_LLCB->q_tail = qohdr;
	else				/* Q has something in it, add to end */
		{
		active_LLCB->q_tail->next = qohdr;
		active_LLCB->q_tail = qohdr;
		}
/*
 *	If the output process is blocked, wake it up.
 */
	if(active_LLCB->link_out != NOPROCESS)
		{
		PRun(active_LLCB->link_out);
		active_LLCB->link_out = NOPROCESS;
		}
/*
 *	If the message originated on this node, block until the packet has been
 *	transmitted.
 */
	if(qohdr->in_link < 0)
		PHalt(&qohdr->misc);
	}

/*
 *	VChan
 *
 *	Translate a logical virtual channel into a channel pointer which can
 *	be used with VChanIn/VChanOut/etc.
 *
 *	Returns:
 *		A pointer to the pre-allocated channel.  If the logical channel
 *		doesn't exist, return NULL.  Note that if the virtual channel
 *		is the host debug channel we actually use the regular host I/O
 *		channel status information to determine the return value since
 *		the host debug channel has both the destination node and
 *		channel set to zero in accordance with the special way that
 *		CIO uses the debugging kernel.
 */
Channel	*
VChan(int vchan)
	{
	if(vchan == VC_DEBUGIO)
		{
		if((_VC_VLCB[VC_HOSTIO].chan | _VC_VLCB[VC_HOSTIO].dest) == 0)
			return (NULL);
		}
	else if((vchan >= _VC_NVLCB) ||
		((_VC_VLCB[vchan].chan | _VC_VLCB[vchan].dest) == 0))
		{
		return (NULL);
		}

	return ((Channel *) (((char *) &_VC_VLCB[vchan]) + 1));
	}

/*
 *	VChanIn
 *
 *	Reads the specified number of bytes from the specified "virtual"
 *	channel into memory starting at the specified location.  Note that if
 *	the low order bit of the channel address is non-zero then it is assumed
 *	that additional software processing will be needed to complete the
 *	transfer ("routing", etc.), and a helper function to that end is
 *	called.
 *
 *	Returns:
 *		None.
 */
void
VChanIn(Channel *chan,void *buf,int size)
	{
	#pragma	asm
	ldl	[chan]
	ldc	1
	and				;See if special processing needed
	cj	@VCi1
	j	@_VChanIn
;
;	Either a "soft" channel or a hardware virtual link.
;
VCi1
	ldl	[buf]			;Buffer address
	ldl	[chan]			;Channel address
	ldl	[size]			;Transfer size
	in				;Read bytes
	#pragma	endasm
	}

static	void
_VChanIn(_VLCB *vchan,void *buf,int size)
	{
	int			old_priority;

	if(size < 0)
		return;

	vchan = (_VLCB *) (((char *) vchan) - 1);	/* Remove bottom bit */

	old_priority = __VChanIn(vchan,buf,size);	/* Get input message */
/*
 *	If the input message didn't match the output message size, kill the
 *	process.
 */
	if(vchan->in_len != size)
		PStop();
/*
 *	If this process was originally at low priority we need to switch back.
 */
	vchan->status &= (~_VCS_ORDY);	/* In case it was used */
	vchan->in_ws = NOPROCESS;

	if(old_priority)
		ProcToLow();
	}

int
__VChanIn(_VLCB *vchan,void *buf,int size)
	{
	int			old_priority;
	_VLCB			*other_vchan;
	_QOHdr			packet;

	old_priority = ProcToHigh();	/* Change to high priority */
/*
 *	Duplicate use of a virtual channel is a programming error!  The
 *	"second" process is killed in place.  This test isn't fool proof since
 *	if the input process is in the middle of communicating rather than
 *	blocked waiting it is possible for "vchan->in_ws" to be NOPROCESS
 *	anyway...
 */
	if(vchan->in_ws != NOPROCESS)
		PStop();		/* Already in use, kill this process */
/*
 *	Setup the appropriate _VLCB parameters for the input operation.
 */
	vchan->in_buf = buf;		/* Starting input buffer pointer */
	vchan->in_len = 0;		/* Current input buffer size */
	vchan->in_max = size;		/* Maximum input buffer size */
/*
 *	Tell the destination virtual channel that we are ready for input.
 *	Note that if both ends of the channel are on this node we can simplify
 *	matters considerably.
 */
	if(vchan->dest == _node_number)	/* Both ends on this node! */
		{
		other_vchan = &_VC_VLCB[vchan->chan];
		other_vchan->status |= _VCS_IRDY;
		if(other_vchan->out_ws != NOPROCESS)
			{
			PRun(other_vchan->out_ws);
			other_vchan->out_ws = NOPROCESS;
			}
		}
	else				/* Data is for another node */
		{
		_QueueOutput(_RouteLink(vchan->dest),
			_LoadOutput(&packet,vchan,NULL,0,
				(int) (_VCH_CNTL | _VCH_IRDY)));
		}
/*
 *	Block waiting for the input to be complete.  Note that in the unlikely
 *	case where the message has ALREADY been received prior to executing the
 *	next statement we check to see if "in_ws" has been set to zero by
 *	"_LinkInput" indicating that we can avoid blocking.
 */
	if(vchan->in_ws != 0)
		PHalt(&vchan->in_ws);

	return (old_priority);
	}

/*
 *	VChanOut
 *
 *	Writes the specified number of bytes to the specified "virtual"
 *	channel from memory starting at the specified location.  Note that if
 *	the low order bit of the channel address is non-zero then it is assumed
 *	that additional software processing will be needed to complete the
 *	transfer ("routing", etc.), and a helper function to that end is
 *	called.
 *
 *	Returns:
 *		None.
 */
void
VChanOut(Channel *chan,void *buf,int size)
	{
	#pragma	asm
	ldl	[chan]
	ldc	1
	and				;See if special processing needed
	cj	@VCo1
	j	@_VChanOut
;
;	Either a "soft" channel or a hardware virtual link.
;
VCo1
	ldl	[buf]			;Buffer address
	ldl	[chan]			;Channel address
	ldl	[size]			;Transfer size
	out				;Write bytes
	#pragma	endasm
	}

void
_VChanOut(_VLCB *vchan,void *buf,int size)
	{
	int			link;
	int			old_priority;
	_VLCB			*other_vchan;
	_QOHdr			packet;
	int			packet_length;
	int			status;

	if(size < 0)
		return;

	old_priority = ProcToHigh();	/* Change to high priority */

	vchan = (_VLCB *) (((char *) vchan) - 1);	/* Remove bottom bit */
/*
 *	Duplicate use of a virtual channel is a programming error!  The
 *	"second" process is killed in place.  This test isn't fool proof since
 *	if the input process is in the middle of communicating rather than
 *	blocked waiting it is possible for "vchan->out_ws" to be NOPROCESS
 *	anyway...
 */
	if(vchan->out_ws != NOPROCESS)
		PStop();		/* Already in use, kill this process */
/*
 *	If the other end of the virtual channel is also on this node, create
 *	a pointer to the other VLCB.
 */
	if(vchan->dest == _node_number)	/* Both ends on this node! */
		other_vchan = &_VC_VLCB[vchan->chan];
	else				/* Other end is on other node */
		{
		link = _RouteLink(vchan->dest);
		(void) _LoadOutput(&packet,vchan,NULL,0,
			(int) (_VCH_CNTL | _VCH_ORDY));
		}
/*
 *	If an input process is ready, start sending data.  Otherwise send a
 *	message to the input process telling it that the output process is
 *	ready (so that ALT stuff works), and block until the input process is
 *	ready.
 */
	if(! (vchan->status & _VCS_IRDY))
		{
		if(vchan->dest == _node_number)	/* Both ends on this node! */
			{
			other_vchan->status |= _VCS_ORDY;
/*
 *	If there is an ALT construct waiting on this input channel, write to
 *	the "in_ws" word which the ALT will be using as the "fake" input
 *	channel.  This will cause the microcode to trigger the ALT.  The
 *	various virtual channel ALT wrapper functions will restart this
 *	process if this actually happens (with a RUNP).
 */
			if(other_vchan->status & _VCS_IALT)
				ChanOut(&other_vchan->in_ws,&old_priority,1);
			}
		else			/* Data is for another node */
			_QueueOutput(link,&packet);

		if(! (vchan->status & _VCS_IRDY))
			PHalt(&vchan->out_ws);
		}
/*
 *	Send the message.
 */
	if(vchan->dest == _node_number)	/* Both ends on this node! */
		{
		vchan->status &= (~_VCS_IRDY);	/* Clear out status */
		if(size > other_vchan->in_max)	/* Too long? */
			other_vchan->in_len = other_vchan->in_max + 1;
		else
			{
			if(size > 0)
				{
				(void) memcpy(other_vchan->in_buf,buf,
					(size_t) size);
				}
			other_vchan->in_len = size;
			}
		PRun(other_vchan->in_ws);
		}
	else				/* Data is for another node */
		{
/*
 *	Determine the maximum message packet size depending on whether the
 *	message is "routed" or not.  The message is "routed" if the final
 *	destination is NOT an immediately adjacent node.
 */
		if(vchan->dest != _VC_LLCB[link].node)
			packet_length = _VC_PLen_R;
		else
			packet_length = _VC_PLen_A;
/*
 *	Send message.
 */
		do
			{
			if(size > packet_length)
				{
/*
 *	Break long regular messages up into multiple shorter messages.
 */
				size -= packet_length;
				status = 0;
				}
			else
				{
				vchan->status &= (~_VCS_IRDY);	/* !IRDY */
				packet_length = size;
				size = 0;
/*
 *	Send short messages (and the last packet of longer ones), through
 *	with the _VCH_EOMP flag set.
 */
				status = _VCH_EOMP;
				}
/*
 *	Send some or all of the message.
 */
			packet.buf = buf;
			packet.len = packet_length;
			packet.p_hdr.len = status | packet_length;
			packet.p_hdr.chan = _VCH_CNTL;	/* Default "routed" */
			_QueueOutput(link,&packet);

			((char *) buf) += packet_length;

			} while(size > 0);
		}
/*
 *	If this process was originally at low priority we need to switch back.
 */
	if(old_priority)
		ProcToLow();
	}

/*
 *	_DebugKernel
 *
 *	This process handles debugging information requests from the host.
 *
 *	Returns:
 *		None.
 */
static	void
_DebugKernel(void)
	{
	void 			*dadr;
	unsigned char		debug_buf[_DEBUG_REGS * 4];
	int			dlen;
	Channel			*host_req;

	if((host_req = VChan(VC_DEBUGIO)) == NULL)
		PStop();		/* Not much we can do about it */

	FOREVER				/* Never terminates */
		{
		VChanOut(host_req,debug_buf,1);	/* Synchronization host msg */
		VChanIn(host_req,debug_buf,9);
		dlen = _Dec4B(&debug_buf[1]);	/* Data length */
		dadr = (void *) _Dec4B(&debug_buf[5]);	/* Data address */
		switch((int) debug_buf[0])
			{
			case	NV_RMEM:	/* Handle read mem requests */
				VChanOut(host_req,dadr,dlen);
				break;

			case	NV_WMEM:	/* Handle write mem requests */
				VChanIn(host_req,dadr,dlen);
				break;

			case	NV_GBRKPT:	/* Proceed from breakpoint */
				VChanIn(host_req,debug_buf,_DEBUG_REGS * 4);
				dlen = _Dec4B(&debug_buf[_DEBUG_WS * 4]);
/*
 *	If we have a T212 we don't need to restore the floating point status.
 */
#ifdef	T212
/* WS without priority */	dadr = (void *) (((int *) (dlen & ~1)) - 4);
/* Temporary PC */		((int *) dadr)[-1] = (int) _GoBrkpt;
/* PC */			((int *) dadr)[0] = _Dec4B(&debug_buf[
					_DEBUG_PC * 4]);
/* Areg */			((int *) dadr)[1] = _Dec4B(&debug_buf[
					_DEBUG_A * 4]);
/* Breg */			((int *) dadr)[2] = _Dec4B(&debug_buf[
					_DEBUG_B * 4]);
/* Creg */			((int *) dadr)[3] = _Dec4B(&debug_buf[
					_DEBUG_C * 4]);
#else	/* !T212 */
/* WS without priority */	dadr = (void *) (((int *) (dlen & ~1)) -
					(_DEBUG_REGS - 1));
/* Temporary PC */		((int *) dadr)[-1] = (int) _GoBrkpt;
/* FAreg/FBreg/FCreg */		(void) memcpy(dadr,(void *) &debug_buf[
					_DEBUG_FA_LOW * 4],(size_t)
						((_DEBUG_REGS-_DEBUG_FA_LOW) *
							4));
/* PC/Areg/Breg/Creg */		(void) memcpy(((void *) &(((int *) dadr)[
						_DEBUG_REGS - _DEBUG_FA_LOW])),
					((void *) &debug_buf[_DEBUG_PC * 4]),
						(size_t) ((_DEBUG_FA_LOW -
							_DEBUG_PC) * 4));
#endif	/* T212 */

/* Restart process */		PRun(((int) dadr) | (dlen & 1));
				break;

			case	NV_PMSTART:	/* Start perf. monitoring */
				if(! _PerfMon)	/* Already enabled? */
					{	/* No, start stuff */
/* Initial time stamp */		_PerfTime = _LowPriTime();
					_PerfMon = TRUE;
					_IdleTimeStart();
					}
				dadr = debug_buf;
				bzero(dadr,(size_t) 8);
				*((int *) dadr) = IDLE_CAL_TICKS;
				dadr = ((char *) dadr) + 4;
				*((int *) dadr) = _IdleCalib;
				VChanOut(host_req,debug_buf,2 * 4);
				break;

			case	NV_PMPOLL:	/* Get perf. monitoring info */
#if	((MAX_LINKS + 2) > _DEBUG_REGS)
#error	"Size of 'debug_buf' needs attention in 'vchan.c'"
#endif	/* ((MAX_LINKS + 1) > DEBUG_REGS) */
				
				dadr = debug_buf;
				bzero(dadr,(size_t) 8);
				dlen = _LowPriTime();
				*((int *) dadr) = dlen - _PerfTime;
				_PerfTime = dlen;
				dadr = ((char *) dadr) + 4;
				*((int *) dadr) = _IdleCount;
				_IdleCount = 0;
				for(dlen = 0; dlen < MAX_LINKS; dlen++)
					{
					dadr = ((char *) dadr) + 4;
					(void) memcpy(dadr,
						&_VC_LLCB[dlen].traffic,
							(size_t) 4);
					bzero(&_VC_LLCB[dlen].traffic,
						(size_t) 4);
					}
				VChanOut(host_req,debug_buf,
					(MAX_LINKS + 2) * 4);
				break;
			}
		}
	}

/*
 *	_Dec4B
 *
 *	Convert four characters representing the little-endian value of an
 *	integer into an integer in the native word length.
 *
 *	Returns:
 *		The converted integer.
 */
static	int
_Dec4B(unsigned char *sptr)
	{
	int			rval;

	rval = *sptr++;
	rval |= (*sptr++ << 8);
#ifndef	T212
	rval |= (*sptr++ << 16);
	rval |= (*sptr << 24);
#endif	/* !T212 */

	return (rval);
	}

/*
 *	_GoBrkpt
 *
 *	Helper function for _DebugKernel to allow a process to continue from a
 *	breakpoint.
 *
 *	Returns:
 *		The WS, PC, Areg, Breg and Creg have the specified values.
 */
static	void
_GoBrkpt(int a,int b,int c)
	{
/*
 *	Restore the floating point ALU state.
 */
#ifdef	T800				/* T800 CPU:  There be Magik here! */
	int			fpstate[_DEBUG_REGS - _DEBUG_FA_LOW];
	;
	#pragma	asm
	ajw	_DEBUG_REGS - _DEBUG_FA_LOW
	_asm_eval1(&fpstate[_DEBUG_FC_LOW - _DEBUG_FA_LOW]);
	FPLDTEST			/* Load FCreg & test bits */
	_asm_eval1(&fpstate[_DEBUG_FB_LOW - _DEBUG_FA_LOW]);
	FPLDTEST			/* Load FBreg & test bits */
	_asm_eval1(&fpstate[_DEBUG_FA_LOW - _DEBUG_FA_LOW]);
	FPLDTEST			/* Load FAreg & test bits */
	#pragma	endasm
#else	/* !T800 */
#ifndef	T212				/* Must be T414 CPU */
	int			fpstate[_DEBUG_REGS - _DEBUG_FA_LOW];
	;
	#pragma	asm
	ajw	_DEBUG_REGS - _DEBUG_FA_LOW
	.EXT	_FPstate
	_asm_eval1(&fpstate[_DEBUG_FC_LOW - _DEBUG_FA_LOW]);
	.LDC	_FPstate+16		/* Address of simulated FCreg */
	.LDC	12			/* Also copy the round mode bits */
	MOVE
	_asm_eval1(&fpstate[_DEBUG_FB_LOW - _DEBUG_FA_LOW]);
	.LDC	_FPstate+8		/* Address of simulated FBreg */
	.LDC	8
	MOVE
	_asm_eval1(&fpstate[_DEBUG_FA_LOW - _DEBUG_FA_LOW]);
	.LDC	_FPstate+0		/* Address of simulated FAreg */
	.LDC	8
	MOVE
	#pragma	endasm
#endif	/* !T212 */
#endif	/* T800 */
/*
 *	Restore integer ALU state.
 */
	#pragma	asm
	ldl	[c]
	ldl	[b]
	ldl	[a]
	#pragma	endasm
	}

/*
 *	The following code implements a simple "idle time" monitoring facility.
 */

/*
 *	_IdleTime
 *
 *	This process is spawned to keep track of how much the processor is
 *	idle waiting for something to do.  It does this by keeping track of how
 *	often when it is invoked it is the only process ready to run.  The
 *	"_IdleCount" variable is used to keep track of how often this was the
 *	case.  This process requires a four word workspace, two of which are
 *	used to temporarily hold the low priority queue pointers, and the other
 *	two hold the descheduled process info.
 *
 *	Returns:
 *		Never!
 */
static	void
_IdleTime()
	{
	int			q_ptrs[2];

	FOREVER
		{
/*
 *	The following code checks to see if there is anything else in the
 *	low priority task queue by adding another copy of this process, and
 *	then seeing if the queue front pointer points to this process.  This
 *	saves a few microseconds over simply checking for a NOPROCESS queue
 *	pointer (we have to add ourselves back onto the queue anyway).
 */
		PRun(((int) q_ptrs) | 1);	/* Schedule ourselves again */
		GetLoPriQ(q_ptrs);	/* Get low priority task queue ptrs */
		_IdleCount += (q_ptrs[0] == ((int) q_ptrs));	/* Us? */
		PStop();		/* Die (and fill in WS[-1] with PC) */
		}
	}

/*
 *	_IdleTimeStart
 *
 *	Start the "_IdleTime" process to monitor how busy the processor is.
 *
 *	Returns:
 *		None.
 */
static	void
_IdleTimeStart()
	{
	_IdleWS[3] = (int) _IdleTime;	/* Initial process PC */
	PRun(((int) &_IdleWS[4]) | 1);	/* Start _IdleTime process */
	}

/*
 *	_LowPriTime
 *
 *	Switch to low priority, read the low priority timer, and switch back to
 *	high priority.
 *
 *	Returns:
 *		The value of the low priority timer.
 */
static	int
_LowPriTime(void)
	{
	int			result;

	ProcToLow();
	result = Time();
	ProcToHigh();

	return (result);
	}
