/************************************************************************/
/*									*/
/*	File:	cio.c							*/
/*									*/
/*	Host side of "C" I/O server for the INMOS Transputer.		*/
/*									*/
/*	This program must be linked with a version of the INMOS		*/
/*	"link.c" I/O interface to customize it to the associated	*/
/*	hardware!							*/
/*									*/
/*	Written by Dave Regan & Kirk Bailey of Logical Systems.		*/
/*									*/
/*		Copyright (c) 1987-1992 by Logical Systems.		*/
/*									*/
/*				3/15/92					*/
/*									*/
/************************************************************************/

#include	<time.h>
#include	<fcntl.h>
#include	<conio.h>
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<io.h>
#include	<stdio.h>
#include	<errno.h>
#include	<stdlib.h>
#include	<string.h>
#include	<signal.h>

#ifdef	NV
#include	<setjmp.h>
#endif	/* NV */

#include	"taldef.h"
#include	"tload.h"
#include	"cio.h"
#include	"cioint.h"

#define	MAXFILES	100		/* Maximum # of legal open files */

#if	((MAX_MSG_DATA < MAX_APLEN) || (MAX_MSG_DATA < MAX_RPLEN))
#error	"Beware of configuring large maximum VCHAN packet sizes"
#endif	/* MAX_MSG_DATA < */

#if	((RPC_EOF >= 0) || (EOF >= 0))
#error	"RPC_EOF/EOF should be negative for correct CIO operation"
#endif	/* RPC_EOF/EOF */

/*
 *	Create POSIX O_NONBLOCK from SYSV O_NDELAY if possible.
 */
#ifndef	O_NONBLOCK
#ifdef	O_NDELAY
#define	O_NONBLOCK	O_NDELAY
#endif	/* O_NDELAY */
#endif	/* !O_NONBLOCK */

/*
 *	Configure the minimum # of per-node information structures to be
 *	allocated whenever we run out.
 */
#define	PER_NODE_ALLOCATION	10

/************************************************************************/
/*									*/
/*		Host Library Configuration Options			*/
/*									*/
/*	TRUE to enable, FALSE to disable.  Note that these options are	*/
/*	required since the host "C" environment may not conform to ANSI	*/
/*	library specifications.  In general, the two "standard"		*/
/*	libraries at the current moment are ANSI and SYS5.  Set the	*/
/*	macros to FALSE if you have SYS5 (many "C" libraries are a	*/
/*	mixture: you will probably have to configure them individually	*/
/*	in conjunction with your "C" compiler library reference		*/
/*	manual).							*/
/*									*/
/************************************************************************/

/*
 *	ANSIFSEEK should be set TRUE if the return value from "fseek" conforms
 *	to ANSI (zero if no errors, non-zero for errors).  If you set this
 *	option to FALSE the return value will be complemented before being
 *	sent to the Transputer (to simulate ANSI).
 */
#define	ANSIFSEEK	TRUE

/*
 *	ANSISYSTEM should be set to TRUE if the "system" function can be passed
 *	a NULL pointer to inquire if a command processor exists.
 */
#define	ANSISYSTEM	TRUE

/************************************************************************/
/*									*/
/*			Global Type Definitions				*/
/*									*/
/************************************************************************/

struct	blocked_io			/* Per-node blocked I/O info	*/
	{
	STREAM		fp;		/* Blocked I/O file pointer	*/
	BOOL		(*func)();	/* Blocked I/O completion func	*/
	int		index;		/* Blocked I/O strm tbl index	*/
	int		length;		/* Blocked I/O length		*/
	int		pushb;		/* Blocked I/O pushback char	*/
	};

struct	node_strm			/* Per-node, per-file structure	*/
	{
	int			gstrm;	/* Global stream table index	*/
	long			pos;	/* Previous stream position	*/
	UCHAR			pushb;	/* Pushback char from "ungetc"	*/
	UCHAR			status;	/* Stream status		*/
#define	STREAM_EOF	0x01		/* TRUE/stream @ EOF		*/
#define	STREAM_ERR	0x02		/* TRUE/stream had error	*/
#define	STREAM_PUSHB	0x04		/* TRUE/stream has "ungetc" ch	*/
	};

struct	node_info			/* Per-node structure		*/
	{
	struct	blocked_io	blkio;	/* Blocked I/O information	*/
	int			bpw;	/* Bytes per word for this node	*/
	int			nerrno;	/* "errno" for this node	*/
	UINT			next;	/* I/O request queue next node	*/
	struct	node_strm	strms[MAXFILES];	/* Node strm tbl*/
#define	STDIN_INDEX	0		/* "strms" slot for "stdin"	*/
#define	STDOUT_INDEX	1		/* "strms" slot for "stdout"	*/
#define	STDERR_INDEX	2		/* "strms" slot for "stderr"	*/
	};

struct	strm_table			/* Global per-stream structure	*/
	{
	int			cindex;	/* Current file index 'cnode'	*/
	UINT			cnode;	/* Current using node		*/
	STREAM			fp;	/* File pointer if open		*/
	STRING			name;	/* File name			*/
	int			ocnt;	/* # of times currently open	*/
	};

/************************************************************************/
/*									*/
/*			Global Data Definitions				*/
/*									*/
/************************************************************************/

static	int	argc;			/* Argument count		*/
static	char	**argv;			/* Argument vector		*/
static	int	args16_size;		/* Arg data size for 16 bit CPU	*/
static	int	args32_size;		/* Arg data size for 32 bit CPU	*/
static	UCHAR	*cmd_buf;		/* Points to cmd ("from"), buf	*/
static	UCHAR	*cmd_ptr;		/* "cmd_buf" buffer pointer	*/
static	UINT	current_chan;		/* Current destination VC	*/
static	UINT	current_node;		/* Node which we are servicing	*/

#ifndef	NV
static	UINT	exit_count;		/* # of non-"exit"ed nodes	*/
#endif	/* !NV */

static	int	exit_status;		/* "Final" exit status		*/

#ifdef	NV
static	UINT	head_irdy;		/* Blocking input request queue	*/
#endif	/* NV */

static	UINT	head_ordy;		/* I/O request queue head index	*/

#ifdef	NV
static	jmp_buf	jexit;			/* Save info for setjmp/longjmp	*/
#endif	/* NV */

static	int	LinkId;			/* LinkId from "OpenLink"	*/
static	UINT	max_nodes;		/* # of entries in "node_info"	*/
static	UINT	max_open;		/* Max "open" global streams	*/
static	struct	node_info	*nodes;	/* Per node information		*/
static	STRING	open_str;		/* Link name for "OpenLink"	*/
static	UCHAR	*rpy_buf;		/* Points to reply ("to"), buf	*/

#ifndef	NV
#ifdef	__TURBOC__
extern	unsigned	_stklen = 8192;	/* Set size of Borland stack	*/
#endif	/* __TURBOC__ */
#endif	/* !NV */

#ifdef	NV
/*
 *	The "sio" variable records whether the last logical->physical stream
 *	conversion performed by the "i_stream" function was associated with
 *	standard I/O ("stdin"/"stdout"/"stderr").
 */
static	BOOL	sio;
#endif	/* NV */

static	struct	strm_table	streams[MAXFILES];	/* Glb strm tbl	*/

#ifdef	NV
static	UINT	tail_irdy;		/* Blcking inp. req. tail index	*/
#endif	/* NV */

static	UINT	tail_ordy;		/* I/O request queue tail index	*/
static	int	timeout;		/* Current comm. timeout	*/
static	BOOL	use_vchans;		/* TRUE if using VCHAN protocol	*/
static	BOOL	used_stdin;		/* "stdin" not used yet		*/
static	BOOL	used_stdout;		/* "stdout" not used yet	*/
static	BOOL	used_stderr;		/* "stderr" not used yet	*/
static	int	vc_aplen;		/* Adj. dst. max. VCHAN packet	*/
static	int	vc_plen;		/* Current packet length	*/
static	int	vc_rplen;		/* Remote dst max VCHAN packet	*/
static	BOOL	vc_rrdy;		/* TRUE/adj routing buf in use	*/

/************************************************************************/
/*									*/
/*			Function Type Declarations			*/
/*									*/
/************************************************************************/

extern	int	_cioext();		/* In "cioext.c" file		*/

#ifndef	NV
static	VOID	banner();
#endif	/* !NV */

static	int	dec_idx();
static	int	dec_1b();
static	int	dec_2b();
static	SLONG	dec_4b();
static	VOID	enc_str();
static	VOID	enc_time();
static	VOID	enc_2b();
static	VOID	enc_4b();

#ifndef	NV
static	VOID	err_dump();
#endif	/* !NV */

static	VOID	fatal();
static	int	get_lstr();
static	VOID	get_str();
static	VOID	i_clearerr();
static	int	i_close();
static	BOOL	i_feof();
static	BOOL	i_ferror();
static	VOID	i_init();
static	int	i_pushb();
static	STREAM	i_stream();
static	int	i_ungetc();
static	VOID	i_update();
static	VOID	node_alloc();
static	VOID	ordy_add();

#ifdef	NV
static	VOID	q_add();
static	BOOL	q_fgetc();
static	BOOL	q_fgets();
static	BOOL	q_getch();
static	BOOL	q_getche();
static	VOID	q_r_2b();
static	VOID	q_r_2beof();
#endif	/* NV */

static	VOID	read_node();
static	UINT	read_vchdr();
static	VOID	reply_open();
static	VOID	reply_str();
static	VOID	reply_zeof();
static	VOID	reply_1b();
static	VOID	reply_2b();
static	VOID	reply_2beof();
static	VOID	reply_4b();
static	VOID	select_node();
static	VOID	warning();
static	VOID	write_node();
static	VOID	write_vchdr();

static	VOID	w_clearerr();
static	int	w_feof();
static	int	w_ferror();
static	int	w_fflush();
static	VOID	w_fgetc();
static	VOID	w_fgets();
static	int	w_fputc();
static	VOID	w_fread();
static	int	w_fseek();
static	long	w_ftell();
static	UINT	w_fwrite();

#if	H_OPSYS == MS_DOS
#ifdef	NV
static	VOID	w_getch();
static	VOID	w_getche();
#endif	/* NV */
#endif	/* MS_DOS */

/*
 *	The following functions may be called only for side effects.
 */
/*lint +fvr */

/*lint -fvr */

/************************************************************************/
/*									*/
/*	Routine:	main						*/
/*									*/
/*	Opens up communication with the transputer link adapter on a	*/
/*	INMOS Transputer and handles "C" file I/O requests from the	*/
/*	connected transputer.						*/
/*									*/
/*	Parameters:							*/
/*		"argc"	An integer count of the command line arguments.	*/
/*		"argv"	A pointer to an array of pointers which point	*/
/*			to individual command line arguments.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

#ifndef	NV
VOID
main(real_argc,real_argv)
	int			real_argc;
	char			*real_argv[];
	{
	(VOID) _cio_open(real_argc,real_argv);
	while(_cio_execute())
		;
	(VOID) _cio_close();
	exit(exit_status);		/* Head for the barn */
	}
#endif	/* !NV */

/************************************************************************/
/*									*/
/*	Routine:	banner						*/
/*									*/
/*	Display sign-on message to user.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

#ifndef	NV

static	VOID
banner()
	{
	fprintf(stderr,
		"CIO ('C' I/O Server), %s [Link I/O Driver: '%s']\n",
			VERSION,IODRIVER);
	fprintf(stderr,"%s\n\n",COPYRIGHT);
	}
#endif	/* !NV */

/************************************************************************/
/*									*/
/*	Routine:	_cio_close					*/
/*									*/
/*	Close CIO server.						*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	Non-zero to indicate a fatal error was		*/
/*			encountered when linked with NV.		*/
/*									*/
/************************************************************************/

BOOL
_cio_close()
	{
#ifdef	NV
	int			i;
/*
 *	Register setjmp/longjmp information for possible fatal error return.
 */
	if(setjmp(jexit) != 0)
		return (ERRORS);
/*
 *	Close all open files (except STDIN/STDOUT/STDERR).
 */
	for(i = 3; i < MAXFILES; i++)
		{
		if(streams[i].fp != NULL)
			(VOID) fclose(streams[i].fp);
/*
 *	Return possible free storage pointed to by "name" field.
 */
		if(streams[i].name != NULL)
			{
			free(streams[i].name);
			streams[i].name = NULL;
			}
		}
/*
 *	Free memory associated with the "node_info" structure.
 */
	if(nodes != NULL)
		{
		free(nodes);
		nodes = NULL;
		}
/*
 *	Free memory associated with the send/receive message buffers.
 */
	if(cmd_buf != NULL)
		{
		free(cmd_buf);
		cmd_buf = NULL;
		}
	if(rpy_buf != NULL)
		{
		free(rpy_buf);
		rpy_buf = NULL;
		}
#endif	/* NV */
/*
 *	Close the Transputer link.
 */
	if(LinkId > 0)			/* Only needed if link was opened */
		{
		if(CloseLink(LinkId) < 0)
			{
			LinkId = 0;
			fatal(FALSE,"Unable to close link: ",open_str);
			}
		LinkId = 0;
		}
	return (NOERRORS);
	}

/************************************************************************/
/*									*/
/*	Routine:	_cio_execute					*/
/*									*/
/*	Execute a single command to or from the Transputer network.  If	*/
/*	CIO is being used with NodeView this function may optionally be	*/
/*	called with a debugging message to be sent to a specific node. 	*/
/*									*/
/*	Parameters:							*/
/*		"node"	An optional destination for use with NV.	*/
/*		"tout"	An optional "timeout" setting for use with NV.	*/
/*		"func"	An optional function code for use with NV.	*/
/*		"dlen"	An optional length for the "data" parameter.	*/
/*		"dadr"	An optional destination node address.		*/
/*		"data"	An optional data src/dst for use with NV.	*/
/*									*/
/*	Returns:	If running standalone: FALSE if time to exit.	*/
/*			If running as part of NV, non-zero as fatal	*/
/*			error flag.					*/
/*									*/
/************************************************************************/

#ifdef	MSC_OPT
#pragma	optimize("e",off)		/* Disable global reg optimize	*/
#pragma	optimize("g",off)		/* Disable global common subs	*/
#pragma	optimize("l",off)		/* Disable loop optimizations	*/
#endif	/* MSC_OPT */

BOOL
#ifdef	NV
_cio_execute(node,tout,func,dlen,dadr,data)
	UINT			node;
	int			tout;
	UINT			func;
	UINT			dlen;
	SLONG			dadr;
	char			*data;
	{
	SLONG			debug_reg[14];
	UINT			*q_ptr;
#else	/* NV */
_cio_execute()
	{
#endif	/* NV */
	char			numbuf[20];
	UCHAR			*rpy_ptr;
	int			t_int1;
	int			t_int2;
	int			t_int3;
	SLONG			t_long1;
	STRING			t_str;
	time_t			t_time;
	struct	tm		*t_tm;
#ifdef	NV
/*
 *	If NodeView doesn't wish to send any debug messages to the Transputer
 *	network, and we have no pending I/O requests from the network, and
 *	no I/O requests waiting, simply return.
 */
	if(node == 0)			/* NodeView has a request? */
		{
		if(use_vchans)		/* No, are we using vchans? */
			{
/*
 *	Since we are using virtual channels deciding whether we have anything
 *	to do is a bit complex since we must handle the various protocol
 *	header messages.
 */
			if(head_ordy == 0)	/* Anything waiting? */
				{
/*
 *	No I/O requests have been previously noted.  See if any new ones are
 *	pending.  Note that "read_vchdr" sets the I/O request queue head
 *	pointer ("head_ordy"), as a side effect if it processes an I/O
 *	request.  This is in addition to the "_VCH_CNTL | _VCH_ORDY" return
 *	value...
 */
				current_node = 0;	/* Anybody is OK */
				while(TestRead(LinkId))	/* Any input bytes? */
					{
					if(read_vchdr() ==
						(_VCH_CNTL | _VCH_ORDY))
						{
						break;	/* Got an I/O req */
						}
					}
				}
			if((head_ordy == 0) && (head_irdy == 0))
				return (NOERRORS);	/* Nothing to do */
			}
		else
			{
/*
 *	Since we aren't using virtual channels we can tell whether we have
 *	something to do simply by checking if an input byte is waiting on the
 *	link.
 */
			if(! TestRead(LinkId))
				return (NOERRORS);
			}
		}
/*
 *	We evidently have something to do, register setjmp/longjmp information
 *	for possible fatal error return.
 */
	if(setjmp(jexit) != 0)
		return (ERRORS);
/*
 *	Copy "tout" setting into "timeout" global variable.  This lets the user
 *	specify the desired communication timeout to use at runtime.
 */
	timeout = tout;
#endif	/* NV */
/*
 *	Record the value of "errno" in the appropriate node info structure.
 */
	if(errno != 0)
		{
 		nodes[current_node - 1].nerrno = errno;
		errno = 0;
		}
	current_chan = VC_HOSTIO;	/* Assume regular I/O channel */
#ifdef	NV
	if(node != 0)
		{
/*
 *	Handle NV debug requests to Transputer network.
 */
		if(! use_vchans)
			{
	fatal(FALSE,"NV debug command prior to network being ready","");
			}
		else
			{
			current_node = node;	/* Node to send debug msg to */
			current_chan = VC_DEBUGIO;	/* Use debug channel */
/*
 *	Read dummy one byte message from debugging channel on node being
 *	debugged to get things synchronized.
 */
			read_node(cmd_buf,1);
/*
 *	Send debugging request message to node.
 */
			cmd_ptr = cmd_buf;
			*cmd_ptr++ = (UCHAR) func;
			enc_4b(&cmd_ptr,(SLONG) dlen);
			enc_4b(&cmd_ptr,(SLONG) dadr);
			reply_str(cmd_buf,9);
/*
 *	Perform request-specific processing.
 */
			switch(func)
				{
				case	NV_RMEM:
/*
 *	Read data from the memory of the specified Transputer node.
 */
					if(dlen > MAX_MSG_DATA)
						{
						fatal(FALSE,"NV read too long",
							"");
						}
					get_str((UCHAR *) data,(int) dlen);
					break;

				case	NV_WMEM:
/*
 *	Write data into the memory of the specified Transputer node.
 */
					if(dlen > MAX_MSG_DATA)
						{
						fatal(FALSE,
							"NV write too long",
								"");
						}
					reply_str((UCHAR *) data,(int) dlen);
					break;

				case	NV_GBRKPT:
/*
 *	Restart process after hitting a breakpoint.
 */
					cmd_ptr = cmd_buf;
					for(t_int1 = 0; t_int1 < _DEBUG_REGS;
						t_int1++)
						{
/* CPU state to restore */			enc_4b(&cmd_ptr,((SLONG *)
							data)[t_int1]);
						}
					reply_str(cmd_buf,_DEBUG_REGS * 4);
					break;

				case	NV_PMSTART:
/*
 *	Enable performance monitoring.
 */
					get_str((UCHAR *) data,2 * 4);
					break;

				case	NV_PMPOLL:
/*
 *	Poll node for performance data.
 */
					get_str((UCHAR *) data,
						(MAX_LINKS + 2) * 4);
					break;

				default:
					fatal(FALSE,
						"Unknown NV debug command","");
					break;	/* Unnecessary */
				}
			return (NOERRORS);
			}
		}
/*
 *	Poll NV to see if any blocked I/O requests are now ready.
 */
	if(head_irdy != 0)
		{
		for(q_ptr = &head_irdy; *q_ptr != 0;
			q_ptr = &nodes[*q_ptr - 1].next)
			{
			if((*nodes[*q_ptr - 1].blkio.func)(*q_ptr))
				{
/*
 *	We have successfully completed a blocked I/O request, delete it from
 *	the queue and return.
 */
				*q_ptr = nodes[*q_ptr - 1].next;
				return (NOERRORS);
				}
			}
		if(head_ordy == 0)
			return (NOERRORS);
		}
#endif	/* NV */
/*
 *	Handle server requests from Transputer.  Header for server request
 *	consists of a one byte function code and a one byte length for the
 *	immediately following data.
 */
	if(use_vchans)			/* Choose node to service */
		select_node();

	read_node(cmd_buf,2);

	t_int1 = cmd_buf[0];		/* Function code */
	get_str(cmd_buf,cmd_buf[1]);	/* Get rest of command */
	cmd_ptr = cmd_buf;
	switch(t_int1)
		{
		case	FN_CIOEXT:
			t_int1 = dec_2b(&cmd_ptr);
			t_int2 = dec_2b(&cmd_ptr);
			t_int3 = dec_2b(&cmd_ptr);
			get_str(cmd_buf,t_int2);
			t_int1 = _cioext(t_int1,cmd_buf,t_int2,rpy_buf,t_int3);
			cmd_ptr = cmd_buf;
			enc_4b(&cmd_ptr,(SLONG) t_int1);
/*
 *	If errno is non-zero, signal the fact and zero it back out.
 */
			if(errno != 0)
				{
				t_int1 = errno;
				errno = 0;
				}
			else
				t_int1 = 0;

			enc_4b(&cmd_ptr,(SLONG) t_int1);
			reply_str(cmd_buf,8);
			reply_str(rpy_buf,t_int3);
			break;

		case	FN_CLEARERR:
			i_clearerr(dec_idx(&cmd_ptr));
			break;

		case	FN_CLOSE:
			reply_zeof(close(dec_2b(&cmd_ptr)));
			break;

		case	FN_CREAT:
/*
 *	Read and translate "mode".
 */
			t_int1 = dec_2b(&cmd_ptr);
			t_int2 = 0;
			if(_S_IREAD & t_int1)
				t_int2 |= S_IREAD;
			if(_S_IWRITE & t_int1)
				t_int2 |= S_IWRITE;
			(VOID) get_lstr();
			reply_2beof(creat((STRING) cmd_buf,t_int2));
			break;

		case	FN_DUP:
			reply_2beof(dup(dec_2b(&cmd_ptr)));
			break;

		case	FN_DUP2:
			t_int1 = (int) dec_2b(&cmd_ptr);
			reply_zeof(dup2(t_int1,dec_2b(&cmd_ptr)));
			break;

		case	FN_ERRNO:
			reply_4b((SLONG) nodes[current_node - 1].nerrno);
/*
 *	Zot node copy of "errno".
 */
			nodes[current_node - 1].nerrno = 0;
			break;

		case	FN_FCLOSE:
			reply_1b(i_close(dec_idx(&cmd_ptr)));
			break;

		case	FN_FCLOSEALL:
			t_int1 = 0;
/*
 *	Skip STDIN/STDOUT/STDERR.
 */
			for(t_int2 = 3;	t_int2 < MAXFILES; t_int2++)
				{
/* If file open */		if(nodes[current_node - 1].strms[t_int2].
					gstrm > 0)
					{
/* Close it */				if(i_close(t_int2) == RPC_EOF)
/* Record that we had an error */		t_int1 = RPC_EOF;
					else if(t_int1 != RPC_EOF)
/* Keep track of how many we closed */		t_int1++;
					}
				}
			reply_2b(t_int1);
			break;
						
		case	FN_FDOPEN:
			t_int1 = (int) dec_2b(&cmd_ptr);
			(VOID) get_lstr();
			reply_open(fdopen(t_int1,(STRING)  cmd_buf),-1,NULL,0,
				NULL);
			break;

		case	FN_FEOF:
			reply_zeof(i_feof(dec_idx(&cmd_ptr)));
			break;

		case	FN_FERROR:
			reply_zeof(i_ferror(dec_idx(&cmd_ptr)));
			break;

		case	FN_FFLUSH:
			if((t_int1 = dec_idx(&cmd_ptr)) < 0)
				{
				reply_1b(RPC_EOF);
				break;
				}
			reply_zeof(w_fflush(i_stream(t_int1)));
			i_update(t_int1);
			break;

		case	FN_FGETC:
			if((t_int1 = dec_idx(&cmd_ptr)) < 0)
				{
				reply_2b(RPC_EOF);
				break;
				}
			if((t_int2 = i_pushb(t_int1)) >= 0)
				reply_2b(t_int2);
			else
				w_fgetc(t_int1);
			break;

		case	FN_FGETS:
			if((t_int1 = dec_idx(&cmd_ptr)) < 0)
				{
/* Bad stream index */		reply_2b(RPC_NULL);
				break;
				}
			t_int2 = dec_2b(&cmd_ptr);
/*
 *	If we have a pushback character from a previous "ungetc" call, and
 *	it was either a newline, or the user has requested a 2 character max
 *	line, just pass back the pushback character (made into a string).
 */
			if((t_int3 = i_pushb(t_int1)) >= 0)
				{
				rpy_buf[0] = (UCHAR) t_int3;
				rpy_buf[1] = '\0';
				if((t_int2 == 2) || (t_int3 == '\n'))
					{
					reply_2b(2);
					reply_str(rpy_buf,2);
					break;
					}
				t_int3 = 1;
				}
			else
				t_int3 = 0;
/*
 *	Either we don't have a pushback character, or we need to call "fgets"
 *	to get additional stuff.
 */
			w_fgets(t_int1,t_int2,t_int3);
			break;

		case	FN_FILENO:
			if((t_int1 = dec_idx(&cmd_ptr)) < 0)
				{
				reply_2b(RPC_EOF);
				break;
				}
			reply_2beof(fileno(i_stream(t_int1)));
			i_update(t_int1);
			break;

		case	FN_FOPEN:
			t_int1 = dec_2b(&cmd_ptr);
			t_int2 = dec_2b(&cmd_ptr);
			get_str(cmd_buf,t_int1);
			get_str(rpy_buf,t_int2);
			reply_open(NULL,-1,(STRING) cmd_buf,t_int1,
				(STRING) rpy_buf);
			break;

		case	FN_FPUTC:
			if((t_int1 = dec_idx(&cmd_ptr)) < 0)
				{
				reply_2b(RPC_EOF);
				break;
				}
			reply_2beof(w_fputc(dec_1b(&cmd_ptr),
				i_stream(t_int1)));
			i_update(t_int1);
			break;

		case	FN_FREAD:
			if((t_int1 = dec_idx(&cmd_ptr)) < 0)
				{
				reply_2b(0);
				break;
				}
/* Desired read len */	t_int2 = dec_2b(&cmd_ptr) * dec_2b(&cmd_ptr);
/*
 *	If we have a pushback character from a previous "ungetc" call, and
 *	the user has requested a 1 character read, just pass back the pushback
 *	character.
 */
			if((t_int3 = i_pushb(t_int1)) >= 0)
				{
				rpy_buf[0] = (UCHAR) t_int3;
				if(t_int2 == 1)
					{
					reply_2b(1);
					reply_str(rpy_buf,1);
					break;
					}
				t_int3 = 1;
				}
			else
				t_int3 = 0;
/*
 *	Either we don't have a pushback character, or we need to call "fread"
 *	to get additional stuff.
 */
			w_fread(t_int1,t_int2,t_int3);
			break;

		case	FN_FREOPEN:
			t_int1 = dec_2b(&cmd_ptr);
			t_int2 = dec_2b(&cmd_ptr);
			t_int3 = dec_2b(&cmd_ptr);
			get_str(cmd_buf,t_int2);
			get_str(rpy_buf,t_int3);
			if((t_int1 <= 0) || (t_int1 > MAXFILES))
				{
				reply_2b(RPC_NULL);
				break;
				}
			(VOID) i_close(t_int1 - 1);
			reply_open(NULL,t_int1 - 1,(STRING) cmd_buf,t_int2,
				(STRING) rpy_buf);
			break;

		case	FN_FSEEK:
			if((t_int1 = dec_idx(&cmd_ptr)) < 0)
				{
				reply_1b(RPC_EOF);
				break;
				}
/*
 *	Clear ERR and EOF flags and flush potential "pushback" char.
 */
			i_init(t_int1);

			t_long1 = dec_4b(&cmd_ptr);
			switch(t_int2 = dec_1b(&cmd_ptr))
				{
				case	_SEEK_CUR:
					t_int2 = SEEK_CUR;
					break;

				case	_SEEK_END:
					t_int2 = SEEK_END;
					break;

				case	_SEEK_SET:
					t_int2 = SEEK_SET;
					break;

				default:
					sprintf(numbuf,"0x%X",t_int2);
			warning("'fseek' origin value undefined: ",numbuf);
					t_int2 = SEEK_CUR;
					break;
				}
#if	ANSIFSEEK
			reply_zeof(
#else	/* !ANSIFSEEK */
			reply_zeof(!
#endif	/* ANSIFSEEK */
					w_fseek(i_stream(t_int1),t_long1,
						t_int2));
			i_update(t_int1);
			break;

		case	FN_FTELL:
			if((t_int1 = dec_idx(&cmd_ptr)) < 0)
				{
				reply_4b((SLONG) RPC_EOF);
				break;
				}
			t_long1 = w_ftell(i_stream(t_int1));
			if(t_long1 == EOF)
				reply_4b((SLONG) RPC_EOF);
			else
				reply_4b(t_long1);
			i_update(t_int1);
			break;

		case	FN_FWRITE:
			t_int1 = dec_idx(&cmd_ptr);
			t_int2 = (int) dec_2b(&cmd_ptr);
			t_int3 = (int) dec_2b(&cmd_ptr);
			get_str(cmd_buf,t_int2 * t_int3);
			if(t_int1 < 0)
				{
				reply_2b(0);
				break;
				}
			reply_2b((int) w_fwrite((STRING) cmd_buf,(UINT) t_int2,
				(UINT) t_int3,i_stream(t_int1)));
			i_update(t_int1);
			break;
#if	H_OPSYS == MS_DOS
		case	FN_GETCH:
#ifdef	NV
			w_getch();
#else	/* NV */
			reply_2b(getch());
#endif	/* NV */
			break;

		case	FN_GETCHE:
#ifdef	NV
			w_getche();
#else	/* NV */
			reply_2b(getche());
#endif	/* NV */
			break;
#endif	/* MS_DOS */					
		case	FN_GETENV:
			(VOID) get_lstr();
			if((t_str = getenv((STRING) cmd_buf)) == NULL)
				reply_2b(0);
			else
				{
				t_int1 = strlen(t_str) + 1;
				reply_2b(t_int1);
				get_str(cmd_buf,1);
				if(cmd_buf[0] != RPC_NULL)
					reply_str((UCHAR *) t_str,t_int1);
				else
					{
warning("Client processor unable to malloc storage for 'getenv' call","");
					}
				}
			break;
#if	H_OPSYS == MS_DOS
		case	FN_KBHIT:
#ifdef	NV
			reply_2b(_nv_kbhit(current_node));
#else	/* NV */
			reply_2b(kbhit());
#endif	/* NV */
			break;
#endif	/* MS_DOS */					
		case	FN_LSEEK:
			t_int1 = (int) dec_2b(&cmd_ptr);
			t_long1 = dec_4b(&cmd_ptr);
			switch(t_int2 = dec_1b(&cmd_ptr))
				{
				case	_SEEK_CUR:
					t_int2 = SEEK_CUR;
					break;

				case	_SEEK_END:
					t_int2 = SEEK_END;
					break;

				case	_SEEK_SET:
					t_int2 = SEEK_SET;
					break;

				default:
					sprintf(numbuf,"0x%X",t_int2);
			warning("'lseek' origin value undefined: ",numbuf);
					t_int2 = SEEK_CUR;
					break;
				}
			t_long1 = lseek(t_int1,t_long1,t_int2);
			if(t_long1 == EOF)
				reply_4b((SLONG) RPC_EOF);
			else
				reply_4b(t_long1);
			break;

		case	FN_OPEN:
/*
 *	Read and translate "access" method.
 */
			t_int1 = dec_2b(&cmd_ptr);
			t_int2 = 0;
			if(_O_APPEND & t_int1)
				t_int2 |= O_APPEND;
#ifdef	O_BINARY
			if(_O_BINARY & t_int1)
				t_int2 |= O_BINARY;
#endif	/* O_BINARY */
			if(_O_CREAT & t_int1)
				t_int2 |= O_CREAT;
			if(_O_EXCL & t_int1)
				t_int2 |= O_EXCL;
#ifdef	O_NONBLOCK
			if(_O_NONBLOCK & t_int1)
				t_int2 |= O_NONBLOCK;
#endif	/* O_NONBLOCK */
			if(_O_RDONLY & t_int1)
				t_int2 |= O_RDONLY;
			if(_O_RDWR & t_int1)
				t_int2 |= O_RDWR;
#ifdef	O_TEXT
			if(_O_TEXT & t_int1)
				t_int2 |= O_TEXT;
#endif	/* O_TEXT */
			if(_O_TRUNC & t_int1)
				t_int2 |= O_TRUNC;
			if(_O_WRONLY & t_int1)
				t_int2 |= O_WRONLY;
/*
 *	Read and translate "mode".
 */
			t_int1 = dec_2b(&cmd_ptr);
			t_int3 = 0;
			if(_S_IREAD & t_int1)
				t_int3 |= S_IREAD;
			if(_S_IWRITE & t_int1)
				t_int3 |= S_IWRITE;

			(VOID) get_lstr();

			reply_2beof(open((STRING) cmd_buf,t_int2,t_int3));
			break;

		case	FN_READ:
			t_int1 = (int) dec_2b(&cmd_ptr);
			t_int1 = read(t_int1,(STRING) rpy_buf,
				(UINT) dec_2b(&cmd_ptr));
			reply_2beof(t_int1);
			reply_str(rpy_buf,t_int1);
			break;

		case	FN_REMOVE:
			(VOID) get_lstr();
			reply_zeof(remove((STRING) cmd_buf));
			break;

		case	FN_RENAME:
			t_int1 = dec_2b(&cmd_ptr);
			t_int2 = dec_2b(&cmd_ptr);
			get_str(cmd_buf,t_int1);
			get_str(rpy_buf,t_int2);
			reply_zeof(rename((STRING) cmd_buf,(STRING) rpy_buf));
			break;

		case	FN_SYSTEM:
			if((t_int1 = dec_2b(&cmd_ptr)) == 0)
				{
#if	!ANSISYSTEM
warning("Unable to pass NULL pointer to 'system' in non-ANSI library","");
				t_str = "";
#else	/* ANSISYSTEM */
				t_str = NULL;
#endif	/* !ANSISYSTEM */
				}
			else
				{
				get_str(cmd_buf,t_int1);
				t_str = (STRING) cmd_buf;
				}
			reply_4b((SLONG) system(t_str));
			break;

		case	FN_TIME:
			reply_4b((SLONG) time(NULL));
			break;
/*
 *	Note if GMT/UTC time is not available on a particular host platform
 *	the following code can be changed to unconditionally do the "reply_1b"
 *	of RPC_NULL.
 */
		case	FN_TIMEGMT:
			t_time = (time_t) dec_4b(&cmd_ptr);
			if((t_tm = gmtime(&t_time)) == NULL)
				reply_1b(RPC_NULL);
			else
				{
				reply_1b(RPC_NOTNULL);
				rpy_ptr = rpy_buf;
				enc_time(&rpy_ptr,t_tm);
				reply_str(rpy_buf,RPC_TIME_DATA);
				}
			break;

		case	FN_TIMELOC:
			t_time = (time_t) dec_4b(&cmd_ptr);
			rpy_ptr = rpy_buf;
			enc_time(&rpy_ptr,localtime(&t_time));
			reply_str(rpy_buf,RPC_TIME_DATA);
			break;

		case	FN_TMPFILE:
			reply_open(tmpfile(),-1,NULL,0,NULL);
			break;

		case	FN_TMPNAM:
			if((tmpnam((STRING) rpy_buf)) == NULL)
				reply_2b(0);
			else
				{
				reply_2b(t_int1 = strlen((STRING)rpy_buf) + 1);
				get_str(cmd_buf,1);
				if(cmd_buf[0] != RPC_NULL)
					reply_str(rpy_buf,t_int1);
				else
					{
warning("Client processor unable to malloc storage for 'tmpnam' call","");
					}
				}
			break;

		case	FN_UNGETC:
			t_int1 = dec_idx(&cmd_ptr);
			reply_2b(i_ungetc(dec_2b(&cmd_ptr),t_int1));
			break;

		case	FN_WRITE:
			t_int1 = (int) dec_2b(&cmd_ptr);
			t_int2 = get_lstr();
			reply_2beof(write(t_int1,(STRING) cmd_buf,
				(UINT) t_int2));
			break;

/************************************************************************/
/*									*/
/*			Low usage rate I/O routines.			*/
/*									*/
/************************************************************************/

#ifdef	NV
		case FN_BRKPT:
/*
 *	Tell NV we hit a breakpoint and pass along the applicable node status.
 */
			for(t_int1 = 0; t_int1 < _DEBUG_REGS; t_int1++)
				debug_reg[t_int1] = dec_4b(&cmd_ptr);
			_nv_brkpt(current_node,debug_reg);
			break;
#endif	/* NV */

		case FN_ARG_GET:
/*
 *	Send argc/argv info and strings to client processor.
 */
/* Start arg space */	t_long1 = dec_4b(&cmd_ptr) + ((argc + 1L) *
				((SLONG) nodes[current_node - 1].bpw));
/*
 *	Send "argc" value and the argument data size to client.
 */
			rpy_ptr = rpy_buf;
			enc_2b(&rpy_ptr,(int) argc);
			if(nodes[current_node - 1].bpw == 2)
				enc_2b(&rpy_ptr,(t_int1 = args16_size));
			else
				enc_2b(&rpy_ptr,(t_int1 = args32_size));
			reply_str(rpy_buf,4);
/*
 *	Fill in "argv" table.
 */
			rpy_ptr = rpy_buf;
			for(t_int2 = 0; t_int2 < argc; t_int2++)
				{
				if(nodes[current_node - 1].bpw == 2)
					enc_2b(&rpy_ptr, (int) t_long1);
				else
					enc_4b(&rpy_ptr,t_long1);

				t_long1 += (strlen(argv[t_int2]) + 1);
				}
			if(nodes[current_node - 1].bpw == 2)
				enc_2b(&rpy_ptr,0);
			else
				enc_4b(&rpy_ptr,0L);
/*
 *	Now add argument strings.
 */
			for(t_int2 = 0; t_int2 < argc; t_int2++)
				enc_str(&rpy_ptr,argv[t_int2]);
/*
 *	Send the argument data.
 */
			reply_str(rpy_buf,t_int1);
			break;

		case FN_ARG_SIZE:
#ifdef	NV
			if(! use_vchans)
				{
	fatal(FALSE,"Root node not configured to use virtual channels","");
				}
#endif	/* NV */
/*
 *	Record the breakpoint entry function address for this node.
 */
			t_long1 = dec_4b(&cmd_ptr);
/*
 *	Tell requesting node how many bytes (rounded up to nearest word), of
 *	workspace will be required to hold the "argc"/"argv" information and
 *	the actual argument strings.
 */
			if((t_int1 = dec_2b(&cmd_ptr)) == 2)
				reply_2b(args16_size);
			else if(t_int1 == 4)
				reply_2b(args32_size);
			else
				{
			fatal(TRUE,"Unexpected target system word length","");
				}
			nodes[current_node - 1].bpw = t_int1;
/*
 *	Since this is the first command that a node sends, perform the node
 *	specific information initialization here.
 */
#ifndef	NV
/* # of active nodes */	exit_count++;
#endif	/* !NV */

/* Init node "errno" */	nodes[current_node - 1].nerrno = 0;
/*
 *	Initialize the per-node stream table.
 */
			for(t_int2 = 0; t_int2 < MAXFILES; t_int2++)
				{
/* File isn't open */		nodes[current_node-1].strms[t_int2].gstrm = 0;
/* Zero file position */	nodes[current_node - 1].strms[t_int2].pos = 0;
/* Clear stream status */	i_init(t_int2);
				}
/* Open 'stdin' */	nodes[current_node - 1].strms[STDIN_INDEX].gstrm =
				STDIN_INDEX + 1;
/* Open 'stdout' */	nodes[current_node - 1].strms[STDOUT_INDEX].gstrm =
				STDOUT_INDEX + 1;
/* Open 'stderr' */	nodes[current_node - 1].strms[STDERR_INDEX].gstrm =
				STDERR_INDEX + 1;
#ifdef	NV
/*
 *	If we are running as part of NodeView, then we need to indicate that
 *	the specified node is up and running and may be issued debugging
 *	messages.  Also, tell the debugger the node word size and the
 *	breakpoint function entry address.
 */
			_nv_nstart(current_node,(UINT) t_int1,(SLONG) t_long1);
#endif	/* NV */
			break;

		case	FN_EXIT:
/*
 *	Record the first non-zero exit status we get (if any).
 */
			t_int1 = (int) dec_4b(&cmd_ptr);
			if(exit_status == 0)
				exit_status = t_int1;
/*
 *	If all nodes which have used CIO have also called "exit", quit.  If
 *	we are running as part of NV then report back that the specified node
 *	has terminated.
 */
#ifdef	NV
			_nv_nexit(current_node,t_int1);
			break;
#else	/* NV */
			if(--exit_count == 0)
				return (FALSE);	/* All done */
			else
				break;
#endif	/* NV */

		case	FN_VCHAN_MODE:
			vc_aplen = (int) dec_2b(&cmd_ptr);
			vc_rplen = (int) dec_2b(&cmd_ptr);
			use_vchans = TRUE;
			break;

		default:
			cmd_ptr = cmd_buf;
			sprintf(numbuf,"0x%X",dec_1b(&cmd_ptr));
			fatal(TRUE,"Invalid function code: ",numbuf);
			break;	/* Not really needed */
		}
#ifdef	NV
	return (NOERRORS);
#else	/* NV */
	return (TRUE);
#endif	/* NV */
	}

#ifdef	MSC_OPT
#pragma	optimize("e",on)		/* Enable global reg optimizing */
#pragma	optimize("g",on)		/* Enable global common subs */
#pragma	optimize("l",on)		/* Enable loop optimizing */
#endif	/* MSC_OPT */

/************************************************************************/
/*									*/
/*	Routine:	_cio_open					*/
/*									*/
/*	Initialize CIO server for operation.				*/
/*									*/
/*	Parameters:							*/
/*		"argc"	An integer count of the command line arguments.	*/
/*		"argv"	A pointer to an array of pointers which point	*/
/*			to individual command line arguments.		*/
/*									*/
/*	Returns:	Non-zero to indicate a fatal error was		*/
/*			encountered when linked with NV.		*/
/*									*/
/************************************************************************/

BOOL
_cio_open(real_argc,real_argv)
	int			real_argc;
	char			*real_argv[];
	{
	int			i;

	argc = real_argc;
	argv = real_argv;
/*
 *	Initialize global variables.
 */
	cmd_buf = NULL;			/* No initial command buffer	*/

#ifndef	NV
	exit_count = 0;			/* # of non-"exit"ed nodes	*/
#endif	/* !NV */

	exit_status = 0;		/* Default to "OK" exit		*/

#ifdef	NV
	head_irdy = 0;			/* Blcking input Q head index	*/
#endif	/* NV */

	head_ordy = 0;			/* I/O request queue head index	*/
	LinkId = 0;			/* No link open initially	*/
	max_nodes = 0;			/* # of entries in "node_info"	*/
	nodes = NULL;			/* Per node information		*/
	rpy_buf = NULL;			/* No initial reply buffer	*/
	timeout = INFINITE_TIMEOUT;	/* Default to infinite time out	*/
	use_vchans = FALSE;		/* TRUE if using VCHAN protocol	*/
	used_stdin = FALSE;		/* "stdin" not used yet		*/
	used_stdout = FALSE;		/* "stdout" not used yet	*/
	used_stderr = FALSE;		/* "stderr" not used yet	*/
	vc_rrdy = FALSE;		/* TRUE/adj routing buf in use	*/
/*
 *	Initialize global stream table.
 */
	for(i = 0; i < MAXFILES; i++)
		{
		streams[i].fp = NULL;
		streams[i].name = NULL;
		streams[i].ocnt = 0;
		}
/*
 *	Initialize the global versions of stdin/stdout/stderr.  Note that
 *	since we initialize the "ocnt" field to -1, there is no way that these
 *	files will ever be "globally" closed.  From a per-node perspective of
 *	course, they can be...
 */
	streams[STDIN_INDEX].fp = stdin;
	streams[STDIN_INDEX].name = NULL;
	streams[STDIN_INDEX].ocnt = -1;

	streams[STDOUT_INDEX].fp = stdout;
	streams[STDOUT_INDEX].name = NULL;
	streams[STDOUT_INDEX].ocnt = -1;

	streams[STDERR_INDEX].fp = stderr;
	streams[STDERR_INDEX].name = NULL;
	streams[STDERR_INDEX].ocnt = -1;

	max_open = 3;			/* We've opened stdin/stdout/stderr */
#ifdef	NV
/*
 *	Register setjmp/longjmp information for possible fatal error return.
 */
	if(setjmp(jexit) != 0)
		return (ERRORS);
#endif	/* NV */
/*
 *	Allocate memory for the send and reply message buffers.
 */
	if((cmd_buf = (UCHAR *) malloc(MAX_MSG_DATA + MAX_MSG_OVER)) == NULL)
		fatal(FALSE,"Insufficient command buffer memory","");
	if((rpy_buf = (UCHAR *) malloc(MAX_MSG_DATA + MAX_MSG_OVER)) == NULL)
		fatal(FALSE,"Insufficient reply buffer memory","");
/*
 *	Allocate the initial "node_info" structure for the root node.
 */
	node_alloc(current_node = 1);
/*
 *	Print sign-on message and open link to Transputer.
 */
#ifndef	NV
	banner();
#endif	/* !NV */
	open_str = getenv(LINK_NAME);	/* Get desired link name if present */
	LinkId = OpenLink(open_str);	/* Open specified link channel */
	if(open_str == NULL)
		open_str = "<default>";
	if(LinkId <= 0)
		fatal(FALSE,"Unable to open link: ",open_str);
/*
 *	Compute amount of Transputer workspace required to hold the "argv"
 *	array and the arguments themselves.  This data may be requested by the
 *	Transputer if desired to initialize the "argc"/"argv" values.  Note
 *	that we actually compute it two ways for either a 16 bit or 32 bit
 *	processor.
 *
 */
	for(args16_size = 0, i = 0; i < argc; i++)
		args16_size += (strlen(argv[i]) + 1);
	args32_size = args16_size;
	args16_size += ((argc + 1) * 2);
	if(args16_size & (2 - 1))
		args16_size = (args16_size & (~(2 - 1))) + 2;
	args32_size += ((argc + 1) * 4);
	if(args32_size & (4 - 1))
		args32_size = (args32_size & (~(4 - 1))) + 4;

	return (NOERRORS);
	}

/************************************************************************/
/*									*/
/*	Routine:	dec_idx						*/
/*									*/
/*	Read a two byte integer from a buffer and translate it into a	*/
/*	stream number taken from the stream table.  Note the table	*/
/*	index has been offset by 1 to allow the RPC_NULL return value	*/
/*	to be sent without conflicting with a valid stream table index.	*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to pointer to UCHAR (the buffer).	*/
/*									*/
/*	Returns:	The per-node stream table index (-1 if error).	*/
/*									*/
/************************************************************************/

static	int
dec_idx(sptr)
	UCHAR			**sptr;
	{
	int			index;
	char			numbuf[20];
/*
 *	Translate the stream index into a table offset for the per-node stream
 *	table.
 */
	index = dec_2b(sptr) - 1;	/* Stream table index */
	if((index < 0) || (index >= MAXFILES))
		{
		sprintf(numbuf,"%d",index);
		warning("Stream number is out of range: ",numbuf);
		return (-1);		/* Error */
		}
/*
 *	Verify that the stream is open.
 */
	if((nodes[current_node - 1].strms[index].gstrm - 1) < 0)
		{
		sprintf(numbuf,"%d",index);
		warning("Operation with a closed file/stream not in use: ",
			numbuf);
		return (-1);
		}
	return (index);
	}

/************************************************************************/
/*									*/
/*	Routine:	dec_1b, dec_2b, dec_4b				*/
/*									*/
/*	Read a one, two, or four byte signed value from the specified	*/
/*	buffer and return the resulting value.  The buffer pointer is	*/
/*	advanced past the data which was read.				*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to pointer to UCHAR (the buffer).	*/
/*									*/
/*	Returns:	The resulting value.				*/
/*									*/
/************************************************************************/

static	int
dec_1b(sptr)
	UCHAR			**sptr;
	{
	return ((int) *(*sptr)++);
	}

static	int
dec_2b(sptr)
	UCHAR			**sptr;
	{
	int			rval;

	rval = *(*sptr)++;
	rval |= (*(*sptr)++ << 8);
	return (rval);
	}

static	SLONG
dec_4b(sptr)
	UCHAR			**sptr;
	{
	SLONG			rval;

	rval = (SLONG) *(*sptr)++;
	rval |= (((SLONG) *(*sptr)++) << 8);
	rval |= (((SLONG) *(*sptr)++) << 16);
	rval |= (((SLONG) *(*sptr)++) << 24);
	return (rval);
	}

/************************************************************************/
/*									*/
/*	Routine:	enc_str						*/
/*									*/
/*	Place a "\0" terminated string into the specified buffer and	*/
/*	advance the buffer pointer past the placed object.		*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to pointer to UCHAR (the buffer).	*/
/*		"s"	The string to place in the buffer.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

static	VOID
enc_str(sptr,s)
	UCHAR			**sptr;
	register STRING		s;
	{
	register STRING		cptr;

	for(cptr = (STRING) *sptr; *s != '\0';)
		*cptr++ = *s++;
	*cptr++ = '\0';
	*sptr = (UCHAR *) cptr;
	}

/************************************************************************/
/*									*/
/*	Routine:	enc_time					*/
/*									*/
/*	Place a broken down time from either "gmtime" or "localtime"	*/
/*	into the specified buffer and advance the buffer pointer past	*/
/*	the placed object.						*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to pointer to UCHAR (the buffer).	*/
/*		"tmptr"	A pointer to the broken down time to be sent.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

static	VOID
enc_time(sptr,tmptr)
	UCHAR			**sptr;
	struct	tm		*tmptr;
	{
	struct	rpc_time	*t;

	t = (struct rpc_time *) *sptr;
	t->tm_sec = tmptr->tm_sec;
	t->tm_min = tmptr->tm_min;
	t->tm_hour = tmptr->tm_hour;
	t->tm_mday = tmptr->tm_mday;
	t->tm_mon = tmptr->tm_mon;
	t->tm_year = tmptr->tm_year;
	t->tm_wday = tmptr->tm_wday;
	t->low_tm_yday = (tmptr->tm_yday & 0xFF);
	t->high_tm_yday = ((tmptr->tm_yday >> 8) & 0xFF);
	t->tm_isdst = tmptr->tm_isdst;

	*sptr += RPC_TIME_DATA;
	}

/************************************************************************/
/*									*/
/*	Routine:	enc_2b, enc_4b					*/
/*									*/
/*	Place a two or four byte signed value into the specified buffer	*/
/*	and advance the buffer pointer past the placed object.		*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to pointer to UCHAR (the buffer).	*/
/*		"value"	The value to place in the buffer.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

static	VOID
enc_2b(sptr,value)
	UCHAR			**sptr;
	int			value;
	{
	*(*sptr)++ = value;
	*(*sptr)++ = (value >> 8);
	}

static	VOID
enc_4b(sptr,value)
	UCHAR			**sptr;
	SLONG			value;
	{
	*(*sptr)++ = value;
	*(*sptr)++ = (value >>= 8);
	*(*sptr)++ = (value >>= 8);
	*(*sptr)++ = (value >> 8);
	}

/************************************************************************/
/*									*/
/*	Routine:	err_dump					*/
/*									*/
/*	Dump the contents of the first 8 bytes of the last message from	*/
/*	the Transputer to the user to aid in diagnosing problems with	*/
/*	the protocol.							*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

#ifndef	NV

static	VOID
err_dump()
	{
	UINT			i;
	UINT			j;

	fprintf(stderr,"Message (node %u): ",current_node);
	for(i = 0; i < 8; i++)
		fprintf(stderr,"%02X ",cmd_buf[i]);
	fprintf(stderr,"\nActive global stream table slots: ");
	for(i = 0; i < MAXFILES; i++)
		{
		if(streams[i].fp != NULL)
			fprintf(stderr,"%u ",i);
		}
	fprintf(stderr,"\n");
	for(i = 0; i < max_nodes; i++)
		{
		if(nodes[i].bpw != 0)
			{
			fprintf(stderr,
				"Active stream table slots for node %d: ",
					i + 1);
			for(j = 0; j < MAXFILES; j++)
				{
				if(nodes[i].strms[j].gstrm != 0)
					fprintf(stderr,"%d ",j);
				}
			fprintf(stderr,"\n");
			}
		}
	}
#endif	/* !NV */

/************************************************************************/
/*									*/
/*	Routine:	fatal						*/
/*									*/
/*	Display error message to user and play dead.			*/
/*									*/
/*	Parameters:							*/
/*		"dump"	TRUE if the message buffer from the Transputer	*/
/*			should be displayed.				*/
/*		"str1"	The first part of the error message to display.	*/
/*		"str2"	The second part of the error message.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

#ifdef	__TURBOC__
#pragma	warn	-par
#endif	/* __TURBOC__ */

static	VOID
fatal(dump,str1,str2)
	BOOL			dump;
	STRING			str1;
	STRING			str2;
	{
#ifdef	NV
	_nv_fatal(current_node,str1,str2);
	longjmp(jexit,1);
#else	/* NV */
	fprintf(stderr,"FATAL: %s%s\n",str1,str2);
	if(dump)
		err_dump();
	if(LinkId > 0)
		(VOID) CloseLink(LinkId);
	exit(ERRORS);
#endif	/* NV */
	}

#ifdef	__TURBOC__
#pragma	warn	.par
#endif	/* __TURBOC__ */

/************************************************************************/
/*									*/
/*	Routine:	get_lstr					*/
/*									*/
/*	Read a string length and contents from the client processor	*/
/*	into the "cmd_buf" buffer.					*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The string length read.				*/
/*									*/
/************************************************************************/

static	int
get_lstr()
	{
	int		len;

	if((len = (int) dec_2b(&cmd_ptr)) > 0)
		read_node(cmd_buf,len);

	return (len);
	}

/************************************************************************/
/*									*/
/*	Routine:	get_str						*/
/*									*/
/*	Read a string from the client processor into the specified	*/
/*	buffer.								*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to pointer to UCHAR (the buffer).	*/
/*		"len"	The length of the string being read.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

static	VOID
get_str(sptr,len)
	UCHAR			*sptr;
	int			len;
	{
	if(len > 0)
		read_node(sptr,len);
	}

/************************************************************************/
/*									*/
/*	Routine:	i_clearerr					*/
/*									*/
/*	Clear the EOF and ERR status indicators for a stream.		*/
/*									*/
/*	Parameters:							*/
/*		"index"	The per-node stream table index (-1 if error).	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

static	VOID
i_clearerr(index)
	int			index;
	{
	if(index < 0)
		return;

	nodes[current_node - 1].strms[index].status &=
		(~ (STREAM_EOF | STREAM_ERR));
	}

/************************************************************************/
/*									*/
/*	Routine:	i_close						*/
/*									*/
/*	Given a index into the stream table for the local node, mark	*/
/*	the file "locally closed".  If the local node is the only one	*/
/*	accessing this particular file then this function will also	*/
/*	actually close the file and mark the global stream table	*/
/*	appropriately.							*/
/*									*/
/*	Parameters:							*/
/*		"index"	The current node local stream table index.	*/
/*									*/
/*	Returns:	Zero on success, RPC_EOF if errors.		*/
/*									*/
/************************************************************************/

static	int
i_close(index)
	int			index;
	{
	int			glb_idx;
	int			rval;

	if(index < 0)
		return (RPC_EOF);
/*
 *	Find the corresponding global stream table entry (if one exists).
 */
	if((glb_idx = nodes[current_node - 1].strms[index].gstrm - 1) < 0)
		return (RPC_EOF);
/*
 *	If this is the only node using the file, REALLY close it.
 */
	rval = 0;
	if((streams[glb_idx].ocnt > 0) && (--streams[glb_idx].ocnt == 0))
		{
		if(fclose(streams[glb_idx].fp) != 0)
			rval = RPC_EOF;

		streams[glb_idx].fp = NULL;	/* Indicate closed */
/*
 *	Return possible free storage pointed to by "name" field.
 */
		if(streams[glb_idx].name != NULL)
			free(streams[glb_idx].name);
		}
	else
		{
/*
 *	Since more than one node/stream is connected to the same file, make
 *	sure if this stream was the last to use the file that the global
 *	"restore" information is zotted.
 */
		if((streams[glb_idx].cnode == current_node) &&
			(streams[glb_idx].cindex == index))
			{
			streams[glb_idx].cnode = 0;
			}
		}
/*
 *	In any event, logically close the file.
 */
	nodes[current_node - 1].strms[index].gstrm = 0;

	return (rval);
	}

/************************************************************************/
/*									*/
/*	Routine:	i_feof						*/
/*									*/
/*	Given a per-node stream table index, determine the EOF status.	*/
/*									*/
/*	Parameters:							*/
/*		"index"	The per-node stream table index (-1 if error).	*/
/*									*/
/*	Returns:	TRUE if the stream @ EOF, FALSE otherwise.	*/
/*									*/
/************************************************************************/

static	BOOL
i_feof(index)
	int			index;
	{
	int			temp;

	if(index < 0)
		return (FALSE);

	temp = nodes[current_node - 1].strms[index].status;
	return ((temp & STREAM_EOF) && (! (temp & STREAM_PUSHB)));
	}

/************************************************************************/
/*									*/
/*	Routine:	i_ferror					*/
/*									*/
/*	Given a per-node stream table index, determine the ERR status.	*/
/*									*/
/*	Parameters:							*/
/*		"index"	The per-node stream table index (-1 if error).	*/
/*									*/
/*	Returns:	TRUE if the stream had an error, FALSE		*/
/*			otherwise.					*/
/*									*/
/************************************************************************/

static	BOOL
i_ferror(index)
	int			index;
	{
	if(index < 0)
		return (FALSE);

	return ((nodes[current_node-1].strms[index].status & STREAM_ERR) != 0);
	}

/************************************************************************/
/*									*/
/*	Routine:	i_init						*/
/*									*/
/*	Initialize the stream "status" field.  This has the effect of	*/
/*	clearing the EOF/ERR and "pushback" character status.		*/
/*									*/
/*	Parameters:							*/
/*		"index"	The per-node stream table index (-1 if error).	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

static	VOID
i_init(index)
	int			index;
	{
	if(index < 0)
		return;

	nodes[current_node - 1].strms[index].status = 0;
	}

/************************************************************************/
/*									*/
/*	Routine:	i_pushb						*/
/*									*/
/*	Given a per-node stream table index, return any character	*/
/*	"pushed back" by a previous "i_ungetc" call.			*/
/*									*/
/*	Parameters:							*/
/*		"index"	The per-node stream table index (-1 if error).	*/
/*									*/
/*	Returns:	The character if available, otherwise EOF.	*/
/*									*/
/************************************************************************/

static	int
i_pushb(index)
	int			index;
	{
	if(index < 0)
		return (EOF);

	if(nodes[current_node - 1].strms[index].status & STREAM_PUSHB)
		{
		nodes[current_node - 1].strms[index].status &= (~STREAM_PUSHB);
		return ((int) nodes[current_node - 1].strms[index].pushb);
		}
	return (EOF);
	}

/************************************************************************/
/*									*/
/*	Routine:	i_stream					*/
/*									*/
/*	Translate a per-node stream table index into the actual stream	*/
/*	pointer obtained from the global stream table.			*/
/*									*/
/*	Parameters:							*/
/*		"index"	The per-node stream table index (-1 if error).	*/
/*									*/
/*	Returns:	The translated stream pointer.			*/
/*									*/
/************************************************************************/

static	STREAM
i_stream(index)
	int			index;
	{
	int			glb_idx;
	int			old_index;
	int			old_node;

	if(index < 0)
		return (NULL);
/*
 *	Find the corresponding global stream table entry.  If the access will
 *	be to one of the pre-opened streams, and this is the first access to
 *	that stream, record the accessing node and stream.  This avoids doing
 *	an unnecessary "fseek" on stdin/stdout/stderr...
 */
	if((glb_idx = nodes[current_node - 1].strms[index].gstrm - 1) <=
		STDERR_INDEX)
		{
		switch(glb_idx)
			{
			case	STDIN_INDEX:
				if(! used_stdin)
					{
					used_stdin = TRUE;
					streams[STDIN_INDEX].cnode =
						current_node;
					streams[STDIN_INDEX].cindex = index;
					}
				break;

			case	STDOUT_INDEX:
				if(! used_stdout)
					{
					used_stdout = TRUE;
					streams[STDOUT_INDEX].cnode =
						current_node;
					streams[STDOUT_INDEX].cindex = index;
					}
				break;

			case	STDERR_INDEX:
				if(! used_stderr)
					{
					used_stderr = TRUE;
					streams[STDERR_INDEX].cnode =
						current_node;
					streams[STDERR_INDEX].cindex = index;
					}
				break;
			}
#ifdef	NV
		sio = TRUE;
#endif	/* NV */
		}
#ifdef	NV
	else
		sio = FALSE;
#endif	/* NV */
/*
 *	If we must "save" the old global file position, do so by using "ftell"
 *	to get the current file position and storing it in the per-node data
 *	associated with the node that used the file previously.  Then seek
 *	the file to the position appropriate to the current node context.  Note
 *	that this is only necessary if the previous user and the current node
 *	are different!
 */
	old_index = streams[glb_idx].cindex;
	old_node = streams[glb_idx].cnode;
	if((current_node != old_node) || (index != old_index))
		{
		if((old_node--) != 0)
			{
/*
 *	Save old context (file position), if needed.
 */
			nodes[old_node].strms[old_index].pos =
				ftell(streams[glb_idx].fp);
			}
/*
 *	Restore new context.
 */
		(VOID) fseek(streams[glb_idx].fp,nodes[current_node - 1].
			strms[index].pos,SEEK_SET);
		streams[glb_idx].cnode = current_node;
		streams[glb_idx].cindex = index;
		}
/*
 *	Return the appropriate stream pointer.
 */
	return (streams[glb_idx].fp);
	}

/************************************************************************/
/*									*/
/*	Routine:	i_ungetc					*/
/*									*/
/*	Push the specified character "back" on the specified stream.	*/
/*									*/
/*	Parameters:							*/
/*		"c"	The character to "pushback".			*/
/*		"index"	The per-node stream table index (-1 if error).	*/
/*									*/
/*	Returns:	RPC_EOF if we already had a character pushed	*/
/*			back (or had an error), otherwise the character	*/
/*			value.						*/
/*									*/
/************************************************************************/

static	int
i_ungetc(c,index)
	UCHAR			c;
	int			index;
	{
	if(index < 0)
		return (RPC_EOF);

	nodes[current_node - 1].strms[index].pushb = c;
	
	if(nodes[current_node - 1].strms[index].status & STREAM_PUSHB)
		return (RPC_EOF);
	else
		nodes[current_node - 1].strms[index].status |= STREAM_PUSHB;

	return ((int) c);
	}

/************************************************************************/
/*									*/
/*	Routine:	i_update					*/
/*									*/
/*	Set the ERR and EOF statuses for the specified stream.		*/
/*									*/
/*	Parameters:							*/
/*		"index"	The per-node stream table index (-1 if error).	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

static	VOID
i_update(index)
	int			index;
	{
	BOOL			needs_clearerr;
	STREAM			fp;

	if(index < 0)
		return;

	needs_clearerr = FALSE;
	fp = i_stream(index);
	if(w_ferror(fp))
		{
		nodes[current_node - 1].strms[index].status |= STREAM_ERR;
		needs_clearerr = TRUE;
		}
	if(w_feof(fp))
		{
		nodes[current_node - 1].strms[index].status |= STREAM_EOF;
		needs_clearerr = TRUE;
		}
	if(needs_clearerr)
		w_clearerr(fp);
	}

/************************************************************************/
/*									*/
/*	Routine:	node_alloc					*/
/*									*/
/*	Allocate enough storage to handle the per-node data structures	*/
/*	at least up through the specified node.				*/
/*									*/
/*	Parameters:							*/
/*		"node"	The requesting node #.				*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

static	VOID
node_alloc(node)
	UINT			node;
	{
	UINT			old_max_nodes;
/*
 *	Need to allocate some additional "node_info" structures.  Allocate at
 *	least PER_NODE_ALLOCATION slots, or if the requested node is larger
 *	than this allocate the minimum # needed to support the requested node.
 *	Also, if "node" is 1 (ie. allocating the first node), ignore the
 *	allocation increment.
 */
	old_max_nodes = max_nodes;
	if((node == 1) || ((max_nodes += PER_NODE_ALLOCATION) < node))
		max_nodes = node;

	if((nodes = (struct node_info *) realloc(nodes,
		(max_nodes * sizeof(struct node_info)))) == NULL)
		{
		fatal(FALSE,"Insufficient per-node info memory","");
		}
/*
 *	Initialize "bytes-per-word" field in newly allocated node info
 *	structures to zero, to mark them as currently unused.
 */
	while(old_max_nodes < max_nodes)
		nodes[old_max_nodes++].bpw = 0;
	}

/************************************************************************/
/*									*/
/*	Routine:	ordy_add					*/
/*									*/
/*	Add a node to the list of those with pending I/O requests.  If	*/
/*	this is the first time this node has been encountered we may	*/
/*	need to allocate storage for the per-node data structure.	*/
/*									*/
/*	Parameters:							*/
/*		"node"	The requesting node #.				*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

static	VOID
ordy_add(node)
	UINT			node;
	{
	if(node > max_nodes)
		node_alloc(node);
/*
 *	Add the node to the queue of those which are requesting I/O services.
 */
	nodes[node - 1].next = 0;
	if(head_ordy == 0)
		head_ordy = tail_ordy = node;
	else
		{
		nodes[tail_ordy - 1].next = node;
		tail_ordy = node;
		}
	}
#ifdef	NV

/************************************************************************/
/*									*/
/*	Routine:	q_add						*/
/*									*/
/*	Add an I/O request to the "blocked" queue.			*/
/*									*/
/*	Parameters:							*/
/*		"func"	The polling function to queue.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

static	VOID
q_add(func)
	BOOL			(*func)();
	{
/*
 *	Wait until we get an "Input Process Ready" back from the selected
 *	node.  We need to do this here so that later when we actually complete
 *	the I/O request we know that the node is ready for the results.
 */
	while(read_vchdr() != (_VCH_CNTL | _VCH_IRDY))
		;
/*
 *	Add the node to the queue of those which are blocked awaiting I/O.
 */
	nodes[current_node - 1].blkio.func = func;
	nodes[current_node - 1].next = 0;
	if(head_irdy == 0)
		head_irdy = tail_irdy = current_node;
	else
		{
		nodes[tail_irdy - 1].next = current_node;
		tail_irdy = current_node;
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	q_fgetc						*/
/*									*/
/*	This function complements "w_fgetc" and is called to poll NV	*/
/*	to see if a "_nv_fgetc" operation can now be completed.		*/
/*									*/
/*	Parameters:							*/
/*		"node"	The node which is blocked awaiting I/O.		*/
/*									*/
/*	Returns:	TRUE if the I/O is completed, FALSE otherwise.	*/
/*									*/
/************************************************************************/

static	BOOL
q_fgetc(node)
	UINT			node;
	{
	int			rval;

	if(! _nv_fgetc(node,nodes[node - 1].blkio.fp,&rval))
		return (FALSE);
/*
 *	We have completed the I/O operation!  Inform the appropriate node of
 *	the results.
 */
	current_node = node;
	q_r_2beof(rval);
	i_update(nodes[current_node - 1].blkio.index);

	return (TRUE);
	}

/************************************************************************/
/*									*/
/*	Routine:	q_fgets						*/
/*									*/
/*	This function complements "w_fgets" and is called to poll NV	*/
/*	to see if a "_nv_fgets" operation can now be completed.		*/
/*									*/
/*	Parameters:							*/
/*		"node"	The node which is blocked awaiting I/O.		*/
/*									*/
/*	Returns:	TRUE if the I/O is completed, FALSE otherwise.	*/
/*									*/
/************************************************************************/

static	BOOL
q_fgets(node)
	UINT			node;
	{
	STRING			rval;
	int			temp;

	temp = ((nodes[node - 1].blkio.pushb < 0) ? 0 : 1);
	if(! _nv_fgets(node,((STRING) rpy_buf) + temp,
		nodes[node - 1].blkio.length - temp,
			nodes[node - 1].blkio.fp,&rval))
		{
		return (FALSE);
		}
/*
 *	We have completed the I/O operation!  Inform the appropriate node of
 *	the results.
 */
	if(temp > 0)			/* Include pushback char if present */
		rpy_buf[0] = nodes[node - 1].blkio.pushb;

	current_node = node;
/*
 *	Return NULL if result was EOF or error.
 */
	if((rval == NULL) && ((temp == 0) ||
		w_ferror(nodes[current_node - 1].blkio.fp)))
		{
		q_r_2b(RPC_NULL);
		}
	else
		{
/*
 *	Otherwise, return the string that was read.
 */
		temp = strlen((STRING) rpy_buf) + 1;
		q_r_2b(temp);
		reply_str(rpy_buf,temp);
		}
	i_update(nodes[current_node - 1].blkio.index);

	return (TRUE);
	}

/************************************************************************/
/*									*/
/*	Routine:	q_fread						*/
/*									*/
/*	This function complements "w_fread" and is called to poll NV	*/
/*	to see if a "_nv_fread" operation can now be completed.		*/
/*									*/
/*	Parameters:							*/
/*		"node"	The node which is blocked awaiting I/O.		*/
/*									*/
/*	Returns:	TRUE if the I/O is completed, FALSE otherwise.	*/
/*									*/
/************************************************************************/

static	BOOL
q_fread(node)
	UINT			node;
	{
	UINT			rval;
	int			temp;

	temp = ((nodes[node - 1].blkio.pushb < 0) ? 0 : 1);
	if(! _nv_fread(node,((STRING) rpy_buf) + temp,(UINT) 1,
		(UINT) (nodes[node - 1].blkio.length - temp),
			nodes[node - 1].blkio.fp,&rval))
		{
		return (FALSE);
		}
/*
 *	We have completed the I/O operation!  Inform the appropriate node of
 *	the results.
 */
	if(temp > 0)			/* Include pushback char if present */
		rpy_buf[0] = nodes[node - 1].blkio.pushb;

	current_node = node;
/*
 *	Return NULL if result was EOF or error.
 */
	rval += temp;
	q_r_2b((int) rval);
	reply_str(rpy_buf,(int) rval);
	i_update(nodes[current_node - 1].blkio.index);

	return (TRUE);
	}
#if	H_OPSYS == MS_DOS

/************************************************************************/
/*									*/
/*	Routine:	q_getch						*/
/*									*/
/*	This function complements "w_getch" and is called to poll NV	*/
/*	to see if a "_nv_getch" operation can now be completed.		*/
/*									*/
/*	Parameters:							*/
/*		"node"	The node which is blocked awaiting I/O.		*/
/*									*/
/*	Returns:	TRUE if the I/O is completed, FALSE otherwise.	*/
/*									*/
/************************************************************************/

static	BOOL
q_getch(node)
	UINT			node;
	{
	int			rval;

	if(! _nv_getch(node,&rval))
		return (FALSE);

	current_node = node;
	q_r_2b(rval);

	return (TRUE);
	}

/************************************************************************/
/*									*/
/*	Routine:	q_getche					*/
/*									*/
/*	This function complements "w_getche" and is called to poll NV	*/
/*	to see if a "_nv_getche" operation can now be completed.	*/
/*									*/
/*	Parameters:							*/
/*		"node"	The node which is blocked awaiting I/O.		*/
/*									*/
/*	Returns:	TRUE if the I/O is completed, FALSE otherwise.	*/
/*									*/
/************************************************************************/

static	BOOL
q_getche(node)
	UINT			node;
	{
	int			rval;

	if(! _nv_getche(node,&rval))
		return (FALSE);

	current_node = node;
	q_r_2b(rval);

	return (TRUE);
	}
#endif	/* MS_DOS */

/************************************************************************/
/*									*/
/*	Routine:	q_r_2b						*/
/*									*/
/*	Returns a two byte result code to the Transputer.  Doesn't	*/
/*	require an "Input Process Ready" since one was read when the	*/
/*	blocked I/O request was originally queued.			*/
/*									*/
/*	Parameters:							*/
/*		"value"	The value to return.				*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

static	VOID
q_r_2b(value)
	int			value;
	{
	UCHAR			buf[2];

	buf[0] = value;
	buf[1] = (value >> 8);

	write_node(buf,2,FALSE);
	}

/************************************************************************/
/*									*/
/*	Routine:	q_r_2beof					*/
/*									*/
/*	Returns a two byte result code to the Transputer.  If the	*/
/*	parameter is EOF the value is converted to RPC_EOF, otherwise	*/
/*	the value is passed through unchanged.  Doesn't	require an	*/
/*	"Input Process Ready" since one was read when the blocked I/O	*/
/*	request was originally queued.					*/
/*									*/
/*	Parameters:							*/
/*		"value"	The value to return.				*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

static	VOID
q_r_2beof(value)
	int			value;
	{
	if(value == EOF)
		value = RPC_EOF;
	q_r_2b(value);
	}
#endif	/* NV */

/************************************************************************/
/*									*/
/*	Routine:	read_node					*/
/*									*/
/*	Read a message from the currently selected node.		*/
/*									*/
/*	Parameters:							*/
/*		"buf"	The address to store the data.			*/
/*		"len"	The amount of data to read (in bytes).		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

static	VOID
read_node(buf,len)
	UCHAR			*buf;
	int			len;
	{
	int			tlen;

	if(use_vchans)
		{
/*
 *	We must send the "Input Process Ready" control message to the
 *	selected node.
 */
		write_vchdr((_VCH_CNTL | _VCH_IRDY),current_chan,current_node);
/*
 *	Read message.
 */
		while(len > 0)
			{
/*
 *	Read message headers.  Ignore control and status messages.
 */
			while((tlen = read_vchdr()) & _VCH_CNTL)
				;
/*
 *	If this is the last packet in the message, make sure the read/write
 *	message lengths match.
 */
			if((tlen & _VCH_EOMP) &&
				((tlen &= (~_VCH_EOMP)) != len))
				{
				fatal(TRUE,"Bad read packet message length",
					"");
				}
/*
 *	We have a data packet, read it.
 */
			if(ReadLink(LinkId,(STRING) buf,tlen,timeout) != tlen)
				{
				fatal(FALSE,
					"Error/timeout reading from link: ",
						open_str);
				}
			buf += tlen;
			len -= tlen;
			}
		}
	else
		{
		if(ReadLink(LinkId,(STRING) buf,len,timeout) != len)
			{
			fatal(FALSE,"Error/timeout reading from link: ",
				open_str);
			}
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	read_vchdr					*/
/*									*/
/*	Read a message packet header.  If the packet is a control	*/
/*	message, record it.  If it is a data record, do nothing.	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The packet header length/status field.		*/
/*									*/
/************************************************************************/

static	UINT
read_vchdr()
	{
	BOOL			debug_hdr;
	UCHAR			header[4];
	UINT			node;
	UINT			status;
/*
 *	Read message packet headers.
 */
	if(ReadLink(LinkId,(STRING) header,4,timeout) != 4)
		fatal(FALSE,"Error/timeout reading from link: ",open_str);

	status = header[0] | (header[1] << 8);
	node = (header[2] | (header[3] << 8));
/*
 *	If the message was from one of the debug kernels, note the fact and
 *	assume the message is from the currently selected debug kernel.
 */
	if(node == 0)
		{
		node = current_node;
		debug_hdr = TRUE;
		}
	else
		debug_hdr = FALSE;
/*
 *	See if we have a protocol control/status message.
 */
	if(status & _VCH_CNTL)
		{
/*
 *	It was a control/status message.  If it is a "Routing Buffer Available"
 *	message, record the fact and return.
 */
		if(status == (_VCH_CNTL | _VCH_RRDY))
			{
			vc_rrdy = FALSE;	/* OK to re-use */
			return (status);
			}
/*
 *	Otherwise, decode the control/status field and do whatever is required.
 */
		switch((int) status)
			{
			case ((int) (_VCH_CNTL | _VCH_IRDY)):
				if(node != current_node)
					{
				fatal(TRUE,"Input ready on wrong node","");
					}
				break;

			case ((int) (_VCH_CNTL | _VCH_ORDY)):
/*
 *	If the "Output Ready" message was from one of the debug kernels then
 *	make the return value be a generic "_VCH_CNTL" so that it is not
 *	confused with a "Output Ready" generated by a user process.
 */
				if(debug_hdr)
					status = _VCH_CNTL;
				else
					{
					if((current_chan != VC_HOSTIO) ||
						(node != current_node))
						{
						ordy_add(node);
						}
					}
				break;

			default:
				fatal(TRUE,"Control packet error","");
			}
		}
/*
 *	Return the results.
 */
	return (status);
	}

/************************************************************************/
/*									*/
/*	Routine:	reply_open					*/
/*									*/
/*	Given a host I/O system supplied stream pointer translate into	*/
/*	a integer offset into the stream translation table to preserve	*/
/*	host pointer size transportability.  Send the result as a two	*/
/*	byte reply message.						*/
/*									*/
/*	Parameters:							*/
/*		"fp"	The host stream pointer to translate.  If this	*/
/*			value is NULL then it is assumed that "name"	*/
/*			and "mode" are to be used with "fopen" (or an	*/
/*			error has occurred).				*/
/*		"slot"	If less than zero, use the first free per-node	*/
/*			stream table slot.  Otherwise use "slot".	*/
/*		"name"	The file name to use if "fopen" needs calling.	*/
/*		"nlen"	The length of "name".				*/
/*		"mode"	The file mode to use if "fopen" needs calling.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

static	VOID
reply_open(fp,slot,name,nlen,mode)
	STREAM			fp;
	int			slot;
	STRING			name;
	int			nlen;
	STRING			mode;
	{
	UINT			count;
	int			index;
/*
 *	Verify that either the file pointer is non-NULL, or both the name and
 *	mode are non-NULL.
 */
	if((fp == NULL) && ((name == NULL) || (mode == NULL)))
		{
		reply_2b(RPC_NULL);	/* If couldn't be opened */
		return;
		}
/*
 *	If the "name" pointer is non-NULL, then we need to check and see if
 *	the file has already been opened.
 */
	if(name != NULL)
		{
		index = -1;
		for(count = 0; count < max_open; count++)
			{
/*
 *	Determine if the slot is in use.
 */
			if(streams[count].fp == NULL)
				{
/*
 *	Nope, keep track of first free global stream table slot.
 */
				if(index == -1)
					index = count;
				}
/*
 *	See if the slot is already in use.  If so, see if the file name is the
 *	same as our current request.
 */
			else if((streams[count].name != NULL) &&
				(strcmp(name,streams[count].name) == 0))
				{
/*
 *	The file names matched, quit searching.
 */
				break;
				}
			}
/*
 *	If we didn't find a copy of the file already open, see if we had a
 *	unused slot and use that instead.
 */
		if((count == max_open) && (index != -1))
			count = index;
		}
	else
		{
/*
 *	The file has already been opened and we just need to find the first
 *	free global file slot to record the file pointer.
 */
		for(count = 0; count < max_open; count++)
			{
/*
 *	Find first free global stream table slot.
 */
			if(streams[count].fp == NULL)
				break;
			}
		}
/*
 *	Find the per-node stream table slot.  If "slot" is less than zero we
 *	just find the first free per-node stream table slot.  Otherwise,
 *	use the "slot" entry in the per-node table.
 */
	if(slot < 0)
		{
		for(index = 0; index < MAXFILES; index++)
			{
			if(nodes[current_node - 1].strms[index].gstrm == 0)
				break;
			}
		if(index == MAXFILES)
			{
			warning("Out of per-node file stream slots","");
			if(fp != NULL)
				(VOID) fclose(fp);
			reply_2b(RPC_NULL);
			return;
			}
		}
	else
		index = slot;		/* Use parameter per-node slot index */

	nodes[current_node - 1].strms[index].gstrm = count + 1;
	nodes[current_node - 1].strms[index].pos = 0;
/*
 *	Clear ERR and EOF flags and flush potential "pushback" char.
 */
	i_init(index);
/*
 *	If this file pointer will be using a new global stream slot update
 *	"max_open" to keep track of the new maximum.
 */
	if(count == max_open)
		{
		if(count == MAXFILES)
			{
			warning("Out of global file stream slots","");
			if(fp != NULL)
				(VOID) fclose(fp);
			nodes[current_node - 1].strms[index].gstrm = 0;
			reply_2b(RPC_NULL);
			return;
			}
		else
			max_open++;	/* One more file slot used */
		}
/*
 *	If the global stream table slot is non-NULL then the file is already
 *	open.
 */
	if(streams[count].fp == NULL)
		{
/*
 *	If "fp" is NULL, then we must try opening the file.
 */
		if(fp == NULL)
			{
			if((fp = fopen(name,mode)) == NULL)
				{
				nodes[current_node - 1].strms[index].gstrm = 0;
				reply_2b(RPC_NULL);
				return;
				}
			}
/*
 *	Either way, we must initialize the global stream table slot.
 */
		streams[count].cnode = current_node;	/* Last used by node */
		streams[count].cindex = index;		/* Last used by strm */

		streams[count].fp = fp;		/* Host stream pointer */

		if(name == NULL)		/* Filename */
			streams[count].name = NULL;
		else
			{
			if((streams[count].name =
				(STRING) malloc(nlen)) == NULL)
				{
				fatal(FALSE,
					"Insufficient filename string memory",
						"");
				}
			(VOID) strcpy(streams[count].name,name);
			}

		streams[count].ocnt = 1;	/* # times has been opened */
		}
	else
		{
/*
 *	The file was already open, increment the "# of times opened" count.
 *	If "ocnt" is negative, then don't bother since the file can NEVER
 *	be closed in any event (ie. stdin/stdout/stderr from a global
 *	perspective).
 */
		if(streams[count].ocnt > 0)
			streams[count].ocnt++;
		}
/*
 *	Reply with the per-node stream table index (+1 to avoid conflicts with
 *	RPC_NULL).
 */
	reply_2b(index + 1);
	}

/************************************************************************/
/*									*/
/*	Routine:	reply_str					*/
/*									*/
/*	Returns a fixed length string to the Transputer.		*/
/*									*/
/*	Parameters:							*/
/*		"buf"	The pointer to the buffer to send.		*/
/*		"len"	The # of bytes to send.				*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

static	VOID
reply_str(buf,len)
	UCHAR			*buf;
	int			len;
	{
	if(len > 0)
		write_node(buf,len,TRUE);
	}

/************************************************************************/
/*									*/
/*	Routine:	reply_zeof					*/
/*									*/
/*	Returns a one byte result code to the Transputer.  If the	*/
/*	parameter is zero the value is sent unchanged, otherwise the	*/
/*	value is converted to a one byte version of RPC_EOF.		*/
/*									*/
/*	Parameters:							*/
/*		"value"	The value to return.				*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

static	VOID
reply_zeof(value)
	int			value;
	{
	UCHAR			c;

	if(value != 0)
		c = (UCHAR) RPC_EOF;
	else
		c = 0;

	write_node(&c,1,TRUE);
	}

/************************************************************************/
/*									*/
/*	Routine:	reply_1b					*/
/*									*/
/*	Returns a one byte result code to the Transputer.		*/
/*									*/
/*	Parameters:							*/
/*		"value"	The value to return.				*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

static	VOID
reply_1b(value)
	int			value;
	{
	UCHAR			buf[1];

	buf[0] = value;

	write_node(buf,1,TRUE);
	}

/************************************************************************/
/*									*/
/*	Routine:	reply_2b					*/
/*									*/
/*	Returns a two byte result code to the Transputer.		*/
/*									*/
/*	Parameters:							*/
/*		"value"	The value to return.				*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

static	VOID
reply_2b(value)
	int			value;
	{
	UCHAR			buf[2];

	buf[0] = value;
	buf[1] = (value >> 8);

	write_node(buf,2,TRUE);
	}

/************************************************************************/
/*									*/
/*	Routine:	reply_2beof					*/
/*									*/
/*	Returns a two byte result code to the Transputer.  If the	*/
/*	parameter is EOF the value is converted to RPC_EOF, otherwise	*/
/*	the value is passed through unchanged.				*/
/*									*/
/*	Parameters:							*/
/*		"value"	The value to return.				*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

static	VOID
reply_2beof(value)
	int			value;
	{
	if(value == EOF)
		value = RPC_EOF;
	reply_2b(value);
	}

/************************************************************************/
/*									*/
/*	Routine:	reply_4b					*/
/*									*/
/*	Returns a four byte result code to the Transputer.		*/
/*									*/
/*	Parameters:							*/
/*		"value"	The value to return.				*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

static	VOID
reply_4b(value)
	SLONG			value;
	{
	UCHAR			buf[4];

	buf[0] = value;
	buf[1] = (value >>= 8);
	buf[2] = (value >>= 8);
	buf[3] = (value >> 8);

	write_node(buf,4,TRUE);
	}

/************************************************************************/
/*									*/
/*	Routine:	select_node					*/
/*									*/
/*	Tell the node with the oldest pending I/O request to go ahead	*/
/*	and send the request.  If no requests are pending, await the	*/
/*	arrival of one here.  If we are running as part of NV we won't	*/
/*	be called unless there is a pending I/O request...		*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

static	VOID
select_node()
	{
#ifndef	NV
	current_node = 0;		/* Requests from ANYONE go to queue */
/*
 *	If there are any I/O requests in the queue, satisfy the oldest,
 *	otherwise wait until we have an I/O request.
 */
	if(head_ordy == 0)
		{
		while(read_vchdr() != (_VCH_CNTL | _VCH_ORDY))
			;
		}
#endif	/* !NV */
/*
 *	Honor the oldest I/O request in the queue.
 */
	current_node = head_ordy;
	if(current_node == 1)
		vc_plen = vc_aplen;	/* Root node, use adj max packet len */
	else
		vc_plen = vc_rplen;	/* Remote node, use remote max len */
/*
 *	Delete request from queue.
 */
	head_ordy = nodes[current_node - 1].next;
	}

/************************************************************************/
/*									*/
/*	Routine:	warning						*/
/*									*/
/*	Display warning message to user.				*/
/*									*/
/*	Parameters:							*/
/*		"str1"	The first part of the message to display.	*/
/*		"str2"	The second part of the message.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

static	VOID
warning(str1,str2)
	STRING			str1;
	STRING			str2;
	{
#ifdef	NV
	_nv_warning(current_node,str1,str2);
#else	/* NV */
	fprintf(stderr,"WARNING: %s%s\n",str1,str2);
	err_dump();
#endif	/* NV */
	}

/************************************************************************/
/*									*/
/*	Routine:	write_node					*/
/*									*/
/*	Write a message to the currently selected node.			*/
/*									*/
/*	Parameters:							*/
/*		"buf"	The address of the data to end.			*/
/*		"len"	The amount of data to send (in bytes).		*/
/*		"ipr"	TRUE/need to wait for an "Input Process Ready".	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

static	VOID
write_node(buf,len,ipr)
	UCHAR			*buf;
	int			len;
	BOOL			ipr;
	{
	if(use_vchans)
		{
/*
 *	Wait until we get an "Input Process Ready" back from the selected
 *	node (if we haven't already gotten it!)
 */
		if(ipr)
			{
			while(read_vchdr() != (_VCH_CNTL | _VCH_IRDY))
				;
			}
/*
 *	Write header and data to selected node.  Break a long write up into
 *	multiple packets.
 */
		while(len > vc_plen)
			{
			write_vchdr(vc_plen,current_chan,current_node);
			if(WriteLink(LinkId,(STRING) buf,vc_plen,timeout) !=
				vc_plen)
				{
				fatal(FALSE,"Error/timeout writing to link: ",
					open_str);
				}
			buf += vc_plen;
			len -= vc_plen;
			}
		write_vchdr((_VCH_EOMP | len),current_chan,current_node);
		}
/*
 *	Write the last data packet of a virtual channel message, or ALL the
 *	data of a non-protocol message.
 */
	if(WriteLink(LinkId,(STRING) buf,len,timeout) != len)
		fatal(FALSE,"Error/timeout writing to link: ",open_str);
	}

/************************************************************************/
/*									*/
/*	Routine:	write_vchdr					*/
/*									*/
/*	Write a virtual channel header message addressed to the		*/
/*	selected destination node and channel.				*/
/*									*/
/*	Parameters:							*/
/*		"len"	The length/control status field.		*/
/*		"chan"	The destination virtual channel.		*/
/*		"dest"	The destination node.				*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

static	VOID
write_vchdr(len,chan,dest)
	UINT			len;
	UINT			chan;
	UINT			dest;
	{
	UCHAR			header[6];

	if(dest == 1)			/* Non-routed virtual channel */
		{
		header[0] = len;
		header[1] = len >> 8;
		header[2] = chan;
		header[3] = chan >> 8;
		if(WriteLink(LinkId,(STRING) header,4,timeout) != 4)
			{
			fatal(FALSE,"Error/timeout writing to link: ",
				open_str);
			}
		}
	else				/* Routed virtual channel */
		{
/*
 *	If the routing buffer on the root node is already in use, wait
 *	until it is available.
 */
		if(vc_rrdy)
			{
			while(read_vchdr() != (_VCH_CNTL | _VCH_RRDY))
				;
			}
/*
 *	Send routed message.  Set the message length/control field as
 *	externally provided.  Set the channel/hop count field to indicate
 *	that this field holds a hop count of 1 (_VCH_CNTL | 1).
 */
		header[0] = len;
		header[1] = len >> 8;
		header[2] = 1;		/* This is the first hop! */
		header[3] = (_VCH_CNTL >> 8);	/* Indicate hop count field */
		if(WriteLink(LinkId,(STRING) header,4,timeout) != 4)
			{
			fatal(FALSE,"Error/timeout writing to link: ",
				open_str);
			}
		header[0] = dest;
		header[1] = dest >> 8;
		header[2] = chan;
		header[3] = chan >> 8;
/*
 *	Note that although we send two undefined bytes (header[4], header[5]),
 *	for the "reserved buffer tag" information, we don't care since
 *	"read_vchdr" doesn't look at the tag information when it is returned
 *	in any event...
 */
		if(WriteLink(LinkId,(STRING) header,6,timeout) != 6)
			{
			fatal(FALSE,"Error/timeout writing to link: ",
				open_str);
			}
/*
 *	Mark the root "routing" buffer as being in use.
 */
		vc_rrdy = TRUE;
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	w_clearerr					*/
/*									*/
/*	Interface function for "clearerr" library routine.  Having this	*/
/*	interface layer permits standard I/O from different nodes to	*/
/*	be read from, and written to, "logically different terminals".	*/
/*									*/
/************************************************************************/

static	VOID
w_clearerr(fp)
	STREAM			fp;
	{
#ifdef	NV
	if(sio)
		_nv_clearerr(current_node,fp);
	else
#endif	/* NV */
		clearerr(fp);
	}

/************************************************************************/
/*									*/
/*	Routine:	w_feof						*/
/*									*/
/*	Interface function for "feof" library routine.  Having this	*/
/*	interface layer permits standard I/O from different nodes to	*/
/*	be read from, and written to, "logically different terminals".	*/
/*									*/
/************************************************************************/

static	int
w_feof(fp)
	STREAM			fp;
	{
#ifdef	NV
	if(sio)
		return (_nv_feof(current_node,fp));
	else
#endif	/* NV */
		return (feof(fp));
	}

/************************************************************************/
/*									*/
/*	Routine:	w_ferror					*/
/*									*/
/*	Interface function for "ferror" library routine.  Having this	*/
/*	interface layer permits standard I/O from different nodes to	*/
/*	be read from, and written to, "logically different terminals".	*/
/*									*/
/************************************************************************/

static	int
w_ferror(fp)
	STREAM			fp;
	{
#ifdef	NV
	if(sio)
		return (_nv_ferror(current_node,fp));
	else
#endif	/* NV */
		return (ferror(fp));
	}

/************************************************************************/
/*									*/
/*	Routine:	w_fflush					*/
/*									*/
/*	Interface function for "fflush" library routine.  Having this	*/
/*	interface layer permits standard I/O from different nodes to	*/
/*	be read from, and written to, "logically different terminals".	*/
/*									*/
/************************************************************************/

static	int
w_fflush(fp)
	STREAM			fp;
	{
#ifdef	NV
	if(sio)
		return (_nv_fflush(current_node,fp));
	else
#endif	/* NV */
		return (fflush(fp));
	}

/************************************************************************/
/*									*/
/*	Routine:	w_fgetc						*/
/*									*/
/*	Interface function for "fgetc" library routine.  Having this	*/
/*	interface layer permits standard I/O from different nodes to	*/
/*	be read from, and written to, "logically different terminals".	*/
/*									*/
/*	Parameters:							*/
/*		"index"	The per-node stream table index.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

static	VOID
w_fgetc(index)
	int			index;
	{
	STREAM			fp;
	int			rval;

	fp = i_stream(index);
#ifdef	NV
	if(sio)
		{
		if(! _nv_fgetc(current_node,fp,&rval))
			{
			nodes[current_node - 1].blkio.fp = fp;
			nodes[current_node - 1].blkio.index = index;
			q_add(q_fgetc);
			return;
			}
		}
	else
#endif	/* NV */
		rval = fgetc(fp);

	reply_2beof(rval);
	i_update(index);
	}

/************************************************************************/
/*									*/
/*	Routine:	w_fgets						*/
/*									*/
/*	Interface function for "fgets" library routine.  Having this	*/
/*	interface layer permits standard I/O from different nodes to	*/
/*	be read from, and written to, "logically different terminals".	*/
/*									*/
/*	Parameters:							*/
/*		"index"	The per-node stream table index.		*/
/*		"length"The requested read length.			*/
/*		"pushb"	1 if a pushback character available, zero else.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

static	VOID
w_fgets(index,length,pushb)
	int			index;
	int			length;
	int			pushb;
	{
	STREAM			fp;
	STRING			rval;

	fp = i_stream(index);
#ifdef	NV
	if(sio)
		{
		if(! _nv_fgets(current_node,((STRING) rpy_buf) + pushb,
			length - pushb,fp,&rval))
			{
			nodes[current_node - 1].blkio.fp = fp;
			nodes[current_node - 1].blkio.index = index;
			nodes[current_node - 1].blkio.length = length;
			if(pushb)
				{
				nodes[current_node - 1].blkio.pushb =
					rpy_buf[0];
				}
			else
				nodes[current_node - 1].blkio.pushb = -1;

			q_add(q_fgets);
			return;
			}
		}
	else
#endif	/* NV */
		rval = fgets(((STRING) rpy_buf) + pushb,length - pushb,fp);

	if((rval == NULL) && ((! pushb) || w_ferror(fp)))
		reply_2b(RPC_NULL);
	else
		{
		length = strlen((STRING) rpy_buf) + 1;
		reply_2b(length);
		reply_str(rpy_buf,length);
		}
	i_update(index);
	}

/************************************************************************/
/*									*/
/*	Routine:	w_fputc						*/
/*									*/
/*	Interface function for "fputc" library routine.  Having this	*/
/*	interface layer permits standard I/O from different nodes to	*/
/*	be read from, and written to, "logically different terminals".	*/
/*									*/
/************************************************************************/

static	int
w_fputc(c,fp)
	int			c;
	STREAM			fp;
	{
#ifdef	NV
	if(sio)
		return (_nv_fputc(current_node,c,fp));
	else
#endif	/* NV */
		return (fputc(c,fp));
	}

/************************************************************************/
/*									*/
/*	Routine:	w_fread						*/
/*									*/
/*	Interface function for "fread" library routine.  Having this	*/
/*	interface layer permits standard I/O from different nodes to	*/
/*	be read from, and written to, "logically different terminals".	*/
/*									*/
/*	Parameters:							*/
/*		"index"	The per-node stream table index.		*/
/*		"length"The requested read length.			*/
/*		"pushb"	1 if a pushback character available, zero else.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

static	VOID
w_fread(index,length,pushb)
	int			index;
	int			length;
	int			pushb;
	{
	STREAM			fp;
	UINT			rval;

	fp = i_stream(index);
#ifdef	NV
	if(sio)
		{
		if(! _nv_fread(current_node,((STRING) rpy_buf) + pushb,
			(UINT) 1,(UINT) (length - pushb),fp,&rval))
			{
			nodes[current_node - 1].blkio.fp = fp;
			nodes[current_node - 1].blkio.index = index;
			nodes[current_node - 1].blkio.length = length;
			if(pushb)
				{
				nodes[current_node - 1].blkio.pushb =
					rpy_buf[0];
				}
			else
				nodes[current_node - 1].blkio.pushb = -1;

			q_add(q_fread);
			return;
			}
		}
	else
#endif	/* NV */
		{
		rval = fread(((STRING) rpy_buf) + pushb,(UINT) 1,
			(UINT) (length - pushb),fp);
		}
	rval += pushb;
	reply_2b((int) rval);
	reply_str(rpy_buf,(int) rval);
	i_update(index);
	}

/************************************************************************/
/*									*/
/*	Routine:	w_fseek						*/
/*									*/
/*	Interface function for "fseek" library routine.  Having this	*/
/*	interface layer permits standard I/O from different nodes to	*/
/*	be read from, and written to, "logically different terminals".	*/
/*									*/
/************************************************************************/

static	int
w_fseek(fp,offset,origin)
	STREAM			fp;
	long			offset;
	int			origin;
	{
#ifdef	NV
	if(sio)
		return (_nv_fseek(current_node,fp,offset,origin));
	else
#endif	/* NV */
		return (fseek(fp,offset,origin));
	}

/************************************************************************/
/*									*/
/*	Routine:	w_ftell						*/
/*									*/
/*	Interface function for "ftell" library routine.  Having this	*/
/*	interface layer permits standard I/O from different nodes to	*/
/*	be read from, and written to, "logically different terminals".	*/
/*									*/
/************************************************************************/

static	long
w_ftell(fp)
	STREAM			fp;
	{
#ifdef	NV
	if(sio)
		return (_nv_ftell(current_node,fp));
	else
#endif	/* NV */
		return (ftell(fp));
	}

/************************************************************************/
/*									*/
/*	Routine:	w_fwrite					*/
/*									*/
/*	Interface function for "fwrite" library routine.  Having this	*/
/*	interface layer permits standard I/O from different nodes to	*/
/*	be read from, and written to, "logically different terminals".	*/
/*									*/
/************************************************************************/

static	UINT
w_fwrite(ptr,size,n,fp)
	STRING			ptr;
	UINT			size;
	UINT			n;
	STREAM			fp;
	{
#ifdef	NV
	if(sio)
		return (_nv_fwrite(current_node,ptr,size,n,fp));
	else
#endif	/* NV */
		return (fwrite(ptr,size,n,fp));
	}
#if	H_OPSYS == MS_DOS
#ifdef	NV

/************************************************************************/
/*									*/
/*	Routine:	w_getch						*/
/*									*/
/*	Interface function for "getch" library routine.  Having this	*/
/*	interface layer permits standard I/O from different nodes to	*/
/*	be read from, and written to, "logically different terminals".	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

static	VOID
w_getch()
	{
	int			rval;

	if(! _nv_getch(current_node,&rval))
		{
		q_add(q_getch);
		return;
		}
	reply_2b(rval);
	}

/************************************************************************/
/*									*/
/*	Routine:	w_getche					*/
/*									*/
/*	Interface function for "getche" library routine.  Having this	*/
/*	interface layer permits standard I/O from different nodes to	*/
/*	be read from, and written to, "logically different terminals".	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

static	VOID
w_getche()
	{
	int			rval;

	if(! _nv_getche(current_node,&rval))
		{
		q_add(q_getche);
		return;
		}
	reply_2b(rval);
	}
#endif	/* NV */
#endif	/* MS_DOS */
